/*
 mbits.c
  simple transfer routines (that are more difficult w. little endian CPUs)

   Copyright (C) 2014 Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.04.20
  Moved code from mblt to separate source file (this).
2014.04.21
  Implemented: temporary buffering method for overlap case, using _alloca
- rCopyBits only.  (It'd require double the code for a 'reverse' method.)
The equivalent of ScrollRect should work in all depths.
  Implemented NotCopy.
2014.04.22
  Complement bits for And & Or to simplify 'not' transfer
(makes or/and/bic modes slightly slower, but much less code as result).
*/

#include <string.h>
#include "mgraf.h"

/* ---- code */

void rCopyBits(int8 *a,
	const int8 *b,
	int dx,
	int sx,
	int c)
{
	uint32		k, v;

	a+= dx>>3, b+= sx>>3, dx&= 7, sx&= 7;
	if (dx!=sx)
	{ /* unaligned */
		const uint32	*s;
		uint32		*buf, *d, u, sm, ev;

		u = (int)a&3, a-= u, d = (uint32*)a, dx+= u<<3;
		u = (int)b&3, s = (const uint32*)(b-u), sx+= u<<3;
		sm = (uint32)((int)0x80000000>>(MIN(c,32)-1))>>dx;
		c+= dx, u = be32_native(*s);
		/* dx becomes 32-sx, and can be eliminated in assembly with shld */
		if ((sx-=dx)<0)
		{
			dx = 32+sx, sx =-sx, v = u>>sx&sm;
			v = be32_native(v)|(*d&~be32_native(sm));
			if ((c-=32)<=0)
				{*d = v; return;}
			buf = NULL, ev = c&31 ? d[(c>>5)+1]:0;
			if (d>s)
			{
				k = (c+64)>>3;
				if ((int8*)d<(int8*)s+k)
					d = buf = _alloca(k);
			}
			*d++ = v;
			if (c&32)
			{
				v = u<<dx, u = be32_native(s[1]), ++s;
				v|= u>>sx, *d++ = be32_native(v);
			}
			if ((k=c>>6)!=0)
				do
				{
					v = u<<dx, u = be32_native(s[1]);
					v|= u>>sx, *d = be32_native(v);
					v = u<<dx, u = be32_native(s[2]);
					v|= u>>sx, d[1] = be32_native(v);
				} while(s+=2,d+=2,--k);
			if (c&=31)
			{
				k =-1u>>c, v = (u<<dx|be32_native(s[1])>>sx)&~k;
				*d = be32_native(v)|(ev&be32_native(k)), ++d;
			}
			goto release;
		}
		dx = 32-sx;
		v = u<<sx, u = be32_native(s[1]), ++s;
		v = (v|u>>dx)&sm, v = be32_native(v)|(*d&~be32_native(sm));
		if ((c-=32)<=0)
			{*d = v; return;}
		buf = NULL, ev = c&31 ? d[(c>>5)+1]:0;
		if (d>s)
		{
			k = (c+64)>>3;
			if ((int8*)d<(int8*)s+k)
				d = buf = _alloca(k);
		}
		*d++ = v;
		if (c&32)
		{
			v = u<<sx, u = be32_native(s[1]), ++s;
			v|= u>>dx, *d++ = be32_native(v);
		}
		if ((k=c>>6)!=0)
			do
			{
				v = u<<sx, u = be32_native(s[1]);
				v|= u>>dx, *d = be32_native(v);
				v = u<<sx, u = be32_native(s[2]);
				v|= u>>dx, d[1] = be32_native(v);
			} while(s+=2,d+=2,--k);
		if (c&=31)
		{
			k =-1u>>c, v = (u<<sx|be32_native(s[1])>>dx)&~k;
			*d = be32_native(v)|(ev&be32_native(k)), ++d;
		}
	release:
		if (buf)
			memcpy(a,buf,(int8*)d-(int8*)buf);
		return;
	}
	/* aligned */
	if (a>b && a<b+sx)
	{ /* overlap */
		v = 0;
		if (dx)
		{
			k = MIN(c,7), c+= dx;
			k = (uint8)(~0xFF>>k)>>dx; /* src mask */
			v = (*a&~k)|(*b&k), ++a, ++b;
			if ((c-=8)<=0)
				{a[-1] = v; return;}
		}
		sx = c>>3;
		if (c&=7)
			k =~0xFF>>c, a[sx] = (a[sx]&~k)|(b[sx]&k);
		memmove(a,b,sx);
		if (dx)
			a[-1] = v;
	} else
	{
		if (dx)
		{
			k = MIN(c,7), c+= dx;
			k = (uint8)(~0xFF>>k)>>dx; /* src mask */
			*a = (*a&~k)|(*b&k), ++a, ++b;
			if ((c-=8)<=0)
				return;
		}
		sx = c>>3;
		memcpy(a,b,sx);
		if (c&=7)
			a+= sx, k =~0xFF>>c, *a = (*a&~k)|(b[sx]&k);
	}
}

void rNotCopyBits(int8 *a,
	const int8 *b,
	int dx,
	int sx,
	int c,
	uint32 cm)
{
	const uint32	*s;
	uint32		*d, k, v, u;

	a+= dx>>3, b+= sx>>3, dx&= 7, sx&= 7;
	u = (int)a&3, d = (uint32*)(a-u), dx+= u<<3;
	k = (uint32)((int)0x80000000>>(MIN(c,32)-1))>>dx;
	c+= dx;
#if env68k || envppc || envx86
	s = (const uint32*)(b-u), sx+= u<<3;
	if ((dx^sx)&7)
#else
	u = (int)b&3, s = (const uint32*)(b-u), sx+= u<<3;
	if (dx!=sx)
#endif
	{ /* unaligned */
		u = be32_native(*s)^cm;
		if ((sx-=dx)<0)
		{
			dx = 32+sx, sx =-sx, v = u>>sx&k;
			*d = be32_native(v)|(*d&~be32_native(k)), ++d;
			if ((c-=32)<=0)
				return;
			if (c&32)
			{
				v = u<<dx, u = be32_native(s[1])^cm, ++s;
				v|= u>>sx, *d++ = be32_native(v);
			}
			if ((k=c>>6)!=0)
				do
				{
					v = u<<dx, u = be32_native(s[1])^cm;
					v|= u>>sx, *d = be32_native(v);
					v = u<<dx, u = be32_native(s[2])^cm;
					v|= u>>sx, d[1] = be32_native(v);
				} while(s+=2,d+=2,--k);
			if (c&=31)
			{
				k =-1u>>c, v = (u<<dx|(be32_native(s[1])^cm)>>sx)&~k;
				*d = be32_native(v)|(*d&be32_native(k));
			}
			return;
		}
		dx = 32-sx;
		v = u<<sx, u = be32_native(s[1])^cm, ++s;
		v = (v|u>>dx)&k, *d = be32_native(v)|(*d&~be32_native(k)), ++d;
		if ((c-=32)<=0)
			return;
		if (c&32)
		{
			v = u<<sx, u = be32_native(s[1])^cm, ++s;
			v|= u>>dx, *d++ = be32_native(v);
		}
		if ((k=c>>6)!=0)
			do
			{
				v = u<<sx, u = be32_native(s[1])^cm;
				v|= u>>dx, *d = be32_native(v);
				v = u<<sx, u = be32_native(s[2])^cm;
				v|= u>>dx, d[1] = be32_native(v);
			} while(s+=2,d+=2,--k);
		if (c&=31)
		{
			k =-1u>>c, v = (u<<sx|(be32_native(s[1])^cm)>>dx)&~k;
			*d = be32_native(v)|(*d&be32_native(k));
		}
		return;
	}
	/* aligned */
	k = be32_native(k), *d = (*d&~k)|((*s^cm)&k), ++s, ++d;
	if ((c-=32)<=0)
		return;
	if (c&32)
		*d = *s^cm, ++s, ++d;
	if (c&64)
		*d = *s^cm, d[1] = s[1]^cm, s+= 2, d+= 2;
	if ((k=c>>7)!=0)
		do
			*d = *s^cm, d[1] = s[1]^cm, d[2] = s[2]^cm, d[3] = s[3]^cm;
		while(s+=4,d+=4,--k);
	if (c&=31)
		k =-1u>>c, k = be32_native(k), *d = (*d&k)|((*s^cm)&~k);
}

void rOrBits(int8 *a,
	const int8 *b,
	int dx,
	int sx,
	int c,
	uint32 cm)
{
	const uint32	*s;
	uint32		*d, k, v, u;

	a+= dx>>3, b+= sx>>3, dx&= 7, sx&= 7;
	u = (int)a&3, d = (uint32*)(a-u), dx+= u<<3;
	k = (uint32)((int)0x80000000>>(MIN(c,32)-1))>>dx;
	c+= dx;
#if env68k || envppc || envx86
	s = (const uint32*)(b-u), sx+= u<<3;
	if ((dx^sx)&7)
#else
	u = (int)b&3, s = (const uint32*)(b-u), sx+= u<<3;
	if (dx!=sx)
#endif
	{ /* unaligned */
		u = be32_native(*s)^cm;
		/* dx becomes 32-sx, and can be eliminated in assembly with shld */
		if ((sx-=dx)<0)
		{
			dx = 32+sx, sx =-sx, v = u>>sx&k;
			*d++|= be32_native(v);
			if ((c-=32)<=0)
				return;
			if (c&32)
			{
				v = u<<dx, u = be32_native(s[1])^cm, ++s;
				v|= u>>sx, *d++|= be32_native(v);
			}
			if ((k=c>>6)!=0)
				do
				{
					v = u<<dx, u = be32_native(s[1])^cm;
					v|= u>>sx, *d|= be32_native(v);
					v = u<<dx, u = be32_native(s[2])^cm;
					v|= u>>sx, d[1]|= be32_native(v);
				} while(s+=2,d+=2,--k);
			if (c&=31)
			{
				k =-1u>>c, v = (u<<dx|(be32_native(s[1])^cm)>>sx)&~k;
				*d|= be32_native(v);
			}
			return;
		}
		dx = 32-sx;
		v = u<<sx, u = be32_native(s[1])^cm, ++s;
		v = (v|u>>dx)&k, *d++|= be32_native(v);
		if ((c-=32)<=0)
			return;
		if (c&32)
		{
			v = u<<sx, u = be32_native(s[1])^cm, ++s;
			v|= u>>dx, *d++|= be32_native(v);
		}
		if ((k=c>>6)!=0)
			do
			{
				v = u<<sx, u = be32_native(s[1])^cm;
				v|= u>>dx, *d|= be32_native(v);
				v = u<<sx, u = be32_native(s[2])^cm;
				v|= u>>dx, d[1]|= be32_native(v);
			} while(s+=2,d+=2,--k);
		if (c&=31)
		{
			k =-1u>>c, v = (u<<sx|(be32_native(s[1])^cm)>>dx)&~k;
			*d|= be32_native(v);
		}
		return;
	}
	/* aligned */
	*d|= (*s^cm)&be32_native(k), ++s, ++d;
	if ((c-=32)<=0)
		return;
	if (c&32)
		*d|= *s^cm, ++s, ++d;
	if ((k=c>>6)!=0)
		do
			*d|= *s^cm, d[1]|= s[1]^cm;
		while(s+=2,d+=2,--k);
	if (c&=31)
		k =-1u>>c, *d|= (*s^cm)&~be32_native(k);
}

void rXorBits(int8 *a,
	const int8 *b,
	int dx,
	int sx,
	int c,
	uint32 cm)
{
	const uint32	*s;
	uint32		*d, k, v, u;

	a+= dx>>3, b+= sx>>3, dx&= 7, sx&= 7;
	u = (int)a&3, d = (uint32*)(a-u), dx+= u<<3;
	k = (uint32)((int)0x80000000>>(MIN(c,32)-1))>>dx;
	c+= dx;
#if env68k || envppc || envx86
	s = (const uint32*)(b-u), sx+= u<<3;
	if ((dx^sx)&7)
#else
	u = (int)b&3, s = (const uint32*)(b-u), sx+= u<<3;
	if (dx!=sx)
#endif
	{ /* unaligned */
		u = be32_native(*s)^cm;
		/* dx becomes 32-sx, and can be eliminated in assembly with shld */
		if ((sx-=dx)<0)
		{
			dx = 32+sx, sx =-sx, v = u>>sx&k;
			*d++^= be32_native(v);
			if ((c-=32)<=0)
				return;
			if (c&32)
			{
				v = u<<dx, u = be32_native(s[1])^cm, ++s;
				v|= u>>sx, *d++^= be32_native(v);
			}
			if ((k=c>>6)!=0)
				do
				{
					v = u<<dx, u = be32_native(s[1])^cm;
					v|= u>>sx, *d^= be32_native(v);
					v = u<<dx, u = be32_native(s[2])^cm;
					v|= u>>sx, d[1]^= be32_native(v);
				} while(s+=2,d+=2,--k);
			if (c&=31)
			{
				k =-1u>>c, v = (u<<dx|(be32_native(s[1])^cm)>>sx)&~k;
				*d^= be32_native(v);
			}
			return;
		}
		dx = 32-sx;
		v = u<<sx, u = be32_native(s[1])^cm, ++s;
		v = (v|u>>dx)&k, *d^= be32_native(v), ++d;
		if ((c-=32)<=0)
			return;
		if (c&32)
		{
			v = u<<sx, u = be32_native(s[1])^cm, ++s;
			v|= u>>dx, *d^= be32_native(v), ++d;
		}
		if ((k=c>>6)!=0)
			do
			{
				v = u<<sx, u = be32_native(s[1])^cm;
				v|= u>>dx, *d^= be32_native(v);
				v = u<<sx, u = be32_native(s[2])^cm;
				v|= u>>dx, d[1]^= be32_native(v);
			} while(s+=2,d+=2,--k);
		if (c&=31)
		{
			k =-1u>>c, v = (u<<sx|(be32_native(s[1])^cm)>>dx)&~k;
			*d^= be32_native(v);
		}
		return;
	}
	/* aligned */
	*d^= (*s^cm)&be32_native(k), ++s, ++d;
	if ((c-=32)<=0)
		return;
	if (c&32)
		*d^= *s^cm, ++s, ++d;
	if ((k=c>>6)!=0)
		do
			*d^= *s^cm, d[1]^= s[1]^cm;
		while(s+=2,d+=2,--k);
	if (c&=31)
		k =-1u>>c, *d^= (*s^cm)&~be32_native(k);
}

void rAndBits(int8 *a,
	const int8 *b,
	int dx,
	int sx,
	int c,
	uint32 cm)
{
	const uint32	*s;
	uint32		*d, k, v, u;

	a+= dx>>3, b+= sx>>3, dx&= 7, sx&= 7;
	u = (int)a&3, d = (uint32*)(a-u), dx+= u<<3;
	k = (uint32)((int)0x80000000>>(MIN(c,32)-1))>>dx;
	c+= dx;
#if env68k || envppc || envx86
	s = (const uint32*)(b-u), sx+= u<<3;
	if ((dx^sx)&7)
#else
	u = (int)b&3, s = (const uint32*)(b-u), sx+= u<<3;
	if (dx!=sx)
#endif
	{ /* unaligned */
		u = be32_native(*s)^cm;
		/* dx becomes 32-sx, and can be eliminated in assembly with shld */
		if ((sx-=dx)<0)
		{
			dx = 32+sx, sx =-sx, v = u>>sx|~k;
			*d++&= be32_native(v);
			if ((c-=32)<=0)
				return;
			if (c&32)
			{
				v = u<<dx, u = be32_native(s[1])^cm, ++s;
				v|= u>>sx, *d++&= be32_native(v);
			}
			if ((k=c>>6)!=0)
				do
				{
					v = u<<dx, u = be32_native(s[1])^cm;
					v|= u>>sx, *d&= be32_native(v);
					v = u<<dx, u = be32_native(s[2])^cm;
					v|= u>>sx, d[1]&= be32_native(v);
				} while(s+=2,d+=2,--k);
			if (c&=31)
			{
				k =-1u>>c, v = (u<<dx|(be32_native(s[1]^cm))>>sx)|k;
				*d&= be32_native(v);
			}
			return;
		}
		dx = 32-sx;
		v = u<<sx, u = be32_native(s[1])^cm, ++s;
		v = (v|u>>dx)|~k, *d++&= be32_native(v);
		if ((c-=32)<=0)
			return;
		if (c&32)
		{
			v = u<<sx, u = be32_native(s[1])^cm, ++s;
			v|= u>>dx, *d++&= be32_native(v);
		}
		if ((k=c>>6)!=0)
			do
			{
				v = u<<sx, u = be32_native(s[1])^cm;
				v|= u>>dx, *d&= be32_native(v);
				v = u<<sx, u = be32_native(s[2])^cm;
				v|= u>>dx, d[1]&= be32_native(v);
			} while(s+=2,d+=2,--k);
		if (c&=31)
		{
			k =-1u>>c, v = (u<<sx|(be32_native(s[1]^cm))>>dx)|k;
			*d&= be32_native(v);
		}
		return;
	}
	/* aligned */
	*d&= (*s^cm)|~be32_native(k), ++s, ++d;
	if ((c-=32)<=0)
		return;
	if (c&32)
		*d&= *s^cm, ++s, ++d;
	if ((k=c>>6)!=0)
		do
			*d&= *s^cm, d[1]&= s[1]^cm;
		while(s+=2,d+=2,--k);
	if (c&=31)
		k =-1u>>c, *d&= (*s^cm)|be32_native(k);
}
