/*
 mblt.h
  for the more low-level routines

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

/* ---- constants */

enum
{ /* cnvParam direct-pixel mask & shift table indices */
	_cvki,
	_cvkj,
	_cvmg,
	_cvmr,
	_cvsi,
	_cvsj,
	_cvdi,
	_cvdj,
	_cvub,
	_cvug,
	_cvur,
	_cvhb,
	_cvhg,
	_cvhr,
	_cvdSize
};

/* ---- typedefs */

#define DCL_BLT_RTN		__fastcall
#define DEF_BLT_RTN(n)		void DCL_BLT_RTN n(int32 o, int32 c, bltParam *f)
#define TYP_BLT_RTN(n)		void(DCL_BLT_RTN*n)(int32 o, int32 c, struct bltParam*)

#define DEF_CNV_RTN(n)		uint32 n(cnvParam *f, int o)
#define TYP_CNV_RTN(n)		uint32(*n)(struct cnvParam*, int)


/* alignment is done manually here */
#if (__SC__>0x800 || __MRC__ || macintosh)
 #pragma options align=mac68k
#else
 #pragma pack(push,2)
#endif

typedef struct cnvParam/*(66+)*/
{
	int32		cache[8]; /* pixel cache */
	int32		hx, rx; /* cached x & rate */
	int32		cm, mp; /* complement mask (not mode), pitch */
	TYP_CNV_RTN(rtn);
	const void	*s; /* source data */
	void		*t; /* 16 or 32-bit (always 32-bit when 'rtn' here is used) */
	const uint8	*itb;
	uint8		cc, ac; /* conversion & allocation code */
} cnvParam;

typedef struct bltParam
{
	rect32		cl; /* clip - may be narrower than the port clip */
/*x10*/	int32		ox, oy, oz; /* abs origin/size */
	int32		vx, vy, vz; /* src position/value */
	int32		sc; /* acting solid color pen */
	uint32		opr, opg, opb; /* r/g/b masks */
/*x38*/	cnvParam	cnv;
	int16		bmr, bmg, bmb; /* color magnitudes */
	uint32		mx, mh; /* pattern x-mask & height */
	const void	*sb; /* src base (non-solid) */
	int8		*p, *a, *e; /* dst pixels start, current and absolute end */
	TYP_BLT_RTN(rtn); /* vy is graphical-y when rtn is called */
	struct mlvParam	*mlc; /* midlevel convert-process-convert */
	const uint16	*dtb; /* indexed modes only */
} bltParam;

#if (__SC__>0x800 || __MRC__ || macintosh)
 #pragma options align=reset
#else
 #pragma pack(pop)
#endif

typedef struct mlvParam
{
	cnvParam	m2d, d2m;
	TYP_BLT_RTN(rtn);
	TYP_BLT_RTN(w2d);
	/* line pixel data follows */
} mlvParam;

typedef struct trapezoid
{
	int64		x1, dx1; /* manual point-x/slope */
	bltParam	prm;
} trapezoid;

typedef TYP_BLT_RTN(bltRoutine);
typedef TYP_CNV_RTN(cnvRoutine);

/* ---- macros, inlines */

/* 3 bits per channel minimum; only need to double twice (h=3; h'=6; h''=12) */
#define cnvDirect(v,u,t,r)	\
	(u=v<<t[_cvub], u|=u>>t[_cvsi], u|=u>>t[_cvdi], r=u>>t[_cvhb], \
	u=v<<t[_cvug]&t[_cvkj], u|=u>>t[_cvsj], u|=u>>t[_cvdj], r|=u>>t[_cvhg]&t[_cvmg], \
	u=v<<t[_cvur]&t[_cvki], u|=u>>t[_cvsi], u|=u>>t[_cvdi], r|=u>>t[_cvhr]&t[_cvmr])

#define cnvGray(v,u,t,r)	\
	(u=(v<<t[_cvur]&t[_cvki])>>16, u|=u>>t[_cvsi], u|=u>>t[_cvdi], \
	r=(v<<t[_cvug]&t[_cvkj])>>16, r|=r>>t[_cvsj], r|=r>>t[_cvdj], \
	v=v<<t[_cvub]>>16, v|=v>>t[_cvsi], v|=v>>t[_cvdi], formGray(u,r,v,u))

/* initialization */
__forceinline int rInitClip(const bltParam *f, /* must be valid, locked */
	int32 o,
	int8 **a)
{
	int32		c;

	*a = f->p, o-= c=f->cl.top;
	if (o>0)
		c+= o, o = 0;
	*a+= f->oz*(c-f->oy);
	return -o;
}

__forceinline int rInitClip2(const bltParam *f, /* must be valid, locked */
	int32 o,
	int32 c,
	int8 **a,
	int8 **e)
{
	*e = (*a=f->p)+f->oz*(c-f->oy);
	o-= c=f->cl.top;
	if (o>0)
		c+= o, o = 0;
	*a+= f->oz*(c-f->oy);
	return -o;
}

/* ---- prototypes */

/*
 rTrapezoid
	Renders a trapezoid.
	The prm field must be initialized with rInitBlt.
*/
void rTrapezoid(trapezoid*, /* must be valid */
	pt2dq p0,
	pt2dq p1);
/*
 rInitBlt
	Initializes a bltParam, with vertical clipping.
	It calls rGetBltRtn, and lockPixels as necessary.
  IMPORTANT:
	Unlike older versions of mgraf, rEndBlt must be used when done.
  Result:
	The clipping y-skip, or -1 on error or hollow.
*/
/* returns clipping y-skip */
int rInitBlt(bltParam *f, /* must be valid */
	grafPort *p, /* must be valid */
	const void *s, /* optional source data (otherwise the pen pattern/color) */
	const cTab64 *t, /* optional source table */
	int y0,
	int y1,
	int mode); /* OR in `formPxFmt(m)<<8` when passing src img data */
void rEndBlt(bltParam *f);
/*
 rEndLine
	In the case of source data, call this when a line ends.
  Result:
	Zero if f->s hit 'e'.
*/
int rEndLine(bltParam *f,
	const int8 *e);
/*
 rBltOrigin
	Initializes the image/pattern origin for a scanline.
*/
void rBltOrigin(bltParam *f,
	int x,
	int y);
/*
 rInitCnv
	Initializes a conversion routine and possible table (in f->t).
*/
int rInitCnv(cnvParam *f, /* must be valid; f->ac is assumed zero */
	pixMap *m, /* must be valid */
	const cTab64 *ct,
	int fmt,
	ufixed scale,
	uint8 mode,
	uint8 flag);
void rEndCnv(cnvParam *f);

mlvParam *rInitDithering(pixMap *dm);

/*
 rBlt
	Transfers an image set up with rInitBlt.
	The validity of f->s should be consistent with 's' passed to rInitBlt.
	If fHeight>0, the rate, drawn until f->s hits f->e.
	If fHeight<0, the height, upward (set s/d to first line past end).
*/
int rBlt(bltParam *f,
	int pitch,
	int width,
	ifixed fHeight,
	int x,
	int sx,
	int sy,
	int ky);
