/*
 mevent.h

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.04.03
  correction: bit-per-byte with keymap are lsb, not msb.
code only using _MKDOWN for the map will continue to work.
*/

#ifndef __MEVENT_H
 #define __MEVENT_H
 #if !TARGET_OS_MAC
  #include "mgraf.h"
 #endif

/* ---- constants */

/* modifiers */
#define altKey			0x100
#define rightBtnState		0x40
#define capsLock		0x400

/* flags native to Mac OS */
/* NOTE: button state bits are on if up */
#ifndef __EVENTS__
 #define charCodeMask		0xFF
 #define btnState		0x80	/* left mouse */
 #define cmdKey			0x100
 #define shiftKey		0x200
 #define alphaLock		0x400
 #define optionKey		0x800
 #define controlKey		0x1000
 #define rightShiftKey		0x2000
 #define rightOptionKey		0x4000
 #define rightControlKey	0x8000
 /* messages */
 #define activeFlag		0x1	/* activeEvt/mouseDown */
 #define changeFlag		0x2
 #define resumeFlag		0x1
 #define suspendResumeMessage	0x1	/* high 24 bits */
#endif

#define dblClickFlag		0x2	/* mouseDown msg */
#define wheelFlag		0x8	/* mouseUp msg */
#define wheelDownFlag		0x4	/* mouseUp msg */

enum
{
#ifndef __EVENTS__
	nullEvent,
	mouseDown,
	mouseUp,
	keyDown,
	keyUp,
	autoKey,
	updateEvt,
	diskEvt,
	activateEvt,
	osEvt				= 15,
#endif
	/* non-Mac OS */
	closeEvent			= 10
};

#ifndef __EVENTS__
enum
{
	mDownMask		= 1<<mouseDown,
	mUpMask			= 1<<mouseUp,
	keyDownMask		= 1<<keyDown,
	keyUpMask		= 1<<keyUp,
	autoKeyMask		= 1<<autoKey,
	updateMask		= 1<<updateEvt,
	activMask		= 1<<activateEvt,
	everyEvent		= 0xFFFF
};

enum
{
	kNullCharCode,
	kHomeCharCode,
	kEnterCharCode			= 3,
	kEndCharCode,
	kHelpCharCode,
	kBellCharCode			= 7,
	kBackspaceCharCode,
	kTabCharCode,
	kLineFeedCharCode,
	kPageUpCharCode,		/* also, VT */
	kPageDownCharCode,		/* also, FF */
	kReturnCharCode,
	kFunctionKeyCharCode		= 16,
	kCommandCharCode,
	kCheckCharCode,
	kDiamondCharCode,
	kAppleLogoCharCode,
	kEscapeCharCode			= 27,
	kClearCharCode			= 27,
	kLeftArrowCharCode,
	kRightArrowCharCode,
	kUpArrowCharCode,
	kDownArrowCharCode,
	kSpaceCharCode,
	kDeleteCharCode			= 127,
	kBulletCharCode			= 165,	/* MacRoman */
	kNonBreakingSpaceCharCode	= 202	/* MacRoman */
};
#endif

#define KBL			255	/* key-value blank/undefined */
#define MKC			160	/* mKeyMap bit count */

#define MKV_A			0
#define MKV_S			1
#define MKV_D			2
#define MKV_F			3
#define MKV_H			4
#define MKV_G			5
#define MKV_Z			6
#define MKV_X			7
#define MKV_C			8
#define MKV_V			9
#define MKV_NON_US_BSL		10
#define MKV_B			11
#define MKV_Q			12
#define MKV_W			13
#define MKV_E			14
#define MKV_R			15
#define MKV_Y			16
#define MKV_T			17
#define MKV_1			18
#define MKV_2			19
#define MKV_3			20
#define MKV_4			21
#define MKV_6			22
#define MKV_5			23
#define MKV_EQUALPLUS		24
#define MKV_9			25
#define MKV_7			26
#define MKV_HYPHEN_USC		27
#define MKV_8			28
#define MKV_0			29
#define MKV_SRBRACKET		30
#define MKV_O			31
#define MKV_U			32
#define MKV_SLBRACKET		33
#define MKV_I			34
#define MKV_P			35
#define MKV_RETURN		36
#define MKV_L			37
#define MKV_J			38
#define MKV_SQUOTE		39
#define MKV_K			40
#define MKV_SEMICOLON		41
#define MKV_BKSLASH		42
#define MKV_COMMA		43
#define MKV_FWSLASH		44
#define MKV_N			45
#define MKV_M			46
#define MKV_PERIOD		47
#define MKV_TAB			48
#define MKV_SPACE		49
#define MKV_GRAVE		50
#define MKV_BACKSPACE		51
#define MKV_PENTER		52
#define MKV_ESCAPE		53
#define MKV_MOD			54
#define MKV_LCMD		55 /* both L/R on old Macs */
#define MKV_LALT		55 /* both L/R on old Macs */
#define MKV_LSHIFT		56 /* both L/R on old Macs */
#define MKV_CAPSLOCK		57
#define MKV_OPT_APPS		58
#define MKV_LCONTROL		59 /* both L/R on old Macs */
#define MKV_RSHIFT		60
#define MKV_RCMD		61
#define MKV_RALT		61
#define MKV_RCONTROL		62
#define MKV_LGUI		63
#define MKV_F17			64
#define MKV_KPDECIMAL		65
#define MKV_RGUI		66
#define MKV_KPMULTIPLY		67
#define MKV_KPADD		69
#define MKV_CLRNUMLOCK		71
#define MKV_VOLUMEUP		72
#define MKV_VOLUMEDOWN		73
#define MKV_MUTE		74
#define MKV_KPDIVIDE		75
#define MKV_ENTER		76
#define MKV_KPMINUS		78
#define MKV_F18			79
#define MKV_F19			80
#define MKV_KPEQSEP		81
#define MKV_KP0			82
#define MKV_KP1			83
#define MKV_KP2			84
#define MKV_KP3			85
#define MKV_KP4			86
#define MKV_KP5			87
#define MKV_KP6			88
#define MKV_KP7			89
#define MKV_F20			90
#define MKV_KP8			91
#define MKV_KP9			92
#define MKV_YEN			93
#define MKV_RO_USC		94
#define MKV_KPCOMMA		95
#define MKV_F5			96
#define MKV_F6			97
#define MKV_F7			98
#define MKV_F3			99
#define MKV_F8			100
#define MKV_F9			101
#define MKV_EISU		102
#define MKV_F11			103
#define MKV_KANA_HANG		104
#define MKV_F13_PRNSCR		105
#define MKV_F16			106
#define MKV_F14_SCRLCK		107
#define MKV_F10			109
#define MKV_FN_ENTER		110
#define MKV_F12			111
#define MKV_F15_PAUSE		113
#define MKV_INS_HELP		114
#define MKV_HOME		115
#define MKV_PAGEUP		116
#define MKV_FWDELETE		117
#define MKV_F4			118
#define MKV_END			119
#define MKV_F2			120
#define MKV_PAGEDOWN		121
#define MKV_F1			122
#define MKV_ALEFT		123
#define MKV_ARIGHT		124
#define MKV_ADOWN		125
#define MKV_AUP			126
#define MKV_POWER		127
#define MKV_SELECT		128
#define MKV_LESS		129
#define MKV_KPLPAREN		130
#define MKV_KPRPAREN		131
#define MKV_F21			132
#define MKV_F22			133
#define MKV_F23			134
#define MKV_F24			135
#define MKV_JUNJA		136
#define MKV_FINAL		137
#define MKV_HANA_KANJI		138
#define MKV_AGAIN		139
#define MKV_FIND		140
#define MKV_EXECUTE		141
#define MKV_STOP		142
#define MKV_CLEAR		143
#define MKV_UNDO		144
#define MKV_CUT			145
#define MKV_COPY		146
#define MKV_PASTE		147
#define MKV_EJECT		148
#define MKV_MEDIAPLAY		149
#define MKV_MEDIASTOP		150
#define MKV_MEDIAPREV		151
#define MKV_MEDIANEXT		152
#define MKV_ACBACK		153
#define MKV_ACFORWARD		154
#define MKV_ACREFRESH		155
#define MKV_ACSTOP		156
#define MKV_ACSEARCH		157
#define MKV_ACFAVORITE		158
#define MKV_ACHOME		159

/* ---- macros */

#define _MKDOWN(m,i)		((m)[(i)>>3]&1<<((i)&7))
#define _MKSET(m,i)		(void)((m)[(i)>>3]|=1<<((i)&7))
#define _MKCLR(m,i)		(void)((m)[(i)>>3]&=~(1<<((i)&7)))

/* ---- typedefs */

typedef struct eventRecord
{
	uint16		what, modifiers;
	long int	message;
	uint32		when;
	pt2d16		where;
} eventRecord;

typedef struct timerEntry
{
	int64		mcsremaining; /* <=0 when inactive/done */
	void (*routine)(struct timerEntry*);
	void		*userInfo;
} timerEntry;

typedef void (*timerRoutine)(timerEntry*);

typedef struct timerThread
{
	int8		private[1];
} timerThread;

/* ---- globals */

extern eventRecord	mEventState;
extern uint8		mKeyMap[20];
extern const uint8	mkvtab_win[227]; /* amounting to 20 bytes, not Mac OS's 16 */

#ifdef WINUSERAPI
extern WNDCLASS	mEventWClass;
#endif

/* ---- prototypes */

#ifndef __EVENTS__
 typedef eventRecord	EventRecord;

 boolean GetNextEvent(uint32 mask,
	eventRecord *event);
 int PostMEvent(const eventRecord *event);
 void FlushEvents(uint32 mask,
	uint32 stopMask);

 uint32 GetCaretTime(void);

 void mOSVBL_timer(timerEntry *this);
 #ifndef __MT__
 void timerThreadIdle(timerThread *thd);
 #endif
#endif

uint8 *MGetKeys(uint8 keys[20]);

uint8 mKCHRproc(uint8 kchar,
	uint16 mod,
	uint8 *deadKey,
	uint8 *deadChar);

uint16 updateEventState(void);
uint16 eventModifiedSwitch(uint8 v);

timerThread *CreateTimerThread(int maxCount);
int EndTimerThread(timerThread*);

timerEntry *InstallTimer(timerThread*,
	timerRoutine rtn);
int RemoveTimer(timerThread*,
	timerEntry*);
int PrimeTimer(timerThread*, /* must be valid */
	timerEntry*, /* must be valid */
	int64 mcs);
#endif
