/*
 mkv.c
  Mapped-Key-Value, with initial basis on Frameworks/Events.h

  Avoid mapping keys that sound a little like the proper key.
 We don't want more official/modern keys going to the wrong function.
 e.g, it would be bad, just any easy key causing a POWER/RESET hit.
*/

#include <enviro64.h>
#include "mevent.h"
#include <ctype.h>

/* ---- constants */

#define BL KBL		/* blank */
#define MD MKV_MOD	/* modifier - requires proper update */

/* ---- globals */

const uint8	mkvtab_win[227] =
{
	BL, /* 0 - reserved/null */
	BL, /* VK_LBUTTON */
	BL, /* VK_RBUTTON */
	BL, /* VK_CANCEL */
	BL, /* 4 - VK_MBUTTON */
	BL, /* VK_XBUTTON1 */
	BL, /* VK_XBUTTON2 */
	BL, /* undef */
	51, /* 8 - VK_BACK (backspace) */
	48,
	BL, /* undef */
	BL,
	143, /* 12 - VK_CLEAR */
	36, /* VK_RETURN */
	BL,
	BL,
	MD, /* 16 - VK_SHIFT */
	MD, /* VK_CONTROL */
	MD, /* VK_MENU */
	113, /* VK_PAUSE - F15, pulse-only on pre-Windows 7 */
	57, /* 20 - VK_CAPITAL */
	104, /* VK_KANA/HANGEUL/HANGUL */
	BL, /* undef */
	136, /* VK_JUNJA */
	137, /* 24 - VK_FINAL */
	138, /* VK_HANJA/KANJI */
	BL, /* undef */
	53, /* VK_ESCAPE */
	BL, /* 28 - VK_CONVERT */
	BL, /* VK_NONCONVERT */
	137, /* VK_ACCEPT */
	BL, /* VK_MODECHANGE */
	49, /* 32 - VK_SPACE */
	116, /* VK_PRIOR */
	121, /* VK_NEXT */
	119, /* VK_END */
	115, /* 36 - VK_HOME */
	123, /* VK_LEFT */
	126, /* VK_UP */
	124, /* VK_RIGHT */
	125, /* 40 - VK_DOWN */
	128, /* VK_SELECT */
	BL, /* VK_PRINT */
	141, /* VK_EXECUTE */
	105, /* 44 - VK_SNAPSHOT - F13, pulse-only */
	114, /* VK_INSERT */
	117, /* VK_DELETE */
	112, /* VK_HELP */
	29, /* 48 - VK_0 */
	18, /* VK_1 */
	19, /* VK_2 */
	20, /* VK_3 */
	21, /* 52 - VK_4 */
	23, /* VK_5 */
	22, /* VK_6 */
	26, /* VK_7 */
	28, /* 56 - VK_8 */
	25, /* VK_9 */
	BL,
	BL,
	BL, /* 60 - undef */
	BL,
	BL,
	BL,
	BL, /* 64 - undef */
	0, /* VK_A */
	11, /* VK_B */
	8, /* VK_C */
	2, /* 68 - VK_D */
	14, /* VK_E */
	3, /* VK_F */
	5, /* VK_G */
	4, /* 72 - VK_H */
	34, /* VK_I */
	38, /* VK_J */
	40, /* VK_K */
	37, /* 76 - VK_L */
	46, /* VK_M */
	45, /* VK_N */
	31, /* VK_O */
	35, /* 80 - VK_P */
	12, /* VK_Q */
	15, /* VK_R */
	1, /* VK_S */
	17, /* 84 - VK_T */
	32, /* VK_U */
	9, /* VK_V */
	13, /* VK_W */
	7, /* 88 - VK_X */
	16, /* VK_Y */
	6, /* VK_Z */
	63, /* VK_LWIN */
	54, /* 92 - VK_RWIN */
	58, /* VK_APPS - to option */
	BL, /* undef */
	BL, /* VK_SLEEP - not wise to map this, I take it */
	82, /* 96 - VK_NUMPAD0 */
	83, /* VK_NUMPAD1 */
	84, /* VK_NUMPAD2 */
	85, /* VK_NUMPAD3 */
	86, /* 100 - VK_NUMPAD4 */
	87, /* VK_NUMPAD5 */
	88, /* VK_NUMPAD6 */
	89, /* VK_NUMPAD7 */
	91, /* 104 - VK_NUMPAD8 */
	92, /* VK_NUMPAD9 */
	67, /* VK_MULTIPLY/kp* */
	69, /* VK_ADD/kp+ */
	81, /* 108 - VK_SEPARATOR --> KP= */
	78, /* VK_SUBTRACT/kp- */
	65, /* VK_DECIMAL/kp. */
	75, /* VK_DIVIDE/kpslash */
	122, /* 112 - VK_F1 */
	120, /* VK_F2 */
	99, /* VK_F3 */
	118, /* VK_F4 */
	96, /* 116 - VK_F5 */
	97, /* VK_F6 */
	98, /* VK_F7 */
	100, /* VK_F8 */
	101, /* 120 - VK_F9 */
	109, /* VK_F10 */
	103, /* VK_F11 */
	111, /* VK_F12 */
	105, /* 124 - VK_F13 - these keys may not work the same */
	107, /* VK_F14 */
	113, /* VK_F15 - must be done manually */
	106, /* VK_F16 */
	64, /* 128 - VK_F17 */
	79, /* VK_F18 */
	80, /* VK_F19 */
	90, /* VK_F20 */
	132, /* 132 - VK_F21 */
	133, /* VK_F22 */
	134, /* VK_F23 */
	135, /* VK_F24 */
	BL, /* 136 - undef */
	BL,
	BL,
	BL,
	BL, /* 140 */
	BL,
	BL,
	BL,
	71, /* 144 - async:VK_NUMLOCK */
	107, /* async:VK_SCROLL */
	BL, /* VK_OEM_NEC_EQUAL/FJ_JISHO */
	BL, /* VK_OEM_FJ_MASSHOU */
	BL, /* 148 - VK_OEM_FJ_TOUROKU */
	BL, /* VK_OEM_FJ_LOYA */
	BL, /* VK_OEM_FJ_ROYA */
	BL,
	BL, /* 152 - undef */
	BL,
	BL,
	BL,
	BL, /* 156 - undef */
	BL,
	BL,
	BL,
	56, /* 160 - VK_LSHIFT */
	60, /* VK_RSHIFT */
	59, /* VK_LCONTROL */
	62, /* VK_RCONTROL */
	55, /* 164 - VK_LMENU */
	61, /* VK_RMENU */
	153, /* VK_BROWSER_BACK */
	154, /* VK_BROWSER_FORWARD */
	155, /* 168 - VK_BROWSER_REFRESH */
	156, /* VK_BROWSER_STOP */
	157, /* VK_BROWSER_SEARCH */
	158, /* VK_BROWSER_FAVORITES */
	159, /* 172 - VK_BROWSER_HOME */
	74, /* VK_VOLUME_MUTE */
	73, /* VK_VOLUME_DOWN */
	72, /* VK_VOLUME_UP */
	152, /* 176 - VK_MEDIA_NEXT_TRACK */
	151, /* VK_MEDIA_PREV_TRACK */
	150, /* VK_MEDIA_STOP */
	149, /* VK_MEDIA_PLAY_PAUSE */
	BL, /* 180 - VK_LAUNCH_MAIL */
	BL, /* VK_LAUNCH_MEDIA_SELECT */
	BL, /* VK_LAUNCH_APP1 */
	BL, /* VK_LAUNCH_APP2 */
	BL, /* 184 - undef */
	BL,
	41, /* VK_OEM_1 */
	24, /* VK_OEM_PLUS */
	43, /* 188 - VK_OEM_COMMA */
	27, /* VK_OEM_MINUS */
	47, /* VK_OEM_PERIOD */
	44, /* VK_OEM_2/slash */
	50, /* 192 - VK_OEM_3/grave */
	BL, /* undef */
	BL,
	BL,
	BL, /* 196 - undef */
	BL,
	BL,
	BL,
	BL, /* 200 - undef */
	BL,
	BL,
	BL,
	BL, /* 204 */
	BL,
	BL,
	BL,
	BL, /* 208 */
	BL,
	BL,
	BL,
	BL, /* 212 */
	BL,
	BL,
	BL,
	BL, /* 216 */
	BL,
	BL,
	33, /* VK_OEM_4/lbracket */
	42, /* 220 - VK_OEM_5/bkslash */
	30, /* VK_OEM_6/rbracket */
	39, /* VK_OEM_7/squote */
	BL, /* VK_OEM_8 */
	BL, /* 224 - undef */
	BL, /* VK_OEM_AX */
	129, /* VK_OEM_102/less */
};	/* VK_ICO_HELP */
	/* 228 - VK_ICO_00 */
	/* VK_PROCESSKEY */
	/* VK_ICO_CLEAR */
	/* VK_PACKET */
	/* 232 - undef */
	/* VK_OEM_RESET */
	/* VK_OEM_JUMP */
	/* VK_OEM_PA1 */
	/* 236 - VK_OEM_PA2 */
	/* VK_OEM_PA3 */
	/* VK_OEM_WSCTRL */
	/* VK_OEM_CUSEL */
	/* 240 - VK_OEM_ATTN */
	/* VK_OEM_FINISH */
	/* VK_OEM_COPY */
	/* VK_OEM_AUTO */
	/* 244 - VK_OEM_ENLW */
	/* VK_OEM_BACKTAB */
	/* VK_ATTN */
	/* VK_CRSEL */
	/* 248 - VK_EXSEL */
	/* VK_EREOF */
	/* VK_PLAY */
	/* VK_ZOOM */
	/* 252 - VK_NONAME */
	/* VK_PA1 */
	/* VK_OEM_CLEAR */

/* ---- code */

const uint8	KCHR_US[] =
{
 /* modifier table index */
	0,0,1,0,2,0,1,0,3,6,4,4,5,6,4,4,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	1,0,1,0,1,0,1,0,4,4,4,4,4,4,4,4,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	3,6,4,4,5,6,4,4,3,6,4,4,5,6,4,4,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
	4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
	7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,
 /* scan table */
	7, /* table count-1 */
/*00*/	97,115,100,102,104,103,122,120,99,118,164,98,113,119,101,114,
/*10*/	121,116,49,50,51,52,54,53,61,57,55,45,56,48,93,111,
/*20*/	117,91,105,112,13,108,106,39,107,59,92,44,47,110,109,46,
/*30*/	9,32,96,8,3,27,0,0,0,0,0,0,0,0,0,0,
/*40*/	0,46,29,42,0,43,28,27,31,0,0,47,3,30,45,0,
/*50*/	0,61,48,49,50,51,52,53,54,55,0,56,57,0,0,0,
/*60*/	16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
/*70*/	16,16,5,1,11,127,16,4,16,12,16,28,29,31,30,0,
/*100*/	65,83,68,70,72,71,90,88,67,86,177,66,81,87,69,82,
/*110*/	89,84,33,64,35,36,94,37,43,40,38,95,42,41,125,79,
/*120*/	85,123,73,80,13,76,74,34,75,58,124,60,63,78,77,62,
/*130*/	9,32,126,8,3,27,0,0,0,0,0,0,0,0,0,0,
/*140*/	0,46,42,42,0,43,43,27,61,0,0,47,3,47,45,0,
/*150*/	0,61,48,49,50,51,52,53,54,55,0,56,57,0,0,0,
/*160*/	16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
/*170*/	16,16,5,1,11,127,16,4,16,12,16,28,29,31,30,0,
/*200*/	65,83,68,70,72,71,90,88,67,86,164,66,81,87,69,82,
/*210*/	89,84,49,50,51,52,54,53,61,57,55,45,56,48,93,79,
/*220*/	85,91,73,80,13,76,74,39,75,59,92,44,47,78,77,46,
/*230*/	9,32,96,8,3,27,0,0,0,0,0,0,0,0,0,0,
/*240*/	0,46,29,42,0,43,28,27,31,0,0,47,3,30,45,0,
/*250*/	0,61,48,49,50,51,52,53,54,55,0,56,57,0,0,0,
/*260*/	16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
/*270*/	16,16,5,1,11,127,16,4,16,12,16,28,29,31,30,0,
/*300*/	140,167,182,196,250,169,189,197,141,195,164,186,207,183,0,168,
/*310*/	180,160,193,170,163,162,164,176,173,187,166,208,165,188,212,191,
/*320*/	0,210,0,185,13,194,198,190,251,201,199,178,214,0,181,179,
/*330*/	9,202,0,8,3,27,0,0,0,0,0,0,0,0,0,0,
/*340*/	0,46,29,42,0,43,28,27,31,0,0,47,3,30,45,0,
/*350*/	0,61,48,49,50,51,52,53,54,55,0,56,57,0,0,0,
/*360*/	16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
/*370*/	16,16,5,1,11,127,16,4,16,12,16,28,29,31,30,0,
/*400*/	129,234,235,236,238,253,252,254,130,215,177,245,206,227,171,228,
/*410*/	231,255,218,219,220,221,223,222,177,225,224,209,161,226,213,175,
/*420*/	172,211,246,184,13,241,239,174,240,242,200,248,192,247,229,249,
/*430*/	9,202,96,8,3,27,0,0,0,0,0,0,0,0,0,0,
/*440*/	0,46,42,42,0,43,43,27,61,0,0,47,3,47,45,0,
/*450*/	0,61,48,49,50,51,52,53,54,55,0,56,57,0,0,0,
/*460*/	16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
/*470*/	16,16,5,1,11,127,16,4,16,12,16,28,29,31,30,0,
/*500*/	129,234,235,236,238,169,189,197,130,195,164,245,206,183,171,168,
/*510*/	231,160,193,170,163,162,164,176,173,187,166,208,165,188,212,175,
/*520*/	172,210,246,184,13,241,239,174,251,201,199,178,214,247,229,179,
/*530*/	9,202,96,8,3,27,0,0,0,0,0,0,0,0,0,0,
/*540*/	0,46,29,42,0,43,28,27,31,0,0,47,3,30,45,0,
/*550*/	0,61,48,49,50,51,52,53,54,55,0,56,57,0,0,0,
/*560*/	16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
/*570*/	16,16,5,1,11,127,16,4,16,12,16,28,29,31,30,0,
/*600*/	140,167,182,196,250,169,189,197,141,195,164,186,207,183,171,168,
/*610*/	180,160,193,170,163,162,164,176,173,187,166,208,165,188,212,191,
/*620*/	172,210,94,185,13,194,198,190,251,201,199,178,214,126,181,179,
/*630*/	9,202,96,8,3,27,0,0,0,0,0,0,0,0,0,0,
/*640*/	0,46,29,42,0,43,28,27,31,0,0,47,3,30,45,0,
/*650*/	0,61,48,49,50,51,52,53,54,55,0,56,57,0,0,0,
/*660*/	16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
/*670*/	16,16,5,1,11,127,16,4,16,12,16,28,29,31,30,0,
/*700*/	1,19,4,6,8,7,26,24,3,22,48,2,17,23,5,18,
/*710*/	25,20,49,50,51,52,54,53,61,57,55,31,56,48,29,15,
/*720*/	21,27,9,16,13,12,10,39,11,59,28,44,47,14,13,46,
/*730*/	9,32,96,8,3,27,0,0,0,0,0,0,0,0,0,0,
/*740*/	0,46,29,42,0,43,28,27,31,0,0,47,3,30,45,0,
/*750*/	0,61,48,49,50,51,52,53,54,55,0,56,57,0,0,0,
/*760*/	16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,16,
/*770*/	16,16,5,1,11,127,16,4,16,12,16,28,29,31,30,0,
 /* dead keys */
	4, /* table count-1 */
 /* option-mkv; scan/replace table [count-1]; default char */
	14,9, 65,231, 69,131, 73,234, 79,238, 85,242, 97,135,
	101,142, 105,146, 111,151, 117,156, 171,

	50,9, 65,203, 69,233, 73,237, 79,241, 85,244, 97,136,
	101,143, 105,147, 111,152, 117,157, 96,

	34,9, 65,229, 69,230, 73,235, 79,239, 85,243, 97,137,
	101,144, 105,148, 111,153, 117,158, 246,

	32,11, 65,128, 69,232, 73,236, 79,133, 85,134, 89,217,
	97,138, 101,145, 105,149, 111,154, 117,159, 121,216, 172,

	45,5, 65,204, 78,132, 79,205, 97,139, 110,150, 111,155, 247
};

const uint8	*_curKCHR = KCHR_US;

uint8 mKCHRproc(uint8 mkv,
	uint16 mod,
	uint8 *deadKey,
	uint8 *deadChar)
{
	const uint8	*t = _curKCHR;
	int		i;
	uint8		chr;

	i = t[mod>>8&31];
	if (deadChar)
		*deadChar = 0;
	if ((chr=mkv>127 ? 0:t[129+128*i+mkv])==0 && mod&optionKey)
	{ /* scan dead keys */
		if (!deadKey)
			goto end;
		if (mkv-MKV_MOD<MKV_F17-MKV_MOD+0u)
			goto end;
		t+= 257+128*t[128], i = *t+1, ++t;
		do
			if (*t==mkv)
				{*deadKey = mkv; goto end;}
		while(t+=5+t[1]*2,--i);
	} else if (deadKey && deadChar && *deadKey)
	{
		if (mkv-MKV_MOD<MKV_F17-MKV_MOD+0u)
			goto end;
		t+= 257+128*t[128], i = *t+1, ++t;
		do
			if (*t==*deadKey)
			{
				i = t[1]+1, t+= 2;
				do
					if (*t==chr)
						{chr = t[1]; goto cdk;}
				while(t+=2,--i);
				*deadChar = *t;
				break;
			}
		while(t+=5+t[1]*2,--i);
	}
cdk:	if (deadKey)
		*deadKey = 0;
end:	return chr;
}
