/*
 mtbox.c
  text boxes

   Revised: 2015.01.05
    Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.

2014.06.29
  Split from mtext.  Now relies on window input, not merely a grafPort.
*/

#include <string.h>

#include "../utility/interpreter.h"
#include "mwind.h"
#include "mblt.h"
#include "mtext.h"
#include "mevent.h"

/* ---- macros, inlines */

__forceinline static const bmfMemory *styleRunFont(const tStyleRun *sr)
{
	return ffMatch(fmFonts[sr->font>=GET_ARRAY_COUNT(fmFonts) ? 0:sr->font],
		sr->size,sr->style);
}

/* ---- code */

static DEF_WDB_RTN(defWordBreak)
{
	const uint8	*s, *f = (uint8*)tbox->text;
	uint32		z = tbox->length; 
	int		m = (!reverse)*2-1;

	for(s=f+i; s-f+0u<z; s+=m)
		if (interpUCType[*s]>(int8)0xE3) /* f-hook (see uctype.h) */
		{
			if ((*s=='\'' || interpUCType[*s]==0x63)
					&& s+m-f+0u<z && interpUCType[s[m]]-0x8A<0xC2u-0x8A)
				continue;
			s+= m<0;
			break;
		}
	i = s-f;
	return i+0u>=z ? ~i>>31&z:i; /* pin */
}

const tStyleRun *tBoxCharStyle(textBoxObj *tbox,
	int32 i,
	int *remCount)
{
	static tStyleRun	defStyle;

	const tStyleRun	*s, *f;
	int		c;

	if (!tbox)
		return NULL;
	c = 0;
	if ((s=tbox->styles)!=NULL)
		c = tbox->styleCount;
	if (c<=0)
	{
	#ifdef USE_MYDIALOGS
	 extern GrafPtr _curPort;
		defStyle.font = _curPort->txFont;
	#else
		defStyle.font = ffDefault;
	#endif
		s = &defStyle, c = 1;
	}
	if (i>0 && c>1)
	{
		while(c>4)
			if (i<(f=s+(c>>1))->offs)
				c = (c+1)>>1;
			else
				s = f, c-= c>>1;
		do
			if (i<s->offs)
				break;
		while(++s,--c);
		--s;
	} else
		--c;
	if (remCount)
		*remCount = c;
	return s;
}

int tBoxRecalc(textBoxObj *tbox,
	int startLine)
{
	const tStyleRun	*sr;
	const bmfMemory	*fm;
	tLineRun	*lr;
	const char	*s;
	ufixed		w, qx;
	int32		z, i = 0, c;
	int		r, y, li, lc, lh, la, mxlh, mxla, sc = 0;

	if (!tbox)
		return -1;
	if ((z=tbox->length)<=0)
	{
		tbox->lineCount = tbox->tHeight = 0;
		if (tbox->lines)
			{free(tbox->lines); tbox->lines = NULL;}
		return 0;
	}
	if (!tbox->lines)
		tbox->lineCount = 0;
	if (startLine>=tbox->lineCount+0u)
		startLine = 0;
	mxla = mxlh = la = lh = lc = li = qx = y = 0;
	if (startLine>0)
	{
		lr = tbox->lines+startLine;
		li = i = lr->offs, lc = startLine;
		y = lr[-1].vpos+lr[-1].descent;
	}
	sr = tBoxCharStyle(tbox,i,&sc);
	s = tbox->text;
	goto start;
	do
	{
		if (sc>0 && i>=sr[1].offs)
		{
			--sc, ++sr;
		start:	fm = styleRunFont(sr);
			if ((la=fm->leading+fm->ascent)>mxla)
				mxla = la;
			if ((lh=la+fm->descent)>mxlh)
				mxlh = lh;
		}
		w = tbox->maxWidth, c = (sc ? MIN(sr[1].offs,z):z)-i;
		r = rTextBreak(fm,s+li,s+i,&c,&w,qx,tbox->tabAlign,0,0,sr->style);
		i+= c;
		if (r>tNoBreak)
		{
			if ((lr=realloc(tbox->lines,++lc*sizeof(tLineRun)))==NULL)
				return -2;
			tbox->lines = lr, lr+= lc-1;
			lr->offs = li, lr->length = (i-li-(r==tBreakLine))|(r<=tBreakLine)<<31;
			lr->width = w>>16, lr->vpos = y+mxla, lr->descent = mxlh-mxla;
			y+= mxlh, mxla = la, mxlh = lh;
			qx = w = 0, li = i;
		} else
			qx = w;
	} while(i<z);
	if ((lr=realloc(tbox->lines,++lc*sizeof(tLineRun)))==NULL)
		return -2;
	tbox->lineCount = lc;
	tbox->lines = lr, lr+= lc-1;
	lr->offs = li, lr->length = (i-li)|1<<31;
	lr->width = w>>16, lr->vpos = y+mxla, lr->descent = mxlh-mxla;
	tbox->tHeight = y+= mxlh;
	return 1;
}

int tBoxSetMargins(textBoxObj *tbox,
	const rect16 *margins)
{
	int		v;

	if (!tbox)
		return -1;
	tbox->txbounds = tbox->bounds;
	if (margins)
	{
		tbox->txbounds.top = (v=tbox->txbounds.top+margins->top)!=(int16)v
			? v>>31^IMAX(int16):v;
		tbox->txbounds.left = (v=tbox->txbounds.left+margins->left)!=(int16)v
			? v>>31^IMAX(int16):v;
		tbox->txbounds.bottom = (v=tbox->txbounds.bottom-margins->bottom)!=(int16)v
			? v>>31^IMAX(int16):v;
		tbox->txbounds.right = (v=tbox->txbounds.right-margins->right)!=(int16)v
			? v>>31^IMAX(int16):v;
	}
	v = RECTANGLE_WIDTH(&tbox->txbounds);
	tbox->maxWidth = MAX(v,0)<<16;
	return !!tbox->maxWidth;
}

int tBoxMaxScroll(textBoxObj *tbox,
	int *maxv,
	int *maxh)
{
	int		v;

	if (!tbox)
		return -1;
	if (maxv)
	{
		if ((v=tbox->tHeight-RECTANGLE_HEIGHT(&tbox->txbounds))<0)
			v = 0;
		*maxv = v;
	} else if (!maxh)
		return 0;
	if (maxh)
	{
		v = tbox->maxWidth, v = (v>>16)+!!(uint16)v-RECTANGLE_WIDTH(&tbox->txbounds);
		*maxh = v;
	}
	return 1;
}

static void tBoxScrollBars(textBoxObj *tbox,
	int forceUpdate)
{
	if (tbox->ctlScrollH)
		if (forceUpdate)
		{
			tbox->ctlScrollH->value = tbox->scrollH;
			wDraw1Control(tbox->ctlScrollH);
		} else
			wSetControlValue(tbox->ctlScrollH,tbox->scrollH);
	if (tbox->ctlScrollV)
		if (forceUpdate)
		{
			tbox->ctlScrollV->value = tbox->scrollV;
			wDraw1Control(tbox->ctlScrollV);
		} else
			wSetControlValue(tbox->ctlScrollV,tbox->scrollV);
}

int tBoxScroll(textBoxObj *tbox,
	int amtv,
	int amth)
{
	int32		v, maxv, maxh, oldv, oldh;

	if ((v=tBoxMaxScroll(tbox,&maxv,&maxh))<=0)
		return v;
	oldv = tbox->scrollV, oldh = tbox->scrollH;
	if ((tbox->scrollV+=amtv)>maxv+0u)
		tbox->scrollV =~tbox->scrollV>>31&maxv;
	if ((tbox->scrollH+=amth)>maxh+0u)
		tbox->scrollH =~tbox->scrollH>>31&maxh;
	/* TODO: perform a pixel scroll, render less... */
	if ((tbox->scrollV^oldv)|(tbox->scrollH^oldh))
	{
		tBoxScrollBars(tbox,0);
		tBoxDraw(tbox);
		if (tbox->wind)
		{
			rect16		b = tbox->wind->port.clip;

			if (SectRect16(&b,&tbox->bounds))
			{
				update_screen_window_rect(tbox->wind,b.left,b.top,b.right,b.bottom);
				return 2;
			}
		}
		return 1;
	}
	return 0;
}

const tLineRun *tVisLine(const tLineRun *s,
	int c,
	int y)
{
	const tLineRun	*f, *e = s+c;

	if (!s)
		return NULL;
	while(c>4)
		if (y<(f=s+(c>>1))->vpos+f->descent)
			c = (c+1)>>1;
		else
			s = f, c-= c>>1;
	do
		if (y<s->vpos+s->descent)
			return s;
	while(++s,--c);
	return s>=e ? NULL:s;
}

static ufixed tLineMeasure(const textBoxObj *tbox,
	const tLineRun *lr,
	uint8 align,
	ufixed *chExtra,
	ufixed *spExtra)
{
	ufixed		w, qx = 0;

	*chExtra = *spExtra = 0, w = lr ? lr->width<<16:0;
	if ((align&=_special_align_mask) && w<tbox->maxWidth)
	{
		int32		c;

		w = tbox->maxWidth-w;
		if (align==_special_align_justify)
		{
			if ((c=lr ? lr->length:0)>0)
			{
				uint32		denom = memcnt(tbox->text+lr->offs,' ',c);

				if (denom>1)
				{
					--denom;
					/* half & half */
					*spExtra = ((w>>1)+denom-1)/denom;
					w-= w>>1;
				}
				if (c>1)
					--c;
				*chExtra = (w+c-1)/c;
			}
		} else if (align==_special_align_right)
			qx = w;
		else if (align==_special_align_center)
			qx = w>>1;
	}
	return qx;
}

int32 tBoxPtToChar(textBoxObj *tbox,
	ufixed x,
	int y)
{
	const tLineRun	*lr;
	const bmfMemory	*fm;
	const tStyleRun	*sr;
	const char	*s;
	ufixed		w, qx, chExtra, spExtra;
	int32		i, z;
	int		c, lz, sc;

	if (!tbox)
		return -1;
	if (!tbox->lines && tBoxRecalc(tbox,0)<0)
		return -2;
	z = tbox->length;
	if (y<0)
		return 0;
	if ((lr=tVisLine(tbox->lines,tbox->lineCount,y))==NULL)
		return z;
	i = lr->offs;
	if (!x || (s=tbox->text)==NULL)
		return i;
	lz = lr->length&IMAX(int);
	sr = tBoxCharStyle(tbox,i,&sc);
	qx = tLineMeasure(tbox,lr,sr->special,&chExtra,&spExtra);
	goto start;
	do
	{
		if (sc>0 && i>=sr[1].offs)
		{
			--sc, ++sr;
		start:	fm = styleRunFont(sr);
		}
		w = x, c = z;
		if (sc && sr[1].offs<c)
			c = sr[1].offs;
		if ((c-=i)>lz)
			c = lz;
		c = rClipText(fm,s+i,c,&w,qx,tbox->tabAlign,0,chExtra,spExtra,sr->style,1);
		qx = w;
	} while(i+=c,(lz-=c)>0 && w<x && i<z);
	w =-1u;
	if (qx>x)
	{ /* round to nearest edge; only rClipText will get the full char width */
		rClipText(fm,s+i-1,1,&w,0,tbox->tabAlign,0,chExtra,spExtra,sr->style,1);
		qx-= w>>1, i-= x<qx;
	}
	return i;
}

const tLineRun *tBoxCharLine(textBoxObj *tbox,
	int32 offs,
	ufixed *x,
	int *top,
	int *bottom)
{
	tLineRun	tln;
	const bmfMemory	*fm;
	const tStyleRun	*sr;
	const tLineRun	*lr, *f;
	int		c;

	if (!tbox || offs>tbox->length+0u)
		return NULL;
	if (!tbox->lines && tBoxRecalc(tbox,0)<0)
		return NULL;
	if ((lr=tbox->lines)==NULL)
	{
		memset(&tln,0,sizeof(tln));
		fm = styleRunFont(tBoxCharStyle(tbox,0,NULL));
		tln.vpos = fm->leading+fm->ascent;
		tln.descent = fm->descent;
		lr = &tln;
	}
	if ((c=tbox->lineCount)>0)
	{
		while(c>4)
			if (offs<(f=lr+(c>>1))->offs)
				c = (c+1)>>1;
			else
				lr = f, c-= c>>1;
		do
			if (offs<lr->offs)
				break;
		while(++lr,--c);
		--lr;
	}
	if (top)
		*top = (lr==&tln || lr<=tbox->lines ? 0:lr[-1].vpos+lr[-1].descent)
			+tbox->txbounds.top-tbox->scrollV;
	if (bottom)
		*bottom = lr->vpos+lr->descent+tbox->txbounds.top-tbox->scrollV;
	if (x)
	{
		const char	*s;
		ufixed		w, qx, chExtra, spExtra;
		int32		r, i = lr->offs;
		int		lz, sc;

		lz = lr->length&IMAX(int);
		sr = tBoxCharStyle(tbox,i,&sc);
		qx = tLineMeasure(tbox,lr,sr->special,&chExtra,&spExtra);
		s = tbox->text;
		goto start;
		do
		{
			if (sc>0 && i>=sr[1].offs)
			{
				--sc, ++sr;
			start:	fm = styleRunFont(sr);
			}
			c = offs;
			if (sc && sr[1].offs<c)
				c = sr[1].offs;
			if ((c-=i)>lz)
				c = lz;
			w =-1u;
			if ((r=rClipText(fm,s+i,c,&w,qx,tbox->tabAlign,0,
					chExtra,spExtra,sr->style,0))!=c)
				break;
			qx = w;
		} while(i+=r,(lz-=c)>0 && i<offs);
		*x = MIN(qx,tbox->maxWidth-0x8000);
	}
	return lr;
}

int tBoxUpdate(textBoxObj *tbox,
	const rect16 *area,
	int firstLine)
{
	rect16		oldclip;
	grafPort	*p;
	int		i;
	boolean		upd;

	if (!tbox)
		return -1;
	if (!tbox->wind)
		return 0;
	p = &tbox->wind->port, oldclip = p->clip;
	if (area)
		SectRect16(&p->clip,area);
	upd = 0;
	if (firstLine>=0)
	{
		if (firstLine<IMAX(int))
			tBoxRecalc(tbox,firstLine);
		if (tbox->ctlScrollV || tbox->ctlScrollH)
		{
			int		mxv, mxh;

			tBoxMaxScroll(tbox,&mxv,&mxh);
			if (tbox->ctlScrollV)
			{
				if (mxv!=tbox->ctlScrollV->max)
					upd = 1;
				tbox->ctlScrollV->max = mxv;
			}
			if (tbox->ctlScrollH)
			{
				if (mxh!=tbox->ctlScrollH->max)
					upd = 1;
				tbox->ctlScrollH->max = mxh;
			}
		}
	}
	if (tbox->flags&tBoxAutoScroll)
	{
		i = tBoxPinView(tbox);
		if (i&3 && i&12)
			upd = 0;
	}
	if (upd)
		tBoxScrollBars(tbox,1);
	i = tBoxDraw(tbox);
	update_screen_window_rect(tbox->wind,
		p->clip.left,p->clip.top,p->clip.right,p->clip.bottom);
	p->clip = oldclip;
	return i;
}

int tBoxDelete(textBoxObj *tbox,
	int32 start,
	int32 end,
	boolean update)
{
	rect16		b;
	const tLineRun	*lr;
	char		*d;
	int32		length, i;
	int		top;

	if (!tbox)// || !(tbox->flags&tBoxModifiable))
		return -1;
	if (end<start)
		i = end, end = start, start = i;
	if (start<0)
		start = 0;
	if (end>tbox->length)
		end = tbox->length;
	if (start>=end || (lr=tBoxCharLine(tbox,start,NULL,&top,NULL))==NULL)
		return 0;
	length = end-start;
	if (tbox->styles && tbox->styleCount>1)
	{ /* style run adjustment, deletion */
		const tStyleRun	*sr;
		tStyleRun	*styles = tbox->styles;
		int		rc;
		/* keep first style if deleting all text */
		if (length==tbox->length)
		{
			if ((styles=realloc(styles,sizeof(tStyleRun)))==NULL)
				return -2;
			tbox->styles = styles, tbox->styleCount = 1;
			goto dotxt;
		}
		sr = tBoxCharStyle(tbox,start,&rc);
		i = sr-tbox->styles+1;
		if (rc)
		{ /* adjust offsets */
			tStyleRun	*sprdel = NULL; /* post-run deletion */

			styles+= i;
			do
				if ((styles->offs-=length)<=start)
					sprdel = styles;
			while(++styles,--rc);
			if (sprdel)
			{
				int		delLast = end==tbox->length;

				if (sprdel->offs<start)
					sprdel->offs = start;
				if ((styles=tbox->styles+(i-=sr->offs==start))<sprdel || delLast)
				{
					memcpy(styles,sprdel,
						(int8*)(tbox->styleCount+tbox->styles)
						-(int8*)sprdel);
					tbox->styleCount-= delLast+(sprdel-styles);
					if ((styles=realloc(tbox->styles,tbox->styleCount
							*sizeof(tStyleRun)))!=NULL)
						tbox->styles = styles;
				}
			}
		}
	}
dotxt:	memmove(tbox->text+start,tbox->text+end,tbox->length-end);
	tbox->length-= length;
	if ((d=realloc(tbox->text,tbox->length))==NULL && tbox->length)
		return -2;
	tbox->text = d;
	//if (tbox->selStart!=tbox->selEnd)
	//	top = IMIN(top);
	tbox->selStart = tbox->selEnd = start;
	i = lr-tbox->lines;

	b.right = b.bottom = b.left = b.top = top;
	tBoxUpdate(tbox,update ? NULL:&b,
		start && tbox->length ? i-(start==lr->offs
		|| start==lr->offs+(lr->length&IMAX(int))):0);
	return 1;
}

int tBoxInsert(textBoxObj *tbox,
	int32 start,
	const char *s,
	int32 length,
	const tStyleRun *style,
	boolean update)
{
	rect16		b;
	const tLineRun	*lr;
	tStyleRun	*styles;
	char		*d;
	int32		total;
	int		top;

	if (!tbox)// || !(tbox->flags&tBoxModifiable))
		return -1;
	if (length<=0)
	{ /* set style on total blank */
		if (!tbox->length && !start && style && !tbox->styleCount)
		{
			if ((styles=realloc(tbox->styles,sizeof(tStyleRun)))==NULL)
				return -2;
			tbox->styles = styles, tbox->styleCount = 1;
			*styles = *style;
		}
		return 0;
	}
	if (start<0)
		start = 0;
	if (start>tbox->length)
		start = tbox->length;
	if ((total=tbox->length+length)<0)
	{
		total = 0x7FFFFFF0-length;
		if ((length=total-tbox->length)<=0)
			return 0;
	}
	if ((lr=tBoxCharLine(tbox,start,NULL,&top,NULL))==NULL)
		return 0;
	if (tbox->styles || style)
	{
		const tStyleRun	*sr;
		int		i, m, rc;

		if (!tbox->length || tbox->styleCount<=0)
		{ /* set if no styles or the old length is zero */
			if ((styles=realloc(tbox->styles,sizeof(tStyleRun)))==NULL)
				return -2;
			tbox->styles = styles, tbox->styleCount = 1;
			if (style)
				*styles = *style;
			goto dotxt;
		}
		sr = tBoxCharStyle(tbox,start-1,&rc); //lr->offs,&rc
		i = (styles=tbox->styles)!=NULL ? sr-styles:0;
		if (!style || !memcmp(&sr->font,&style->font,
				sizeof(tStyleRun)-OFFSET(tStyleRun,font)))
		{ /* same style; no addition */
			m = 0;
			goto ajs;
		}
		/* 1: new style coming exactly at run point; insert one */
		/* 2: inserting in the middle of a style; split */
		m = (sr->offs!=start)+1;
		if ((styles=realloc(tbox->styles,
				sizeof(tStyleRun)*(tbox->styleCount+m)))==NULL)
			return -2;
		/* ** this part UNTESTED ** */
		tbox->styles = styles;
		memmove(styles+i+m,styles+i,sizeof(tStyleRun)*(tbox->styleCount-i));
		styles[i+1] = *style;
		styles[i+1].offs = start;
		tbox->styleCount+= m;
		i+= m;
		if (m==2)
		{ /* split - must manually update offset */
			styles[i] = styles[i];
			styles[i].offs = start+length;
		}
	ajs:	if (rc) /* adjust offsets */
		{
			styles+= i+1;
			do
				styles->offs+= length;
			while(++styles,--rc);
		}
	}
dotxt:	if ((d=realloc(tbox->text,total))==NULL)
		return -2;
	tbox->text = d;
	tbox->length = total;
	memmove(d+start+length,d+start,total-start-length);
	memcpy(d+start,s,length);
	tbox->selEnd = tbox->selStart = start+length;
	b.right = b.bottom = b.left = b.top = top;
	tBoxUpdate(tbox,update ? NULL:&b,lr-tbox->lines);
	return 1;
}

SLRgn *tBoxSelRgn(textBoxObj *tbox)
{
	SLRgn		*rgn = NULL;
	pt2dq		pt[4];
	ifixed		ox;
	ufixed		x;
	int32		i, ss, se;
	int		r, top, bot;
	uint16		ils;

	if (!tbox)
		return NULL;
	ss = tbox->selStart, se = tbox->selEnd;
	if (ss>se)
		i = se, se = ss, ss = i;
	ils = !(tbox->flags&tBoxModifiable) && tbox->flags&tBoxListSelect;
	if ((ss==se && !ils) || !tBoxCharLine(tbox,ss,&x,&top,&bot))
		return NULL;
	ox = tbox->txbounds.left-tbox->scrollH;
	/* add sel start/end caret lines */
	x = ils ? tbox->txbounds.left:(x>>16)+ox; //ss ? (x>>16)+ox:0;
	FixPt(pt,x,top);
	FixPt(pt+1,x,bot);
	r = rgnXorLine(&rgn,tbox->txbounds.top,tbox->txbounds.bottom,*pt,pt[1]);
	assert(r>=0);
	x = tbox->maxWidth, top = bot = tbox->txbounds.bottom;
	tBoxCharLine(tbox,se,&x,&top,&bot);
	x = ils ? tbox->txbounds.right:(x>>16)+(x>>15&1)+ox;
	FixPt(pt+2,x,top);
	FixPt(pt+3,x,bot);
	r = rgnXorLine(&rgn,tbox->txbounds.top,tbox->txbounds.bottom,pt[2],pt[3]);
	assert(r>=0);
	if (pt[3].y!=pt[1].y)
	{ /* start/end are on different lines, add completion sides */
		pt[2].x = pt->x = tbox->txbounds.right<<16|0x8000;
		r = rgnXorLine(&rgn,tbox->txbounds.top,tbox->txbounds.bottom,*pt,pt[2]);
		assert(r>=0);
		pt[3].x = pt[1].x = tbox->txbounds.left<<16|0x8000;
		r = rgnXorLine(&rgn,tbox->txbounds.top,tbox->txbounds.bottom,pt[1],pt[3]);
		assert(r>=0);
	}
	return rgn;
}

static void tVisBBox(textBoxObj *tbox)
{
	grafPort	*p = &tbox->wind->port;
	rect2dq		frame;

	frame.top = MIN(tbox->txbounds.top,tbox->bounds.top)-1;
	frame.left = MIN(tbox->txbounds.left,tbox->bounds.left)-1;
	frame.bottom = MIN(tbox->txbounds.bottom,tbox->bounds.bottom);
	frame.right = MIN(tbox->txbounds.right,tbox->bounds.right);
	p->penw = p->penh = 0x10000;
	rRect(p,&frame);
	p->penw =-1;
}

int tBoxDraw(textBoxObj *tbox)
{
	penState	pen;
	rect16		oldcl;
	rgb48		oldfg;
	const tStyleRun	*sr;
	grafPort	*p;
	SLRgn		*hRgn;
	const bmfMemory	*oldfm;
	const tLineRun	*lr, *el;
	const char	*s;
	ufixed		oldch, oldsp, oldta, oldto;
	ifixed		qx;
	int32		z, i, lz, c, np;
	int		r = 0, sc = 0, y, v;
	uint8		oldtm, oldfc;

	if (!tbox || !tbox->wind || !tbox->wind->visible)
		return -1;
	z = tbox->length, p = &tbox->wind->port;
	/* get the line information; calculate if necessary */
	if (!tbox->lines && tBoxRecalc(tbox,0)<0)
		return -2;
	el = tbox->lines+tbox->lineCount-1;

	oldfg = p->fgColor;
	oldcl = p->clip, oldtm = p->txMode, oldfm = p->txFM, oldfc = p->txStyle;
	oldch = p->chExtra, oldsp = p->spExtra, oldta = p->tabAlign, oldto = p->tabOrigin;

	/* caret is erased */
	tbox->flags&=~tBoxCaretState, tbox->caretTime = 0;

	RGBFore(p,tbox->flags&tBoxUseCustBgColor ? &tbox->custBgColor:&p->bkColor);
	rPenNormal(p,&pen);
	p->penw =-1;
	if (rFillRect(p,tbox->bounds.left,tbox->bounds.top,tbox->bounds.right,tbox->bounds.bottom)<=0)
		goto end;
	/* set up clipping, and erase */
	if (!SectRect16(&p->clip,&tbox->txbounds))
		goto end;
	if (tbox->flags&tBoxFrame)
		tVisBBox(tbox);
	/* set up origin */
	if ((i=tbox->txbounds.left-tbox->scrollH)!=(int16)i)
		i = i>>31^IMAX(int16);
	v = tbox->scrollV;
	if ((lr=tVisLine(tbox->lines,tbox->lineCount,
			tbox->flags&tBoxFullUpdate ? v
			:v+MAX(p->clip.top-tbox->txbounds.top,0)))==NULL)
		goto end;
	p->tabAlign = tbox->tabAlign, p->tabOrigin = i<<16|0x8000;
	s = tbox->text, i = lr->offs, v-= tbox->txbounds.top;

	/* solid copy interferes with italics, kerning, etc. */
	p->txMode = mTransparent;

	/* set up hilite region (and render layer) */
	if ((hRgn=tBoxSelRgn(tbox))!=NULL)
	{
		p->pnMode = mSrcCopy;
		RGBFore(p,tbox->flags&tHiliteLayer ? &tbox->hiliteColor:&rgb_white);
		rRegion(p,hRgn);
		if (tbox->flags&tHiliteLayer)
			{free(hRgn); hRgn = NULL;}
	}
	r = 0, sr = tBoxCharStyle(tbox,i,&sc);
	goto start;
	do
	{
		if (sc>0 && i>=sr[1].offs)
		{
			--sc, ++sr;
			p->txFM = styleRunFont(sr), p->txStyle = sr->style;
			RGBFore(p,&sr->color);
		}
		c = lz;
		if (sc && (np=sr[1].offs-i)<c)
			c = np;
		if (c>0)
			qx = rDrawText(p,s+i,c,y,qx), i+= c, lz-= c;
		if (lz<=0)
		{
			++lr;
			if (lr>el)
				break;
			i = lr->offs;
			if (sc>0 && i>=sr[1].offs)
			{
				--sc, ++sr;
			start:	p->txFM = styleRunFont(sr), p->txStyle = sr->style;
				RGBFore(p,&sr->color);
				r = 1;
			}
			lz = lr->length&IMAX(int32);
			if ((y=lr->vpos-v)>p->clip.bottom)
				el = lr; /* height includes the line */
			qx = addu(p->tabOrigin,tLineMeasure(tbox,lr,sr->special,
				&p->chExtra,&p->spExtra));
		}
	} while(i<z);

	if (hRgn)
	{
		p->pnMode = mSrcXor;
		rRegion(p,hRgn);
		free(hRgn);
	}
end:	p->clip = oldcl, p->txMode = oldtm, p->txFM = oldfm, p->txStyle = oldfc;
	p->chExtra = oldch, p->spExtra = oldsp, p->tabAlign = oldta, p->tabOrigin = oldto;
	RGBFore(p,&oldfg);
	rRestorePen(p,&pen,1);
	return r;
}

static int tBoxDrawCaret(textBoxObj *tbox)
{
	rect16		oldcl;
	grafPort	*p;
	pixPat		*oldpp;
	ufixed		x;
	int		r, left, top, bot;

	if (!tbox->wind || tbox->flags&tBoxListSelect
			|| !tBoxCharLine(tbox,tbox->selEnd,&x,&top,&bot))
		return 0;
	p = &tbox->wind->port, tbox->flags^= tBoxCaretState;
	left = (x>>16)+tbox->txbounds.left-tbox->scrollH;
	oldcl = p->clip, oldpp = p->pPat, p->pPat = NULL;
	SectRect16(&p->clip,&tbox->bounds);
	if ((r=rInvertRect(p,left,top,left+1,bot))!=0)
		update_screen_window_rect(tbox->wind,left,top,left+1,bot);
	p->clip = oldcl, p->pPat = oldpp;
	return r+1;
}

int tBoxIdle(textBoxObj *tbox)
{
	int32		time;

	if (!tbox)
		return -1;
	if (~tbox->flags&(tBoxActive|tBoxModifiable) || tbox->selEnd!=tbox->selStart)
		return 0;
	time = TickCount();
	if ((time-tbox->caretTime)<(int32)GetCaretTime())
		return 0;
	tbox->caretTime = time;
	return tBoxDrawCaret(tbox);
}

int tBoxPinView(textBoxObj *tbox)
{
	ufixed		x;
	int		i, r, m, top, bot;

	if (!tbox)
		return -1;
	if (tbox->flags&tBoxDragState)
	{
		int32		time = TickCount();

		if ((int16)(time-tbox->dragTime)<3) /* ~1/20 sec */
			return 0;
		tbox->dragTime = time;
	}
	tBoxCharLine(tbox,MAX(tbox->selEnd,0),&x,&top,&bot);
	r = 0;
	if (tbox->length<=0 || (top<tbox->txbounds.top && !(tbox->flags&tBoxBDirection)))
	{
		if (tbox->length<=0 || (tbox->scrollV-=tbox->txbounds.top-top)<0)
			tbox->scrollV = 0;
		r = 1;
	} else if (bot>tbox->txbounds.bottom && !(tbox->flags&tBoxUDirection))
		tbox->scrollV-= tbox->txbounds.bottom-bot, r = 2;

	if ((i=tbox->tHeight-RECTANGLE_HEIGHT(&tbox->txbounds))<0)
		i = 0;
	if (tbox->scrollV>i+0u)
		tbox->scrollV =~tbox->scrollV>>31&i;

	m = (x>>16)+tbox->txbounds.left-tbox->scrollH;
	if (tbox->length<=0 || m<tbox->txbounds.left)
	{
		if (tbox->length<=0 || (tbox->scrollH-=tbox->txbounds.left-m)<0)
			tbox->scrollH = 0;
		r|= 4;
	} else if (m>tbox->txbounds.right)
		tbox->scrollH-= tbox->txbounds.right-m, r|= 8;
	if (r)
		tBoxScrollBars(tbox,0);
	return r;
}

int tBoxSelect(textBoxObj *tbox,
	int32 ss,
	int32 se)
{
	grafPort	*p;
	rect16		oldcl;
	int		oldss, oldse;
	int		top, bot;

	if (!tbox)
		return -1;
	if (ss>tbox->length+0u)
		ss =~ss>>31&tbox->length;
	if (se>tbox->length+0u)
		se =~se>>31&tbox->length;
	oldss = tbox->selStart, oldse = tbox->selEnd;
	if (tbox->flags&tBoxAutoScroll)
	{
		tbox->selEnd = se;
		if (tBoxPinView(tbox)>0)
			{tbox->flags|= tBoxFullUpdate; goto full;}
	}
	if (!((oldss^ss)|(oldse^se)))
		return 0;
full:	tbox->selStart = ss, tbox->selEnd = se;
	if (!tbox->wind)
		return 1;
	p = &tbox->wind->port, oldcl = p->clip;

	if (!(tbox->flags&tBoxFullUpdate))
	{
		int32		i;

		if (!SectRect16(&p->clip,&tbox->bounds))
			goto end;
		if ((i=MIN(ss,se))>oldss)
			i = oldss;
		tBoxCharLine(tbox,MIN(i,oldse),NULL,&top,NULL);
		if ((i=MAX(ss,se))<oldss)
			i = oldss;
		tBoxCharLine(tbox,MAX(i,oldse),NULL,NULL,&bot);
		if (top!=(int16)top)
			top = top>>31^IMAX(top);
		if (bot!=(int16)bot)
			bot = bot>>31^IMAX(bot);
		if (top>p->clip.top)
			p->clip.top = top;
		if (bot<p->clip.bottom)
			p->clip.bottom = bot;
	}
	tbox->flags&=~tBoxFullUpdate;
	tBoxDraw(tbox);
	update_screen_window_rect(tbox->wind,
		p->clip.left,p->clip.top,p->clip.right,p->clip.bottom);
end:	p->clip = oldcl;
	return 2;
}

int tBoxClick(textBoxObj *tbox,
	pt2d16 loc,
	uint16 modifiers,
	boolean dblClick)
{
	int32		i, sm, dblss, dblse;
	ufixed		x;
	int		y;

	if (!tbox || !tbox->wind)
		return -1;
	tbox->flags&=~(tBoxUDirection|tBoxBDirection);
	if (!(tbox->flags&tBoxActive) || !ptInRect16(&tbox->bounds,loc))
		return 0;
	if (!(modifiers&rightBtnState))
		return 1;
	if (tbox->flags&tBoxCaretState)
		tBoxDrawCaret(tbox);
	tbox->caretTime = 0;
	dblse = 0, dblss = IMAX(dblss);
	if ((x=loc.x-tbox->txbounds.left+tbox->scrollH)>0xFFFF)
		x =~x>>31;
	else
		x = (x<<16)-((x && dblClick)<<16);
	i = tBoxPtToChar(tbox,x,loc.y-tbox->txbounds.top+tbox->scrollV);
	sm = modifiers&shiftKey ? tbox->selStart:i;
	if (sm==i && tbox->flags&tBoxModifiable)
		if (dblClick)
		{
			dblss = tbox->wordBreak(tbox,i,1);
			dblse = tbox->wordBreak(tbox,i,0);
		} else
		{ /* QD TextEdit does a solid caret until up or the selection changes */
			tBoxSelect(tbox,i,i);
			tBoxDrawCaret(tbox);
		}
	tbox->flags|= tBoxDragState;
	tbox->dragTime = (int16)TickCount();
	goto start;
	do
	{
		int		ss, se;

		get_mouse_loc(&loc.x,&loc.y);
		x = loc.x-tbox->txbounds.left+tbox->scrollH;
		y = loc.y-tbox->txbounds.top+tbox->scrollV;
		if ((x=loc.x-tbox->txbounds.left+tbox->scrollH)>0xFFFF)
			x =~x>>31;
		else
			x = (x<<16)-((x && dblClick)<<16);
		i = tBoxPtToChar(tbox,x,y);
	start:	if (i<sm)
			ss = MAX(sm,dblse), se = MIN(i,dblss);
		else
			ss = MIN(sm,dblss), se = MAX(i,dblse);
		if (tbox->flags&tBoxInsertOnly)
			ss = se;
		else if (dblClick)
		{
			i = ss;
			ss = tbox->wordBreak(tbox,ss,ss<=se || ss==se+1);
			se = tbox->wordBreak(tbox,se,ss>se);
			if (ss==se)
				ss = i;
			else
				ss-= ss>se+1 && ss==i+1;
		}
		if (loc.y>=tbox->txbounds.bottom)
			tbox->flags|= tBoxBDirection;
		else
			tbox->flags|= tBoxUDirection;
		tBoxSelect(tbox,ss,se);
		tbox->flags&=~(tBoxUDirection|tBoxBDirection);
	} while(mouse_still_down());
	tbox->flags&=~tBoxDragState;
	tbox->selEndX = x;
	return 2;
}

int tBoxKey(textBoxObj *tbox,
	int key,
	uint16 modifiers)
{
	char		b[1];
	const tLineRun	*lr;
	int32		i;
	int		amtv, amth;

	if (!tbox)
		return -1;
	if (~tbox->flags&tBoxActive)
		return 0;
	amtv = amth = 0, i = tbox->selEnd;
	if (modifiers&cmdKey)
	{
		if (key=='a')
		{
			tBoxSelect(tbox,0,IMAX(int32));
			tbox->selEndX =-1u;
		} else if (key==kLeftArrowCharCode)
		{
			if ((lr=tBoxCharLine(tbox,i,NULL,NULL,NULL))==NULL)
				return 1;
			i = lr->offs;
			goto selo;
		} else if (key==kRightArrowCharCode)
		{
			if ((lr=tBoxCharLine(tbox,i,NULL,NULL,NULL))==NULL)
				return 1;
			i = lr->offs+(lr->length&IMAX(int32))-(lr->length>1);
			goto selo;
		} else if (key==kUpArrowCharCode)
			goto selu;
		else if (key==kDownArrowCharCode)
			goto seld;
		return 1;
	}
	switch(key)
	{
	case kPageUpCharCode:
		if (modifiers&controlKey)
			{amth =-16; goto scr;}
		amtv =-RECTANGLE_HEIGHT(&tbox->txbounds)+16;
		goto scr;
	case kPageDownCharCode:
		if (modifiers&controlKey)
			{amth = 16; goto scr;}
		amtv = RECTANGLE_HEIGHT(&tbox->txbounds)-16;
		goto scr;
	case kHomeCharCode:
		if (modifiers&controlKey && tbox->flags&tBoxModifiable)
		{
		selu:	i = 0;
			goto selo;
		}
		amtv =-0x10000000;
		goto scr;
	case kEndCharCode:
		if (modifiers&controlKey && tbox->flags&tBoxModifiable)
		{
		seld:	i = IMAX(int32);
		selo:	tBoxSelect(tbox,modifiers&shiftKey ? tbox->selEnd:i,i);
			tbox->selEndX =-1u;
			break;
		}
		amtv = 0x10000000;
		goto scr;
	case kLeftArrowCharCode:
		if (!(tbox->flags&tBoxModifiable))
			break;
		amtv = i, --i;
		/* prefer controlKey - HP keyboard won't do apps/shift/left */
		if (modifiers&(optionKey|controlKey))
			if ((i=tbox->wordBreak(tbox,i,1))==amtv)
				i = tbox->wordBreak(tbox,i-2,1);
		tbox->selEndX =-1u;
		goto sel;
	case kRightArrowCharCode:
		if (!(tbox->flags&tBoxModifiable))
			break;
		amtv = i, ++i;
		if (modifiers&(optionKey|controlKey))
			i = tbox->wordBreak(tbox,i,0);
		tbox->selEndX =-1u;
		goto sel;
	case kUpArrowCharCode:
		if (modifiers&controlKey || !(tbox->flags&(tBoxModifiable|tBoxListSelect)))
			{amtv =-16; goto scr;}
		amth =-1;
		goto arv;
	case kDownArrowCharCode:
		if (modifiers&controlKey || !(tbox->flags&(tBoxModifiable|tBoxListSelect)))
		{
			amtv = 16;
		scr:	tBoxScroll(tbox,amtv,amth);
			break;
		}
		amth = 1;
	arv:	tBoxCharLine(tbox,i,tbox->selEndX==-1u ? &tbox->selEndX:NULL,
			amth<=0 ? &amtv:NULL,amth>0 ? &amtv:NULL);
		i = tBoxPtToChar(tbox,tbox->selEndX,amtv+amth-tbox->txbounds.top+tbox->scrollV);
	sel:	tBoxSelect(tbox,modifiers&shiftKey ? tbox->selStart:i,i);
		if (tbox->selStart==tbox->selEnd && !(tbox->flags&tBoxCaretState))
			tBoxDrawCaret(tbox);
		tbox->caretTime = (int16)TickCount();
		break;
	case kBackspaceCharCode:
		if (!(tbox->flags&tBoxModifiable))
			return 0;
		tBoxDelete(tbox,tbox->selStart==tbox->selEnd ? i-1:tbox->selStart,i,1);
		break;

	default:
		if (!(tbox->flags&tBoxModifiable))
			return 0;
		tbox->selEndX =-1u;
		if (key==kDeleteCharCode)
		{
			tBoxDelete(tbox,i,tbox->selStart==tbox->selEnd ? i+1:i,1);
			break;
		}
		if (tbox->selStart!=tbox->selEnd)
			tBoxDelete(tbox,tbox->selStart,tbox->selEnd,0);
		*b = (char)key;
		tBoxInsert(tbox,tbox->selEnd,b,1,NULL,1);
		return 2;
	}
	return 1;
}

/* scroll bars, creation/disposal */

DEF_TRK_RTN(textTrackScroll)
{
	textBoxObj	*tbox = (textBoxObj*)ctl->refCon;
	int32		time;
	boolean		horiz = ctl->title[1]=='H';

	if (part==kControlLabelPart || !part)
	{
		if (horiz)
			tBoxScroll(tbox,0,value-tbox->scrollH);
		else
			tBoxScroll(tbox,value-tbox->scrollV,0);
		return;
	}
	time = TickCount();
	if ((int16)(time-tbox->dragTime)<3) /* ~1/20 sec */
		return;
	tbox->dragTime = time;
	if (part==kControlUpButtonPart)
		tBoxKey(tbox,horiz ? kPageUpCharCode:kUpArrowCharCode,controlKey);
	else if (part==kControlDownButtonPart)
		tBoxKey(tbox,horiz ? kPageDownCharCode:kDownArrowCharCode,controlKey);
	else if (part==kPageUpButtonPart)
		if (horiz)
			tBoxScroll(tbox,0,-RECTANGLE_WIDTH(&tbox->txbounds)+16);
		else
			tBoxKey(tbox,kPageUpCharCode,0);
	else if (part==kPageDownButtonPart)
		if (horiz)
			tBoxScroll(tbox,0,RECTANGLE_WIDTH(&tbox->txbounds)-16);
		else
			tBoxKey(tbox,kPageDownCharCode,0);
}

textBoxObj *tBoxNew(windowRef wind,
	const rect16 *bounds,
	const char *s,
	int32 z,
	uint16 flags)
{
	rect16		mbounds;
	textBoxObj	*tbox;

	if (!bounds || (tbox=calloc(sizeof(textBoxObj),1))==NULL)
		return NULL;
	tbox->bounds = *bounds;
	if (!s)
		z = 0;
	if (z>0)
	{
		if ((tbox->text=malloc(z))==NULL)
			goto fail;
		memcpy(tbox->text,s,z);
	}
	tbox->wind = wind;
	tbox->length = z;
	mbounds.left = mbounds.top = mbounds.right = mbounds.bottom = 2;
	tBoxSetMargins(tbox,&mbounds);
	tbox->wordBreak = defWordBreak;
	tbox->flags = flags;
	if (tBoxRecalc(tbox,0)<0)
		{fail: free(tbox->lines); free(tbox); return NULL;}

	if (flags&(tBoxAutoVScroll|tBoxAutoHScroll))
	{
		int		mxv, mxh;

		tBoxMaxScroll(tbox,&mxv,&mxh);
		if (flags&tBoxAutoVScroll)
		{
			mbounds = *bounds;
			mbounds.left = bounds->right;
			mbounds.right = bounds->right+16;
			tbox->ctlScrollV = wNewControl(wind,&mbounds,(uint8*)"\1V",1,0,0,mxv,
				ctlTypeScrollBar,(long int)tbox);
			if (tbox->ctlScrollV)
				tbox->ctlScrollV->trk = textTrackScroll;
		}
		if (flags&tBoxAutoHScroll)
		{
			mbounds = *bounds;
			mbounds.top = bounds->bottom;
			mbounds.bottom = bounds->bottom+16;
			tbox->ctlScrollH = wNewControl(wind,&mbounds,(uint8*)"\1H",1,0,0,mxh,
				ctlTypeScrollBar,(long int)tbox);
			if (tbox->ctlScrollH)
				tbox->ctlScrollH->trk = textTrackScroll;
		}
	}
	return tbox;
}

int tBoxDispose(textBoxObj *tbox)
{
	int		r;

	if (!tbox)
		return -1;
	r = 0;
	if (tbox->text)
		{free(tbox->text); tbox->text = NULL, ++r;}
	if (tbox->lines)
		{free(tbox->lines); tbox->lines = NULL, ++r;}
	if (tbox->styles)
		{free(tbox->styles); tbox->styles = NULL, ++r;}
	if (tbox->ctlScrollV)
		wDisposeControl(tbox->ctlScrollV);
	if (tbox->ctlScrollH)
		wDisposeControl(tbox->ctlScrollH);
	free(tbox);
	return r;
}
