/*
 mwind.c

  Created: 2014.08.17
  Revised: 2014.11.26

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
*/

#define __MWIND_PRIVATE
#include "mevent.h"
#include "mwind.h"

#include <string.h>
#if DEBUG
 #include <stdio.h>
#endif

/* ---- globals */

const rgb48	_defMWinColors[3] =
{
	{0,0,0}, /* frame color */
	{0,0,0}, /* title color */
	{61166,61166,61166} /* content (was 86.667%) */
};

windowRecord	screen_desktop_window;
/* the order here is deliberate for a wPaintBehind hack */
#define frontWindow	screen_desktop_window.below
#define bottomWindow	screen_desktop_window.above

GrafPtr		_curPort;

#if DEBUG
static void checkWindowChain(void)
{
	windowRef	win;
	int		wincount = 1;
#if DUMPCHAIN
	printf("# window chain (bottom up):\n");
#endif
	if ((win=bottomWindow)!=NULL)
	{
		do
		{
		#if DUMPCHAIN
			printf("# %* %08p - %.*s\n",
				wincount,win,*win->title,win->title+1);
		#endif
		} while(++wincount<32 && (win=win->above)!=NULL);
		assert(wincount<32);
	}
#if DUMPCHAIN
	printf("# window chain (top down):\n");
#endif
	if ((win=frontWindow)!=NULL)
	{
		do
		{
		#if DUMPCHAIN
			printf("# %* %08p - %.*s\n",
				wincount,win,*win->title,win->title+1);
		#endif
		} while(--wincount>=-1 && (win=win->below)!=NULL);
		if (wincount<0)
			printf("# unmatching window chain below list");
	}
		
}
#endif

/* ---- code */

void _defMWinModalDraw(const rect16 *box,
	const rgb48 *color,
	boolean fill)
{
	rect2dq		b;
	pt2dq		pt[3];
	cSpec		cstor[17];
	cTab64		*ct = (cTab64*)&cstor;
	CGrafPtr	p = &screen_desktop_window.port;

	if (!color)
		color = _defMWinColors;
	_imgContColors(ct,color+wCContent,NULL,0);
	cstor->rgb = p->fgColor;
	rPenNormal(p,NULL);
	RGBFore(p,color+wCFrame);
	FixRect(&b,box->left-3,box->top-3,box->right+3,box->bottom+3);
	rRect(p,&b);
	/* -1:bl-tl-tr */
	RGBFore(p,&ct->colors[2].rgb);
	FixPt(pt,box->left-1,box->bottom);
	FixPt(pt+1,box->left-1,box->top-1);
	FixPt(pt+2,box->right,box->top-1);
	rLine(p,*pt,pt[1]);
	rLine(p,pt[1],pt[2]);
	/* -1:tr-br-bl */
	RGBFore(p,&ct->colors[6].rgb);
	FixPt(pt,box->right,box->bottom);
	FixPt(pt+1,box->left-1,box->bottom);
	rLine(p,pt[2],*pt);
	rLine(p,*pt,pt[1]);
	/* -2:bl-tl-tr */
	RGBFore(p,&ct->colors[1].rgb);
	FixPt(pt,box->left-2,box->bottom+1);
	FixPt(pt+1,box->left-2,box->top-2);
	FixPt(pt+2,box->right+1,box->top-2);
	rLine(p,*pt,pt[1]);
	rLine(p,pt[1],pt[2]);
	/* -2:tr-br-bl */
	RGBFore(p,&ct->colors[9].rgb);
	FixPt(pt,box->right+1,box->top-2);
	FixPt(pt+1,box->right+1,box->bottom+1);
	FixPt(pt+2,box->left-2,box->bottom+1);
	rLine(p,*pt,pt[1]);
	rLine(p,pt[1],pt[2]);
	if (fill)
	{
		RGBBack(p,color+wCContent);
		rEraseRect(p,box->left,box->top,box->right,box->bottom);
	}
	RGBFore(p,&cstor->rgb);
	update_screen_rect(box->left-3,box->top-3,box->right+3,box->bottom+3);
}

DEF_WIN_RTN(modalProc)
{
	if (msg==wcMsgDraw)
	{
		if (!win->visible)
			return 0;
		_defMWinModalDraw(&win->bounds,win->data,0);
		return 1;
	} else if (msg==wcMsgCalc)
	{
		win->frame = win->bounds;
		OutsetRect16(&win->frame,3,3);
	}
	return -8;
}

winRoutine	f_winr[] =
{
	NULL,		/* doc */
	modalProc	/* modal */
			/* single-pixel frame */
};

windowRef wFrontWindow(void)
	{return frontWindow;}

/* this only draws the frame; wBeginUpdate erases */
int wPaintOne(windowRef win,
	const rect16 *inval)
{
	rect16		b;

	if (!win)
		return -1;
	b = screen_desktop_window.bounds;
	if (!win->visible || !SectRect16(&b,inval))
		return 0;
	b.top-= win->bounds.top;
	b.left-= win->bounds.left;
	b.bottom-= win->bounds.top;
	b.right-= win->bounds.left;
	wInvalRect(win,&b);
	return 1;
}

int wPaintBehind(windowRef behind,
	const rect16 *inval)
{
	rect16		clobbered;
	windowRef	win;
#if 0
	windowRef	*updl = NULL, *p;
#endif
	int		c;

	win = &screen_desktop_window;
	if (!win->above)
		return 0;
	clobbered = *inval;
	SectRect16(&clobbered,&screen_desktop_window.bounds);
#if 0
	for(c=0; win && win!=behind; win=win->next)
	{
		rect16		b;

		b = clobbered;
		if (win->visible && SectRect16(&b,&win->frame))
		{
			if ((p=realloc(updl,++c*sizeof(windowRef)))==NULL)
			{
				free(updl);
				return -2;
			}
			updl = p;
			updl[c-1] = win;
		}
	}
	if (!c)
		return 0;
	do
		wPaintOne(updl[c-1],&clobbered);
	while(--c);
#endif
	for(c=0; win && win!=behind; win=win->above)
		wPaintOne(win,&clobbered);
	update_screen_rect(clobbered.left,clobbered.top,
		clobbered.right,clobbered.bottom);
	return c;
}

static int updateWindow(windowRef win)
{
	rect16		b;
	pixMap		*m, *sm;

	if (!win->visible)
		return 0;
	b = screen_desktop_window.bounds;
	SectRect16(&b,&win->bounds);
	if ((m=win->port.map)==NULL)
		return -1;
	m->w = RECTANGLE_WIDTH(&b);
	m->h = RECTANGLE_HEIGHT(&b);
	m->pOrigin.x =-b.left;
	m->pOrigin.y =-b.top;
	sm = screen_desktop_window.port.map;
	if (memcmp(&m->pitch,&sm->pitch,OFFSET(pixMap,pHand)-OFFSET(pixMap,pitch))
			|| (sm->bpp<9 && sm->pTable
			&& (!m->pTable || m->pTable->seed!=sm->pTable->seed)))
	{
		if (m->iTable)
			{free(m->iTable); m->iTable = NULL;}
		memcpy(&m->pitch,&sm->pitch,OFFSET(pixMap,pHand)-OFFSET(pixMap,pitch));
		copyCTable(&m->pTable,sm->bpp<9 ? sm->pTable:NULL);
	}
	win->frame = win->bounds;
	if (win->rtn)
		win->rtn(win,wcMsgCalc,NULL);
	return wPaintOne(win,&win->frame);
}

void wSendBehind(windowRef sending,
	windowRef behind)
{
	if (!sending || sending==&screen_desktop_window || sending==behind)
		return; /* idiot proofing */
	if (!frontWindow) /* uninitialized FrontWindow - assume empty */
		{frontWindow = bottomWindow = sending; return;}
	if (behind && behind!=(windowRef)-1l && behind->below==sending)
		return; /* already behind this window */
	/* remove, first (will be re-inserted into the list) */
	if (sending->above)
		sending->above->below = sending->below;
	if (sending->below)
		sending->below->above = sending->above;
	sending->below = sending->above = NULL;
	if (behind==(windowRef)-1l)
	{ /* bring to front */
		if (sending==frontWindow)// || !frontWindow)
			return; /* already front window */
		assert(!frontWindow->above);
		frontWindow->above = sending;
		sending->below = frontWindow;
		frontWindow = sending;
		updateWindow(sending); /* only the new front window to be updated */
		return;
	}
	if (!behind) /* to bottom */
		if ((behind=bottomWindow)==NULL)
			return; /* !!! */
	sending->above = behind;
	if ((sending->below=behind->below)!=NULL)
		behind->below->above = sending;
	behind->below = sending;
	wPaintBehind(sending,&sending->frame);
}

void wOpenWindow(windowRef newWin,
	windowRef behind,
	int16 procID)
{
	if (!newWin)
		return;
	if (procID>=0)
		newWin->rtn = (procID&=15)<GET_ARRAY_COUNT(f_winr) ? f_winr[procID]:NULL;
	openGrafPort(&newWin->port,screen_desktop_window.port.map->bpp,
		RECTANGLE_WIDTH(&newWin->bounds),RECTANGLE_HEIGHT(&newWin->bounds),0);

	newWin->port.txFont = screen_desktop_window.port.txFont;
	newWin->port.txSize = screen_desktop_window.port.txSize;
	newWin->port.txStyle = screen_desktop_window.port.txStyle;
	newWin->port.fgColor = screen_desktop_window.port.fgColor;

	wSendBehind(newWin,behind);
	updateWindow(newWin);
	RGBBack(&newWin->port,(newWin->data ? (const rgb48*)newWin->data:_defMWinColors)+2);
	wPaintOne(newWin,&screen_desktop_window.bounds);
#if DEBUG
	checkWindowChain();
#endif
}

void wCloseWindow(windowRef win)
{
	if (!win || win==&screen_desktop_window)
		return; /* idiot proofing */
	win->visible = 0;
	wKillControls(win);
	wPaintBehind(win,&win->frame);

	/* remove from list */
	if (win->above)
		win->above->below = win->below;
	if (win->below)
		win->below->above = win->above;
	closeGrafPort(&win->port);
	/* recalc bottom */
	if (!win->below)
	{ /* bottom's below ==NULL; the desktop's above should be the bottom window */
		assert(bottomWindow==win);
		bottomWindow = win->above;
	}
	/* recalc FrontWindow */
	win->below = win->above = NULL;
	if ((win=bottomWindow)!=NULL)
		while(win->above)
			win = win->above;
	/* avoid recursion (shouldn't really need this line) */
	frontWindow = win==&screen_desktop_window ? NULL:win;
#if DEBUG
	checkWindowChain();
#endif
}

void wDisposeWindow(windowRef win)
{
	wCloseWindow(win);
	free(win);
}

int wShowWindow(windowRef win,
	boolean show)
{
	if (!win)
		return -1;
	if (show)
	{
		if (!win->visible)
			win->visible = 1;
		return updateWindow(win);
	}
	if (win->visible)
	{
		win->visible = 0;
		wPaintBehind(win,&win->frame);
		return 1;
	}
	return 0;
}

void wBeginUpdate(windowRef win)
{
	rgb48		bkc;
	CGrafPtr	p = &screen_desktop_window.port;

	if (win->updateRect.left>=win->updateRect.right
			|| win->updateRect.top>=win->updateRect.bottom)
		return;
	if (win->rtn)
		win->rtn(win,wcMsgDraw,NULL);
	bkc = p->bkColor;
	RGBBack(p,(win->data ? (const rgb48*)win->data:_defMWinColors)+wCContent);
	rEraseRect(p,MAX(win->updateRect.left,0)+win->bounds.left,
		MAX(win->updateRect.top,0)+win->bounds.top,
		MIN(win->updateRect.right,win->port.map->w)+win->bounds.left,
		MIN(win->updateRect.bottom,win->port.map->h)+win->bounds.top);
	win->clipSave = win->port.clip;
	SectRect16(&win->port.clip,&win->updateRect);
	p->bkColor = bkc;
}

void wEndUpdate(windowRef win)
{
	if (!win)
		return;
	if (win->visible==2)
		win->visible = 1;
	if (win->updateRect.left>=win->updateRect.right
			|| win->updateRect.top>=win->updateRect.bottom
			|| !win->visible)
		return;
	update_screen_rect(win->updateRect.left+win->bounds.left,
		win->updateRect.top+win->bounds.top,
		win->updateRect.right+win->bounds.left,
		win->updateRect.bottom+win->bounds.top);
	win->updateRect.right = win->updateRect.bottom = IMIN(int16);
	win->port.clip = win->clipSave;
}

void wInvalRect(windowRef win,
	const rect16 *dirtyRect)
{
	eventRecord	event;

	if (!win || !win->visible || !dirtyRect)
		return;
	UnionRect16(&win->updateRect,dirtyRect);
	if (win->visible!=2)
	{
		win->visible = 2;
		event = mEventState;
		event.what = updateEvt;
		event.message = (long int)win;
		PostMEvent(&event);
	}
}

void update_screen_window_rect(windowRef win,
	int left,
	int top,
	int right,
	int bottom)
{
	int		ox, oy;

	if (!win || win->port.map->baseAddr!=screen_desktop_window.port.map->baseAddr)
		return;
	if (win)
		ox =-win->bounds.left, oy =-win->bounds.top;
	else
		ox = oy = 0;
	if ((ox+=win->port.origin.x)>0)
		left = subu(left,ox), right = subu(right,ox);
	else
		left = addu(left,-ox), right = addu(right,-ox);
	if ((oy+=win->port.origin.y)>0)
		top = subu(top,oy), bottom = subu(bottom,oy);
	else
		top = addu(top,-oy), bottom = addu(bottom,-oy);
	update_screen_rect(left,top,right-left,bottom-top);
}

void MacSetPort(GrafPtr port)
{
	_curPort = port;
}

void GetPort(GrafPtr *port)
{
	*port = _curPort;
}
