/*
 spanner.c
  portable graphics and input demo
  see README.txt for input features

  Created: 2014.02.15 20:58 EST
  Revised: 2014.11.20
   Author: Adam J. P.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#include <enviro64.h>
#include <stdio.h>

#include <windows.h>

#include "my_cseries.h"
#include "mwind.h"
#include "mevent.h"
#include "mtext.h"
#include "interpreter.h"

#define devWindowName		"spanner"
#define DEFAULT_BPP		24
#define DISPLAY_STATS		1

#define CW_TOPLEFT		0

/* ---- required suite */

int8		temporary[512];

void global_idle_proc(void)
{
}

void stop_interface_fade(void)
{
}

uint8 *getpstr(void *d,
	short ID,
	int x)
{
	if (!d || ID<0)
		return NULL;
	*(int8*)d = 0;
	return d;
}

static uint8	_dbgTemp[1024];

#define ATTASKLEVEL0()		1

int dprintf(const char *form,
	...)
{
	va_list		args;
	char		*s;
	int		len;

	va_start(args,form);

	if ((len=vsprintf(1+(char*)_dbgTemp,form,args))>255)
		len = 255;

	va_end(args);

	s = memchr(1+_dbgTemp,';',*_dbgTemp=len);
	//_dbgTemp[len+1] = 0; -- already set by vsprintf

	/* if first ';' follows "g;" */
	if (!ATTASKLEVEL0() || !s || s[1]!='g') // || s[2]!=';')
	{
#if TARGET_OS_MAC
		DebugStr(_dbgTemp);
#else
		__debugbreak();
#endif
	}
	if (s)
		*s = 0;
	if (ATTASKLEVEL0())
		printf("## %s\n",
			1+_dbgTemp);
	return len;
}

void _debugMsg(const char *msg,
	const char *where,
	int n)
{
	static char	*type[] = {"warning","assertion failure","pause","halt"};
	int		z, i = n&0x3FFFFFFF;

	n = (uint32)n>>30;
	z = sprintf((char*)_dbgTemp+1,msg ? "%s in '%s', line %d: %s\n":"%s in '%s', line %d\n",
		type[n],where,i,msg);
	*_dbgTemp = MIN(z,255);
	n&= 1;
	if (n)
		freopen("stderr.txt","w",stderr);
	fprintf(n ? stderr:stdout,n ? "### %s\n":"## %s\n",
		_dbgTemp+1);
	if (n)
		fflush(stderr);
	__debugbreak();
	if (n)
		exit(-1);
}

/* ---- code */

static const char	*base36 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";

static char *hexToStr(char *d,
	const uint8 *s,
	int z)
{
	if (!d || z<=0)
		return d;
	do
		*d++ = base36[*s>>4], *d++ = base36[*s++&15];
	while(--z);
	return d;
}

/* ---- active window */

#define MAXTIMERS		32

static timerThread	*oldtimers;

static uint64	timer_start;

#define SCRW			640
#define SCRH			480

const rgb48	rgb_orange = {65535,32767,0};
const rgb48	rgb_tan = {37008,29041,14906};
const rgb48	rgb_ltyellow = {65535,65535,.65*65535};

static int boxer(const char *path,
	const rect16 *bounds)
{
	rect2dq		frame;
	rect16		clip;
	rgb48		oldfore, oldback;
	FILE		*f;
	textBoxObj	*tbox = NULL;
	ctlObj		*tbctl = NULL;
	char		*b;
	long int	z;
	int		r = 0, oldpenw, oldpenh;

	if (!path)
		return -1;
	if ((f=fopen(path,"r"))==0)
		return -43;
	if (fseek(f,0,SEEK_END))
		{r =-3; goto end;}
	if ((z=ftell(f))>(8<<20))
		{r =-1309; goto end;}
	fseek(f,0,SEEK_SET);
	if ((b=malloc(z))==NULL)
		{r =-2; goto end;}
	z = fread(b,1,z,f);
	fclose(f); f = NULL;
	r = 0;
	clip = screen_desktop_window.port.clip;
	memset(&screen_desktop_window.port.clip,0,sizeof(screen_desktop_window.port.clip));
	//if ((tbox=tBoxNew(&screen_desktop_window,bounds,0,0,tBoxAutoScroll|tBoxActive|tBoxModifiable
	//		|tBoxAutoVScroll))==NULL) //|tBoxAutoHScroll
	if ((tbctl=wNewControl(&screen_desktop_window,bounds,0,1,tBoxAutoVScroll<<16,0,0,
			ctlTypeStaticText,0))==NULL || (tbox=tbctl->data)==NULL)
	{
		free(b);
		screen_desktop_window.port.clip = clip;
		goto end;
	}
	//tbox->maxWidth = 640<<16;
	if ((tbox->styles=calloc(sizeof(tStyleRun),3))!=NULL)
	{
		tbox->styleCount = 3;
		tbox->styles[1].font = 2, tbox->styles[1].size = 14;
		tbox->styles->font = tbox->styles[2].font = 21;
		tbox->styles[2].style = tbox->styles[1].style = _style_bold;
		tbox->styles->special = _special_align_center;
		tbox->styles[2].special = tbox->styles[1].special = _special_align_justify;
		tbox->styles[1].offs = 13;
		tbox->styles[2].offs = 22;
	}
	/* add a softened frame (uses blend transfer mode) */
	screen_desktop_window.port.clip = clip;
	oldfore = screen_desktop_window.port.fgColor;
	oldback = screen_desktop_window.port.bkColor;
	oldpenw = screen_desktop_window.port.penw;
	oldpenh = screen_desktop_window.port.penh;
	FixRect(&frame,bounds->left,bounds->top,bounds->right,bounds->bottom);
	RGBFore(&screen_desktop_window.port,&rgb_ltyellow);
	screen_desktop_window.port.pnMode = mBlend;
	screen_desktop_window.port.opColor = system_colors[_system_gray67];
	screen_desktop_window.port.penw = screen_desktop_window.port.penh = 2<<16;
	rRect(&screen_desktop_window.port,&frame);
	screen_desktop_window.port.penw = screen_desktop_window.port.penh = 4<<16;
	rRect(&screen_desktop_window.port,&frame);
	screen_desktop_window.port.penw = screen_desktop_window.port.penh = 6<<16;
	rRect(&screen_desktop_window.port,&frame);
	RGBBack(&screen_desktop_window.port,&rgb_ltyellow);
	/* have the text box own the buffer (tBoxUpdate is necessary) */
	tbox->text = b, tbox->length = z;
	tBoxUpdate(tbox,NULL,0);
	screen_desktop_window.ctlFocus = tbctl;
	screen_desktop_window.ctlCancel = screen_desktop_window.ctlDefault = screen_desktop_window.ctlList;
	for(r=0;;)
	{
		eventRecord	event;

		if (GetNextEvent(everyEvent,&event))
			if ((event.what==keyDown || event.what==autoKey)
					&& ((uint8)event.message=='q'
					|| (uint8)event.message=='w')
					&& event.modifiers&cmdKey)
				break;
		if (wControlEvent(&screen_desktop_window,&event)==screen_desktop_window.ctlList)
			break;
		if (event.what==closeEvent)
			{r = 1; break;}
	}
	screen_desktop_window.port.penw = oldpenw, screen_desktop_window.port.penh = oldpenh;
	RGBFore(&screen_desktop_window.port,&oldfore);
	RGBBack(&screen_desktop_window.port,&oldback);
	rFillRect(&screen_desktop_window.port,tbox->bounds.left-5,tbox->bounds.top-5,
		tbox->bounds.right+5+17,tbox->bounds.bottom+5+17);
	update_screen_rect(0,0,screen_desktop_window.port.map->w,
		screen_desktop_window.port.map->h);
	set_mouse_cursor(0,0);
end:	if (tbctl)
		wDisposeControl(tbctl);
	//if (tbox)
	//	tBoxDispose(tbox);
	if (f)
		fclose(f);
	return r;
}

#if DISPLAY_STATS
static void displayStats(ffMemory *ff,
	int x,
	int y)
{
	static char	*modenames[] =
	{
		"cpy ","or   ","xor  ","bic  ",
		"ncpy","nor  ","nxor","nbic",
		"tran","blnd","admx","admn",
		"adpn","adov","sbpn","sbov"
	};
	char		b[sizeof(mKeyMap)*2+80];
	rgb48		color;
	pt2d16		where;
	uint64		mcs;
	int32		z, vblc = TickCount();

	get_mouse_loc(&where.x,&where.y);
	getPixelRGB(screen_desktop_window.port.map,
		where.x-screen_desktop_window.port.origin.x,where.y-screen_desktop_window.port.origin.y,&color);
	hexToStr(b,mKeyMap,sizeof(mKeyMap));
	if (oldtimers)
	{
		Microseconds(&mcs);
		z = sprintf(b+sizeof(mKeyMap)*2," %.2fs - %.3f ms; %s; %d,%d=%03X,%03X,%03X",
			vblc/60.,(mcs-timer_start)/(double)vblc/1000.,
			modenames[screen_desktop_window.port.pnMode&15],
			where.x,where.y,color.red>>4,color.green>>4,color.blue>>4);
	} else
		z = sprintf(b+sizeof(mKeyMap)*2," %s; %d,%d=%03X,%03X,%03X",
			modenames[screen_desktop_window.port.pnMode&15],
			where.x,where.y,color.red>>4,color.green>>4,color.blue>>4);
	screen_desktop_window.port.txMode = mSrcCopy;
	screen_desktop_window.port.txSize = 0;
	rTextFont(&screen_desktop_window.port,0);
	screen_desktop_window.port.txStyle = _style_plain;
	x = rDrawText(&screen_desktop_window.port,b,z+sizeof(mKeyMap)*2,y,x<<16|0x8000);
	screen_desktop_window.port.chExtra = screen_desktop_window.port.map->w<<16;
	rDrawText(&screen_desktop_window.port," ",1,y,x);
	screen_desktop_window.port.chExtra = 0;
}
#endif

static void testTime(void)
{
	DWORD		prev = timeGetTime(), next, sum = 0, dif, mint = IMIN(DWORD);
	int		c = 20;

	while((next=timeGetTime())==prev)
		;
	do
	{
		for(prev=next; (next=timeGetTime())==prev;)
			;
		if ((dif=next-prev)<mint)
			mint = dif;
	} while(sum+=dif,--c);
	printf("GetTickCount resolution: %gms (%d lowest)\n",
		((sum+1)>>1)/10.,mint);
}

int WINAPI WinMain(HINSTANCE hInstance,
	HINSTANCE hPrevInstance,
	LPSTR lpCmdLine,
	int iCmdShow)
{
	static int8	txsizes[5] = {9, 12, 14, 24, 36};
	static uint8	weavePat[8] = {0xF8,0x74,0x22,0x47,0x8F,0x17,0x22,0x71};
	static char	txs[] = "Run, rabbit, run! \xD9";
#ifdef TEST_FILTER
	cTab64		*filterTable;
#endif
	rect2dq		ov, rr;
	rect16		ctlb;
	uint8		grayPat[8];
	pt2dq		tr[4], ln[2];
	pt2d16		tx, px, pxz, pxd;
	rgb48		shapeColor;
	const rgb48	*lineColor;
	const bmfMemory	*fm;
	ctlObj		*cctl;
	pixMap		*pic;
	const char	*picName = NULL;
	long int	depth = 0;
	int		r, txhw;
	uint32		ltick = 0;
	boolean		wpat = 0, wpatt = 0;
	boolean		pmode = 0, pmodet = 0;
	boolean		animate = 1, animatet = 0;
	boolean		incrt = 0, decrt = 0;
	boolean		hasDrawn = 0;
	int8		txstyl = 0, txsize = 1, txfont = 3, txgray = 0, txspex = 0;
(void)hPrevInstance; (void)iCmdShow;
	wRegisterDev(hInstance,
		LoadIcon(NULL,IDI_APPLICATION),
		NULL,
		(HBRUSH)GetStockObject(BLACK_BRUSH),
		devWindowName);
	set_mouse_cursor(4,0);

	//srand(GetTickCount());
	remove("stderr.txt");
	freopen("stdout.txt","w",stdout);

	if (lpCmdLine)
		if ((picName=readanyinteger(lpCmdLine,strlen(lpCmdLine),&depth))!=NULL)
			picName = parse_skipwsp(picName,picName+strlen(picName));

	if ((r=wSetVideoMode(depth ? depth:DEFAULT_BPP,SCRW,SCRH,CW_TOPLEFT,0))<=0)
	{
		if (r)
			dprintf("# error setting %dx%d %dbpp mode;g",
				SCRW,SCRH,depth);
		else
			dprintf("# could not create video window;g");
		goto end;
	}
	testTime();
	/* start the timer */
	oldtimers = CreateTimerThread(MAXTIMERS);
	PrimeTimer(oldtimers,InstallTimer(oldtimers,mOSVBL_timer),1);
	//timer_start = GetTickCount();
	Microseconds(&timer_start);

	/* initialize the world port */
	lockPixels(screen_desktop_window.port.map);
	pmErase(screen_desktop_window.port.map,screen_desktop_window.port.fgSolid);
	/* print info after possible inverse table building */
	printf("# depth: %dbpp, pSize=%d, minCmp=%d, pitch=%d\n",
		screen_desktop_window.port.map->bpp,screen_desktop_window.port.map->pSize,
		screen_desktop_window.port.map->minCmp,screen_desktop_window.port.map->pitch);
	/* as of 2014.08.09, origin works in all areas */
//	screen_desktop_window.port.origin.x =-8;
	screen_desktop_window.port.origin.y = 4;
	screen_desktop_window.port.clip.top = 8;
	screen_desktop_window.port.clip.bottom+= 8;

	/* initialize the graphics */
	tr->x = ov.left = 0x08A8000;
	tr->y = 0x0008000;
	tr[1].x = ov.right = 0x1248000;
	tr[1].y = ov.top = 0x01B8000;
	tr[2].x = 0x1208000;
	tr[2].y = ov.bottom = 0x0C08000;
	tr[3].x = 0x0718000;// rand()%(SCRW*0x4000)+SCRW*0x8000;
	tr[3].y = 0x0C88000;// rand()%(SCRH*0x4000)+SCRH*0x8000;
	ln->x = 0x1BE8000;
	ln->y = 0x1248000;
	ln[1].x = 0x24B8000;
	ln[1].y = 0x18A8000;
	rr.left = 0x10E8000;
	rr.top = 0x1448000;
	rr.right = 0x19B8000;
	rr.bottom = 0x1AA8000;
	/* initially white */
	lineColor = &rgb_white;

	_get_gray_pattern(grayPat);

	/* load all fonts */
	fmLoadMaster("Fonts.rsrc");
	fm = ffMatch(fmFonts[txfont&31],txsizes[txsize&7],
		screen_desktop_window.port.txStyle);
	tx.x = 96, tx.y = 400;
	txhw = rTextWidth(fm,txs,-1,NULL,0,0,screen_desktop_window.port.txStyle)/2;

	/* load the image */
	px.x = pxz.x = pxd.x = 96, px.y = pxz.y = pxd.y = 200;
	if ((pic=decompressPictFile(picName && *picName ? picName:"blot.pct",&r))==NULL)
		printf("## couldn't load image (%d)%s\n",
			r,!r ? " - data too short"
			:r==-2 ? " - out of memory"
			:r==-3 ? " - read error"
			:r==-8 ? " - unsupported format":"");
	else
		pxz.x = pxd.x = pic->w, pxz.y = pxd.y = pic->h;

#ifdef TEST_FILTER
	/* will come out half-intensity */
	RGBToPen(screen_desktop_window.port.map,&rgb_white); /* make sure an iTable is built */
	filterTable = calloc(8,2049);
	filterTable->flags = 0x8000;
	ctMakeGrays(filterTable->colors,-(filterTable->count=(screen_desktop_window.port.map->minCmp
		? 1u<<screen_desktop_window.port.map->minCmp:1024))*2);
#endif
	/* create the buttons */
	ctlb.top = 442, ctlb.left = 12;
	ctlb.bottom = 464, ctlb.right = 100;
	wNewControl(&screen_desktop_window,&ctlb,(const uint8*)"\5About",1,0,0,1,ctlTypePushBtn,0);
	ctlb.top = 445, ctlb.left = 112;
	ctlb.bottom = 461, ctlb.right = 184;
	cctl = wNewControl(&screen_desktop_window,&ctlb,(const uint8*)"\6Dither",1,0,0,2,ctlTypeCheckBox,'c');

	set_mouse_cursor(0,0);

	for(;;)
	{
		eventRecord	event;
		int32		tick, c;

		if (GetNextEvent(-1u,&event))
		{
			if (event.what==mouseDown)
			{
				pt2d16		loc = event.where;
				ctlObj		*ctl;
				int		r;

				global_to_local_pt(&loc);
				if ((r=wFindControl(&screen_desktop_window,loc,&ctl))>0)
					if (ctl->rtn(ctl,wcMsgClick,&loc))
						if (ctl->refCon=='c')
						{
							wSetControlValue(ctl,!ctl->value);
							goto upd;
						} else
							goto about;
			} else if (event.what==updateEvt)
			{
			//	wBeginUpdate(&screen_desktop_window);
			//	wEndUpdate(&screen_desktop_window);
			} else if (event.what==keyDown)
				if (event.modifiers&cmdKey)
				{
					if ((uint8)event.message=='q')
						event.what = closeEvent;
					else if ((uint8)event.message=='a'
							|| (uint8)event.message=='/')
					{
					about:	wSetControlTitle(screen_desktop_window.ctlList,(uint8*)"\5Close");
						wHiliteControl(cctl,255);
						ctlb.top = 16, ctlb.left = 32;
						ctlb.bottom = 432, ctlb.right = 608;
						if (boxer("README.txt",&ctlb)>0)
							break;
						wSetControlTitle(screen_desktop_window.ctlList,(uint8*)"\5About");
						wHiliteControl(cctl,0);
						if (!oldtimers)
							goto upd;
					}
				} else if ((uint8)event.message==kEscapeCharCode)
					{EndTimerThread(oldtimers); oldtimers = NULL;}
			if (event.what==closeEvent)
				break;
		}
		tick = TickCount(); //GetTickCount()>>4;
		if (tick==ltick)
			continue;
		ltick = tick;
		/* update for both left and right modifier keys */
		MGetKeys(mKeyMap);
	upd:	if (hasDrawn)
		{
			screen_desktop_window.port.pnMode = screen_desktop_window.port.txMode = mSrcCopy;
			RGBFore(&screen_desktop_window.port,&rgb_black);
			RGBBack(&screen_desktop_window.port,&rgb_black);
			rOval(&screen_desktop_window.port,&ov);
			rPolygon(&screen_desktop_window.port,tr,4);
			rLine(&screen_desktop_window.port,*ln,ln[1]);
			rTriangle(&screen_desktop_window.port,tr[2],tr[3],*tr);
			rRoundRect(&screen_desktop_window.port,&rr,16<<16,16<<16);

			if (pic)
				rFillRect(&screen_desktop_window.port,px.x,px.y,px.x+pxd.x,px.y+pxd.y);
			screen_desktop_window.port.txStyle = txstyl;
			screen_desktop_window.port.txSize = txsizes[txsize&7];
			rTextFont(&screen_desktop_window.port,txfont&31);
			rDrawText(&screen_desktop_window.port,txs,-1,tx.y,(tx.x-txhw)<<16|0x8000);
			RGBBack(&screen_desktop_window.port,&rgb_white);
		}
		if (animate&1)
		{
			tr->x+= (int16)rand();
			tr->y+= (int16)rand();
			tr[1].x+= (int16)rand();
			tr[1].y+= (int16)rand();
			tr[2].x+= (int16)rand();
			tr[2].y+= (int16)rand();
			ln->x+= (int16)rand();
			ln->y+= (int16)rand();
			ln[1].x+= (int16)rand();
			ln[1].y+= (int16)rand();
			ov.left+= (int16)rand();
			ov.top+= (int16)rand();
			ov.right+= (int16)rand();
			ov.bottom+= (int16)rand();
			rr.left+= (int16)rand();
			rr.top+= (int16)rand();
			rr.right+= (int16)rand();
			rr.bottom+= (int16)rand();
		}
		if (pmodet!=_MKDOWN(mKeyMap,MKV_I) && !pmodet)
			pmode = (pmode+1)&15;
		pmodet = _MKDOWN(mKeyMap,MKV_I);
		if (wpatt!=_MKDOWN(mKeyMap,MKV_H) && !wpatt)
			wpat^= 1;
		wpatt = _MKDOWN(mKeyMap,MKV_H);
		if (animatet!=_MKDOWN(mKeyMap,MKV_SPACE) && !animatet)
			animate^= 1;
		animatet = _MKDOWN(mKeyMap,MKV_SPACE);
		if (decrt!=_MKDOWN(mKeyMap,MKV_HYPHEN_USC) && !decrt) /* - key to decrease pensize */
			screen_desktop_window.port.penw = screen_desktop_window.port.penh-= screen_desktop_window.port.penh<0 ? 0:0x4000;
		decrt = _MKDOWN(mKeyMap,MKV_HYPHEN_USC);
		if (incrt!=_MKDOWN(mKeyMap,MKV_EQUALPLUS) && !incrt) /* = key to decrease pensize */
			screen_desktop_window.port.penw = screen_desktop_window.port.penh+= 0x4000;
		incrt = _MKDOWN(mKeyMap,MKV_EQUALPLUS);
		shapeColor.red = 65535;
		if (_MKDOWN(mKeyMap,MKV_L)) /* L to set the intensity (by v) */
			shapeColor.red = event.where.y*140;
		shapeColor.blue = shapeColor.red;
		shapeColor.green = MIN(event.where.x,510)*257>>1;

		/* A, S, D, and F for the quadrilateral */
		if (_MKDOWN(mKeyMap,MKV_A))
			tr[0].x = event.where.x<<16|0x8000, tr[0].y = event.where.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_S))
			tr[1].x = event.where.x<<16|0x8000, tr[1].y = event.where.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_D))
			tr[2].x = event.where.x<<16|0x8000, tr[2].y = event.where.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_F))
			tr[3].x = event.where.x<<16|0x8000, tr[3].y = event.where.y<<16|0x8000;
		/* Z and X for the line coords */
		if (_MKDOWN(mKeyMap,MKV_Z))
			ln->x = event.where.x<<16|0x8000, ln->y = event.where.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_X))
			ln[1].x = event.where.x<<16|0x8000, ln[1].y = event.where.y<<16|0x8000;
		/* C and V for the oval coords */
		if (_MKDOWN(mKeyMap,MKV_C))
			rr.left = event.where.x<<16|0x8000, rr.top = event.where.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_V))
			rr.right = event.where.x<<16|0x8000, rr.bottom = event.where.y<<16|0x8000;
		if (_MKDOWN(mKeyMap,MKV_P))
			px.x = event.where.x-(pxd.x>>1), px.y = event.where.y-(pxd.y>>1);
		if (_MKDOWN(mKeyMap,MKV_SLBRACKET))
		{
			pxd.x+= px.x, pxd.y+= px.y;
			px.x = event.where.x, px.y = event.where.y;
			pxd.x-= px.x, pxd.y-= px.y;
		}
		if (_MKDOWN(mKeyMap,MKV_SRBRACKET))
			pxd.x = event.where.x-px.x, pxd.y = event.where.y-px.y;
		if (_MKDOWN(mKeyMap,MKV_W))
			lineColor = &rgb_white;
		if (_MKDOWN(mKeyMap,MKV_Y))
			lineColor = chromo_colors+_rgb_yellow;
		if (_MKDOWN(mKeyMap,MKV_M))
			lineColor = chromo_colors+_rgb_magenta;
		if (_MKDOWN(mKeyMap,MKV_R))
			lineColor = chromo_colors+_rgb_red;
		if (_MKDOWN(mKeyMap,MKV_N))
			lineColor = chromo_colors+_rgb_cyan;
		if (_MKDOWN(mKeyMap,MKV_G))
			lineColor = chromo_colors+_rgb_green;
		if (_MKDOWN(mKeyMap,MKV_B))
			lineColor = chromo_colors+_rgb_blue;
		if (_MKDOWN(mKeyMap,MKV_K))
			lineColor = &rgb_black;
		if (_MKDOWN(mKeyMap,MKV_O))
			lineColor = &rgb_orange;
		if (_MKDOWN(mKeyMap,MKV_T))
			lineColor = &rgb_tan;
		if (_MKDOWN(mKeyMap,MKV_1))
			tx.x = event.where.x, tx.y = event.where.y;
		if (_MKDOWN(mKeyMap,MKV_2))
		{
			if (!(txstyl&0x80))
				++txstyl, txstyl&= 7;
			txstyl|= 0x80;
		} else
			txstyl&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_3))
		{
			if (!(txsize&0x80))
				txsize = (txsize+1)%GET_ARRAY_COUNT(txsizes);
			txsize|= 0x80;
		} else
			txsize&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_4))
		{
			if (!(txfont&0x80))
			{
				c = 31;
				do
					++txfont, txfont&= 31;
				while(!fmFonts[txfont] && --c);
			}
			txfont|= 0x80;
		} else
			txfont&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_5))
		{
			if (!(txgray&0x80))
				txgray^= mTextGray;
			txgray|= 0x80;
		} else
			txgray&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_6))
		{
			if (!(txspex&0x80) && (screen_desktop_window.port.spExtra+=65536/1.5+.5)>20*65536)
				screen_desktop_window.port.spExtra = 0;
			txspex|= 0x80;
		} else
			txspex&=~0x80;
		if (_MKDOWN(mKeyMap,MKV_0)|_MKDOWN(mKeyMap,MKV_SEMICOLON))
		{
			if (!cctl->hilite)
				cctl->value = !cctl->value*(!!_MKDOWN(mKeyMap,MKV_SEMICOLON)+1);
			cctl->hilite = 255;
		} else
			cctl->hilite = 0;
		if (screen_desktop_window.port.map->bpp==1
				&& formGray(shapeColor.red,shapeColor.green,shapeColor.blue,c)
				-65536u<24576u)
		{
			RGBFore(&screen_desktop_window.port,&rgb_black);
			rPenPat(&screen_desktop_window.port,grayPat);
		} else
			RGBFore(&screen_desktop_window.port,&shapeColor);
		rOval(&screen_desktop_window.port,&ov);
		rPenPat(&screen_desktop_window.port,NULL);

		RGBFore(&screen_desktop_window.port,&shapeColor);

		screen_desktop_window.port.pnMode = screen_desktop_window.port.txMode = pmode|txgray;
		rPolygon(&screen_desktop_window.port,tr,4);

		RGBFore(&screen_desktop_window.port,lineColor);
		rLine(&screen_desktop_window.port,*ln,ln[1]);

		screen_desktop_window.port.opColor.red = 65535*3/9;
		screen_desktop_window.port.opColor.green = 65535*4/9;
		screen_desktop_window.port.opColor.blue = 65535*5/9;
		RGBFore(&screen_desktop_window.port,system_colors+_system_gray87);
		rRoundRect(&screen_desktop_window.port,&rr,16<<16,16<<16);
		rPenPat(&screen_desktop_window.port,wpat ? weavePat:grayPat);
		RGBFore(&screen_desktop_window.port,screen_desktop_window.port.map->bpp==1
			? &rgb_black:system_colors+_system_gray73);
		rTriangle(&screen_desktop_window.port,tr[2],tr[3],*tr);
		rPenPat(&screen_desktop_window.port,NULL);

		hasDrawn = 1;
		RGBFore(&screen_desktop_window.port,pmode==(mSrcCopy|mTransparent)
			? system_colors+_system_gray53:&rgb_black);

		rBltPix(&screen_desktop_window.port,pic,0,0,pxz.x,pxz.y,px.x,px.y,pxd.x,pxd.y,
			pmode|(cctl->value ? mDither:0));

		screen_desktop_window.port.txStyle = txstyl;
		screen_desktop_window.port.txSize = txsizes[txsize&7];
		rTextFont(&screen_desktop_window.port,txfont&31);
		rDrawText(&screen_desktop_window.port,txs,-1,tx.y,(tx.x-txhw)<<16|0x8000);
#if DISPLAY_STATS
		RGBFore(&screen_desktop_window.port,&rgb_black);
		displayStats(*fmFonts,0,screen_desktop_window.port.map->h+2);
#endif
#ifdef TEST_FILTER
		pmFilter(screen_desktop_window.port.map,filterTable);
#endif
		wDrawControls(&screen_desktop_window);
		update_screen_rect(0,0,screen_desktop_window.port.map->w,
			screen_desktop_window.port.map->h);
	}
end:	wKillControls(&screen_desktop_window);
	wSetVideoMode(-1,0,0,0,0);
	EndTimerThread(oldtimers);
	//Sleep(10);
	fmUnloadMaster();
	//if (!ftell(stdout))
	//	fclose(stdout), remove("stdout.txt");
	return 0;
}
