#include <windows.h>
#include <time.h>
#include <stdio.h>
#include <fenv.h>

#include "../mos/mevent.h"

static uint32	mcs_freq;
static double	mcs_freqsub;

void Microseconds(UnsignedWide *mcs)
{
	LARGE_INTEGER	tt;
	uint64		n;

	if (!mcs_freq)
	{
		if (!QueryPerformanceFrequency(&tt) || (mcs_freq=tt.LowPart)==0)
		{
			//mcs_freq = 10000000;
			*mcs = GetTickCount()*1000ll;
			return;
		}
		mcs_freqsub = 1000000./mcs_freq;
	}
	QueryPerformanceCounter(&tt);
	n = tt.QuadPart/mcs_freq;
	*mcs = n*1000000u+(tt.QuadPart-n*mcs_freq)*mcs_freqsub;
}

void native_perf_test(void)
{
	LARGE_INTEGER	freq, count;
	long long	ltick, etick;
	time_t		time1, time2;
	int		i;

	QueryPerformanceFrequency(&freq);
	printf("%lld perf clocks/sec\n",
		freq.QuadPart);
	for(time1=time(NULL); (time2=time(NULL))==time1; )
		;
	QueryPerformanceCounter(&count);
	ltick = count.QuadPart;
	for(i=1; i<=9; ++i)
	{
		for(time1=time2; (time2=time(NULL))==time1; )
			;
		QueryPerformanceCounter(&count);
		etick = count.QuadPart-ltick;
		printf("%d - %lld ticks, %.12g avg\n",
			i,etick,(double)etick/i/freq.QuadPart);
	}
}

void mcs_perf_test(void)
{
	UnsignedWide	ltick, etick;
	time_t		time1, time2;
	int		i;

	Microseconds(&ltick); /* make sure the function is accelerated */
	printf("\nMicroseconds accuracy test\n");
	for(time1=time(NULL); (time2=time(NULL))==time1; )
		;
	Microseconds(&ltick);
	for(i=1000000; i<=9000000; i+=1000000)
	{
		for(time1=time2; (time2=time(NULL))==time1; )
			;
		Microseconds(&etick);
		etick-= ltick;
		printf("%d - %lld mcs, %.10g avg\n",
			i,etick,(double)etick/i);
	}
}

int __cdecl main(void)
{
	native_perf_test();
	mcs_perf_test();
	return 0;
}
