/*
 IIgsRMap.h
  Apple IIgs Sys 5+ resource fork types and structures.
  All numbers here are LSB.

  Corrected: the IIgs prefixes were 'ProDOS'.

  Compared to Mac OS rsrc forks, this format is more expandable,
however less efficient.  Neither system can read the other.
  RMover: Error 1E02 attempting to open "HCGS.Read.Me" (Mac OS rfork).
*/

/* IIgs rsrc forks use 16-bit hex codes, not OSTypes */
enum
{ /* resource types */
	rIcon			= 0x8001,
	rPicture,
	rControlList,
	rControlTemplate,	/* 0x8004 */
	rC1InputString,
	rPString,		/* 'STR ' equivalent */
	rStringList,
	rMenuBar,		/* 0x8008 */
	rMenu,
	rMenuItem,
	rTextForLETextBox2,
	rCtlDefproc,		/* 0x800C - OMF */
	rCtlColorTable,
	rWindParam1,
	rWindParam2,
	rWindColor,		/* 0x8010 */
	rTextBlock,
	rStyleBlock,
	rToolStartup,
	rResNames,		/* 0x8014 */
	rAlertString,
	rText,
	rMenuStrg,
	rCDEVCode,		/* 0x8018 - OMF */
	rCDEVFlags,
	rTwoRects,
	rFileTypes,
	rListRef,		/* 0x801C */
	rCString,
	rXCMD,			/* OMF */
	rXFCN,			/* OMF */
	rErrorString,		/* 0x8020 */
	rKTrans,
	rSoundSample		= 0x8024,
	rCursor			= 0x8027,
	rVersion		= 0x8029,
	rComment,
	rTaggedStrings		= 0x802E,
	rPatternList,
	rRectList		= 0xC001,
	rPrintRecord,
	rFont
};

enum
{ /* resource flags/attrs */
	fLocked			= 0x8000,
	fFixed			= 0x4000,
	fConverter		= 0x0800, /* OMF */
	fAbsoluteLoad		= 0x0400,
	/* purge level bits */
	fWriteProtect		= 0x0080,
	fPreload		= 0x0040,
	fNoBankCross		= 0x0010,
	fNoSpecialMem		= 0x0008,
	fPageAligned		= 0x0004
};

/* 16-bit file type, 32-bit 'auxiliary' type */
typedef struct IIgsHeader
{
	int32		type; /* must be zero */
	int32		mapOffset, mapSize;
	int32		reserved; /* must be zero */
/*0x10*/int16		unused[16];
	/* here, it's the same fInfo copy as Mac OS at offset 0x30 */
	/* (except, of course, the 15-char. limit to a file name,
	  and fTypes aren't meant for OSTypes)
	*/
/*0x30*/Str15		name;
	int16		scratch[9];
/*0x42*/uint8		fSType; /* 0x70 ('p') */
	char		fType[3]; /* auxiliary is around here somewhere */
	OSType		fCreator; /* 'pdos' */
	int16		flag;
	/* the next 4 bytes may be the Finder icon position - not sure */
} IIgsHeader;

typedef struct IIgsMapHeader/*(32+)*/
{
	int32		mptr;		/* active map handle (temporary) */
	int16		index;		/* active index (temporary) */
	/* these two must match those in the fork header */
	int32		mapOffset, mapSize;
	uint16		mapListOffset, flag;
	int16 reserved;
	int32		extWLength;	/* leading map data (temporary?) */
	int32		rsrcCount;
	/* various other things unverified, ignored by ResEx */
} IIgsMapHeader;

/* unlike Mac OS, resource types are 16-bit, IDs are 32-bit that can't be 0 */
/* IDs are 32-bit because names are tied by ID (typeCode+0x10000 for names ID) */
typedef struct IIgsEntry/*(20)*/
{
	uint16		type;
	int32		ID;
	int32		fileOffset;
	uint16		flag;
	int32		length;
	int32		mptr; /* active handle (temporary) */
} IIgsEntry;

/* resource names are in themselves part of a resource */
/* the Tour stack has redundant copies */
typedef struct IIgsName
{
	int32		ID;
	uint8		pstr[1];
} IIgsName;

/* the names resource has some trailing data at the end */
typedef struct IIgsNames
{
	int16 type; /* = 1 */
	int32		count;
	IIgsName		names[1];
} IIgsNames;

/* the length of sound clip data is implied; rsrc (size-10)/chanCount samples */
/* NOTE: the DOC treats sounds like C-strs in that nulls terminate playback */
typedef struct IIgsSoundClip
{
	int16		format; /* must be 0 */
	uint16		pageCount; /* 256-byte page count */
	uint16		relPitch; /* semitone.q */
	uint8		stereoChannel;
	uint8	reserved; /* actually part of LSB 'stereo' field; set to zero */
	uint16		sampleRate; /* 0=use default (11127?) */
	uint8		samples[1]; /* always uint8 */
} IIgsSoundClip;

typedef struct IIgsIcon
{
	int16 type;
	uint16		wsize; /* 2-byte count data size */
	int16		height;
	int16		width; /* double-resolution */
} IIgsIcon;

typedef struct IIgsCursor
{
	uint16		height;
	uint16		width; /* in words */
	uint16		curImg[1];
	/* mask[], hotspot.pt, flags.w, reserved.l[2]=0 */
} IIgsCursor;
