/*
 main.c
  ResEx tool shell layer

  Revised: 2014.03.31

2014.02.28
  added 8027, 802E-F, C001-3, FWID, KCN#, KSWP, MENA, alrx, dftb, dlgx,
dlle, fctb, flac, hovr, kcl4/8, kcs8, mcfg, mgrp, mime, mitq, pcki, pslt,
resf, rsmi, stg#, src#, tab#, thga, thn#, thnd, thnr, trk#, xmnu & wind.
  fixed 802A - rComment, from Finder info (the only use I saw).
*/

#include "../my_cseries.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include "resex.h"
#include "interpreter.h"

/* ---- main code */

typedef char	typeName[32];

static const typeName	rsrcTypeNames[] =
{
	".256Shapes Collection",
	"8001rIcon",
	"8002rPicture",
	"8003rControlList",
	"8004rControlTemplate",
	"8005rC1InputString",
	"8006rPString",
	"8007rStringList",
	"8008rMenuBar",
	"8009rMenu",
	"800ArMenuItem",
	"800BrTextForLETextBox2",
	"800CrCtlDefproc OMF",
	"800DrCtlColorTable",
	"800ErWindParam1",
	"800FrWindParam2",
	"8010rWindColor",
	"8011rTextBlock",
	"8012rStyleBlock",
	"8013rToolStartup",
	"8014rResNames",
	"8015rAlertString",
	"8016rText",
	"8017rMenuStrg", /* may also be rCodeResource */
	"8018rCDEVCode OMF",
	"8019rCDEVFlags",
	"801ArTwoRects",
	"801BrFileTypes",
	"801CrListRef",
	"801DrCString",
	"801ErXCMD OMF",
	"801FrXFCN OMF",
	"8020rErrorString",
	"8021rKTrans",
	"8024rSoundSample",
	"8027rCursor",
	"8029rVersion", /* newer - versnum.l region.b name.pstr copyright.pstr */
	"802ArComment", /* pstr */
	"802ErTaggedStrings",
	"802FrPatternList",
	"ALRTAlert Dialog",
	"BNDLIcon Bundle",
	"C001rRectList",
	"C002rPrintRecord",
	"C003rFont",
	"CCI\xAAProduct Description",
	"CCODM68K Code",
	"CDEFM68K Control Def",
	"CNTLControl Description",
	"CODEM68K Code Segment",
	"CURSCursor-Mono",
	"DESCDescription P-String",
	"DITLDialog Item List",
	"DLOGDialog",
	"DRELData Relocation Table",
	"DRVRM68K Driver/Accessory",
	"FKEYF-Key Shortcut Code",
	"FONDFont Description",
	"FONTBitmapped Font",
	"FREFBundle File Ref",
	"FRSVReserved Fonts-IDs",
	"FWIDFont Width Table",
	"ICN#32x32 Icon/Mask-Mono",
	"ICON32x32 Icon-Mono",
	"INITM68K Extension Code",
	"INTLNumber Formatting",
	"KCHRKey Character Map",
	"KCN#Keyboard Icons-Mono",
	"KSWPKeyboard Swapping",
	"LDEFM68K List Def",
	"MBARMenubar Description",
	"MBDFM68K Menubar Def",
	"MDEFM68K Menu Definition",
	"MDVRM68K Driver",
	"MENAMenu Acc. Key-Win32",
	"MENUBasic Menu",
	"MPSRMPW Record",
	"NFNTBitmap Font",
	"NcmpPPC Decompressor",
	"PACKM68K Code Package",
	"PAT Pattern-Mono",
	"PAT#Pattern List-Mono",
	"PICTQuickdraw Picture",
	"PROCM68K Procedure Code",
	"PPobWindow Template",
	"PTCHM68K Patch Code",
	"SERDSerial Driver",
	"SEngStuffIt Engine",
	"SICNSmall Icon List",
	"SIZEApp Allocation Info",
	"SOCKNet Socket Code",
	"STR P-String",
	"STR#P-String List",
	"TMPLRsrc Template",
	"WDEFM68K Window Def",
	"WINDWindow Description",
	"XCMDM68K External Cmd",
	"XCODM68K External Code",
	"XFCNM68K External Func",
	"actbAlert Color Table",
	"acurAnimated Cursor List",
	"adioAudio Code",
	"aeteApple Event Table",
	"alrxExtended Alert",
	"atlkAppleTalk Code",
	"bcemDiskCopy Info",
	"bootBoot Code",
	"cardSlot Card Description",
	"cctbControl Color Table",
	"cfrgCode Fragment Info",
	"cicnColor Icon",
	"clokClock Timer Code",
	"clutColor Table",
	"cmdoMPW Commando",
	"crsrCursor-Color",
	"dcmpRsrc Decompressor",
	"dctbDialog Color Table",
	"dftbDialog Control Fonts",
	"dimgDisk Image Driver",
	"dlgxExtended Dialog",
	"dlleMultiplatform Things", /* cstr */
	"drvrDriver",
	"fctbFont Color Table", /* added 2014.02.28 */
	"finfFont Info List",
	"flacFlattened Collection",
	"fld#System Folder List",
	"flstFont Info List",
	"fmapFileType Map",
	"fovrPPC Library",
	"ftagFile Tag",
	"fxltFile Translation",
	"gpchGeneral Patch",
	"hdlgHelp Balloon-Dialog",
	"hfdrHelp Balloon-Finder",
	"hmnuHelp Balloon-Menu",
	"hovrHelp Mgr Override",
	"hrctHelp Balloon-Rectangle",
	"hwinHelp Balloon-Windows",
	"icl432x32 Icon-4 bit",
	"icl832x32 Icon-8 bit",
	"icm#Mini Icon/Mask-Mono",
	"icm4Mini Icon-4 bit",
	"icm8Mini Icon-8 bit",
	"icnsIcon Collection",
	"ics#Small Icon/Mask-Mono",
	"ics4Small Icon-4 bit",
	"ics8Small Icon-8 bit",
	"ictbItem Style",
	"itl0Number Formatting",
	"itl1Date Formatting",
	"kcl4Large KB Icon-4 bit",
	"kcl8Large KB Icon-8 bit",
	"kcs#Small KB Icon-Mono",
	"kcs4Small KB Icon-4 bit",
	"kcs8Small KB Icon-8 bit",
	"kindDesktop Kind Info",
	"ldesList Box Description",
	"mcfgQT MIME Import Config",
	"mctbMenu Color Table",
	"mgrpQT Media Group",
	"mimeQT MIME Import Info", /* 'sean' type */
	"mitqMakeITable Queues",
	"mntrM68K Monitor Driver",
	"mst#Item Language Table",
	"mstrMenu Item P-String",
	"ncmpPPC Decompressor",
	"ndrvPPC Driver",
	"nrctRectangle List",
	"nsndPPC Sound Driver",
	"nsrdSerial Driver",
	"ntrbPPC Trap Patch",
	"openDesktop Open Enabler",
	"osaxM68K AppleScript",
	"otlmOpenTransport Loader",
	"pckiQTStreaming Pack Info",
	"perfCode Performance",
	"plttColor Palette",
	"pnotPreview Rsrc List",
	"ppatColor Pattern",
	"ppt#Color Pattern List",
	"procM68K Procedure",
	"psltNuBus pseudo-slots",
	"ptchM68K Patch",
	"puf!Text Mappings",
	"qtcmQuickTime Component",
	"rgb 48-bit RGB Color",
	"resfReserved Fonts",
	"rsmiReassembler Info",
	"scodSystem Code",
	"scptAppleScript",
	"scszScript Requirements",
	"siftM68K Sound Mixer",
	"snd Sound List",
	"snthSound Synthesizer",
	"sprtSprocket Data",
	"src#MovieExporter Sources",
	"ssaiMusical Instrument",
	"stg#QT Presets",
	"stgpQT Preset Platforms",
	"stylText Style",
	"syszSys RAM Requirement",
	"tab#Tab Control",
	"thgaCM Thing",
	"thn#CM Thing Order",
	"thndCM Thing Dependency",
	"thngCM Thing List",
	"thnrCM Thing Public Map",
	"trk#MovieExporter Tracks",
	"ttssPhoneme Symbols",
	"vdigPPC Video Driver",
	"vdpmM68K Video Driver",
	"versVersion Info",
	"wctbWindow Color Table",
	"windExtensible Window-flac",
	"xcmdPPC External Command",
	"xfcnPPC External Func",
	"xmnuExtended Menu"
};

static const char *getResTypeName(int32 t)
{
	const typeName	*s = rsrcTypeNames, *f;
	uint32		c = GET_ARRAY_COUNT(rsrcTypeNames);
	int32		u = be32_native(t), v;
	/* a sorted list means we can speed up the search with this */
	while(c>4)
	{
		v = *(const int32*)(f=s+c/2);
		if (u<be32_native(v))
			c = (c+1)>>1;
		else
			s = f, c-= c/2;
	}
	do
		if (*(const int32*)s==t)
			return 4+(char*)s;
	while(++s,--c);
	return NULL;
}

static const char *getMapTypeName(ResFile_Memory *m)
{
	switch(m->type)
	{
	case TYPE_RFMACOS: return "Mac OS";
	case TYPE_RFIIGS: return "Apple IIgs";
	case TYPE_RSWINRC: return "Windows";
	default: return "???";
	}
}

void listMap(ResFile_Memory *m,
	uint8 printOffs)
{
	RsrcInfo	info;

	ResFile_TypeEntry	*l;

	const char	*s;
	int32		t;
	int		i, j, c, err;

	if (m->map.totalCount<=0)
		return;
	puts("=======================");
	printf(m->options&FAIL_RECONSTRUCT ?
"## abnormal map - list reconstructed\n"
"   index    size %s\n\n"
		:" ix/abs    ID    size  %s flags name\n\n",
		printOffs ? " offset":"");
	for(i=t=0,l=m->map.typeEntries; ++i<=m->map.typeCount; ++l)
	{
		err = 0, c = ResFile_CountResources(m,l->typeCode,&err);
		if (err)
		{
			printf("## listMap: error getting type #%d\n",
				i);
			continue;
		}
		printf("- type #%-2d- ",
			i);
		if (m->type!=TYPE_RFIIGS)
			fputc('\'',stdout);
		print_tchars(stdout,(char*)&l->typeCode,4,'\'');
		if (m->type!=TYPE_RFIIGS)
			fputc('\'',stdout);
		printf((s=getResTypeName(l->typeCode))!=NULL ? " (%s)\n":"\n",
			s);
		for(j=0; j<c; ++t,++j)
		{
			err = 0, ResFile_GetIndResource(m,
				l->typeCode,~j,&info,&err);
			if (err)
			{
				printf("## listMap: error getting #%d\n",
					j);
				continue;
			}
			printf("%3d/%-3d ",
				j,t);
			printf(info.ID>65535 ? "$%X":"%6d",
				info.ID,info.sizeOnDisk);
			printf(printOffs ? " % 7d  %06X":" % 7d",
				info.sizeOnDisk,info.offsetOnDisk);
			printf(m->type!=TYPE_RFMACOS ? "  %04X":"  %02X",
				info.flag);
			if (*info.name)
			{
				fputs(" \"",stdout);
				print_tchars(stdout,(char*)info.name+1,*info.name,'"');
				fputc('"',stdout);
			}
			fputc('\n',stdout);
		}
	}
}

typedef struct options
{
	long int	offset;
	boolean		rsrc, info, list, update;
} options;

int process(const char *in,
	options *opt)
{
	FILE		*f = NULL;
	ResFile_Memory	*m = NULL;
	int		fd =-1, err = 0, r = strlen(in);
	char		*n = malloc(r+15); /* excess for native RF appendage */

	if (!n)
		return -2;
	memcpy(n,in,r); n[r] = 0, r = 0;
	if (opt->rsrc)
	{
#if TARGET_OS_MAC
		if ((fd=_open(n,O_RDWR|O_BINARY|O_RSRC))==NULL)
			goto ferrno;
		if (fd<0)
			{err = fd; goto ferr;}
		f = fdopen(fd,"rb");
#endif
		memcpy(n+r,"/resource",9);
	}
	if (!f)
		f = fopen(n,"rb");
	if (!f)
		goto ferrno;
#if TARGET_OS_MAC
	Debugger();
#endif
	if ((err=ResFile_LoadMap(f,TOPT_MAKERAW|TOPT_VERBOSE,&m,opt->offset))!=0)
	{
		fprintf(stderr,"### ResFile_LoadMap %s (%d)\n",
			err<0 ? "error":"warning",err);
		if (err<0)
			goto end;
	}
	if (opt->update)
	{
		if (fd<0)
			f = freopen(n,"rb+",f);
	#if TARGET_OS_MAC
		else
			f = _fdopen(fd,"rb+");
	#endif
		if (!f)
		{
		ferrno:	err = errno;
	#if TARGET_OS_MAC
		ferr:
	#endif
			fprintf(stderr,"### file error (%d)\n",
				err);
			if (err==ENOENT)
				fprintf(stderr," - '%s' could not be found\n",
					n);
			else if (err==EACCES)
				fprintf(stderr," - could not (re-)open '%s'\n",
					n);
			else if (err==EINVAL)
				fprintf(stderr," - illegal characters name '%s'\n",
					n);
			r = 1;
			goto end;
		}
		if ((err=ResFile_Update(m))!=0)
		{
			fprintf(stderr,"### update %s (%d)\n",
				err<0 ? "error":"warning",err);
			goto end;
		}
	}
/* checked - 02.14
	dispose_mhandle(ResFile_GetResource(m,
		MK4CC('s','n','d',' '),(uint8*)"\005Sound",0,NULL));
	ResFile_GetResourceIndex(m,MK4CC('s','n','d',' '),(uint8*)"\005Sound",0);
*/
	if (opt->info)
	{
		int		enz, ovz = ResFile_OverheadSize(m,&enz);

		printf(
"# file: '%s'\n"
"# RF type: %s\n",
			n,getMapTypeName(m));
		if (m->type==TYPE_RFMACOS) /* don't know what of other types */
			printf("# attributes: %04X%s\n",
				m->map.attributes,
				m->map.attributes&RF_READ_ONLY_ATTR ? " (read-only)":"");
		printf(
"# map length: %d; per-rsrc overhead: %d+%d\n"
"# type count: %d; total: %d\n\n",
			m->mapLen,ovz,enz,m->map.typeCount,m->map.totalCount);
		if (opt->list)
			listMap(m,opt->list-1);
	}

end:	if (f)
		fclose(f);
	if (n)
		free(n);
	if (m)
		free(m);
	return r;
}

int main(int argc,
	char **argv)
{
	options		opt;
	char		**a = argv+1, *s;

	if (argc<1)
		return 0;
	memset(&opt,0,sizeof(opt));
	for(; --argc && *(s=*a)=='-'; ++a)
		if (!(++s)[1])
			if (*s=='i')
				opt.info^= 1;
			else if (*s=='l')
				opt.list = 1;
			else if (*s=='L')
				opt.list = 2;
			else if (*s=='o')
			{
				s = *++a;
				if (!--argc || !readanyinteger(s,strlen(s),&opt.offset))
				{
					fprintf(stderr,"### bad offset value '%s'\n",
						s);
					return 1;
				}
			} else if (*s=='r')
				opt.rsrc^= 1;
			else if (*s=='u')
				opt.update^= 1;
			else
				goto unk;
		else
		{
		unk:	fprintf(stderr,"### unknown option '%s'\n",
				s-1);
			goto usage;
		}
	opt.info|= opt.list;
	if (argc<1)
	{
	usage:	fprintf(stderr,
"# usage: %s [options] files\n"
"#\n"
"# -i            print info\n"
"# -l            print info and listing\n"
"# -L            .. listing with offsets\n"
"# -o            offset\n"
"# -r            native RF (/resource)\n"
"# -u            update\n",
			*argv);
		return 1;
	}
	//cs_initialize(NULL);
	for(; argc; --argc,++a)
		if (process(*a,&opt))
			return 1;
	return 0;
}
