/*
 main.c
  executable shell (demo)
   attempts to play on single path (drag & drop one)
  under Windows:
   escape to stop
   M for mute
   space to pause/resume
   +/= for volume
   KP+/KP- for rate

 Note: there's a small delay in the response time.

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#include <stdio.h>
#include <math.h>
#include <enviro64.h>
#include "snddec.h"

/* ---- library */

const char *readanyinteger(const char *s,
	long z,
	long int *v)
{
	const char	*e;
	uint32		r = 0, m = 0;
	uint8		c;

	if (!s || z<=0)
		return 0;
	e = z+s, c = *s++;
	if ((c+=-'-')==0)
	{
		if (s>=e)
			return 0;
		m =-1, c = *s++;
	} else if ((c+='-'-'+')==0)
	{
		if (s>=e)
			return 0;
		c = *s++;
	} else if ((uint8)((c+='+')-'0')>(uint8)9)
	{
		if (c==36)
			goto hexadecimal;
		return 0;
	}
	if (c!='0')
		goto decimal;
	if (s>=e)
		goto end;
	c = *s++;
	if (c!='x' && c!='X')
	{
		if (c=='o' || c=='O')
		{ /* octal (starting-zero-only not supported here) */
			do
			{
				c = *s++, c-= '0';
				if (c>(uint8)7)
					if (c<=(uint8)('9'-'7')
							|| (c+='0'-'A')<=(uint8)('Z'-'A')
							|| (c+='A'-'a')<=(uint8)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				r = (r<<3)+c;
			} while(s<e);
			goto end;
		}
		if (c=='b' || c=='B')
		{ /* binary */
			do
			{
				c = *s++, c-= '0';
				if (c>(uint8)1)
					if (c<=(uint8)('9'-'1')
							|| (c+='0'-'A')<=(uint8)('Z'-'A')
							|| (c+='A'-'a')<=(uint8)('z'-'a')
							|| (c+='a'-'_')==0)
						{s = 0; break;}
					else
						{--s; break;}
				r+= r, r+= c;
			} while(s<e);
			goto end;
		}
		goto decimal;
		do
		{
			c = *s++;
		decimal:
			c-= '0';
			if (c>(uint8)9)
				if ((c+='0'-'A')<=(uint8)('Z'-'A')
						|| (c+='A'-'a')<=(uint8)('z'-'a')
						|| (c+='a'-'_')==0)
					{s = 0; break;}
				else
					{--s; break;}
			r = r*10+c;
		} while(s<e);
		goto end;
	}
hexadecimal:
	if (s>=e) /* only starting character? */
		return 0;
	do
	{
		c = *s++, c-= '0';
		if (c<=(uint8)('9'-'0'))
			r = r*16+c;
		else if ((c+='0'-'A')<=(uint8)('Z'-'A')
				|| (c+='A'-'a')<=(uint8)('z'-'a'))
			if (c>(uint8)('F'-'A'))
				{s = 0; break;}
			else
				c+= 10, r = r*16+c;
		else if ((c+='a'-'_')==0)
			{s = 0; break;}
		else
			{--s; break;}
	} while(s<e);
end:	r+= m, r^= m;
	if (v)
		*v = r;
	return s;
}

/* ---- main code */

static long readBufferCmdHeader(FILE *s,
	sndFormat *f,
	uint8 *n)
{
	uint8		b[64];

	if (!s || !f)
		return -1;
	if (fread(b,1,sizeof(b),s)<22)
		return -3;
	return parseBufferCmdHeader(b,f,n);
}

#ifndef NO_SNDMGR
 #include <stdlib.h>
 #include <string.h>
 #include <windows.h>	//no longer needed to compile

static void setChSettings(sndChannel *ch,
	int32 vol,
	ufixed rmul)
{
	sndCommand	cmd;

	cmd.cmd = sndLRVolumeCmd;
	cmd.param2 = (void*)vol;
	sndDoCommand(ch,&cmd,1);

	cmd.cmd = sndRateMultiplierCmd;
	cmd.param1 = 0;
	cmd.param2 = (void*)rmul;
	sndDoCommand(ch,&cmd,1);

	//cmd.cmd = sndWaitCmd;
	//cmd.param2 = (void*)1;
	//sndDoCommand(ch,&cmd,1);
}

static void sndWaitLoop(sndManager *mgr,
	sndChannel *ch,
	int32 vol,
	ufixed rmul)
{
#ifdef WINUSERAPI
	int8		r, pauset = 0, incrt = 0, decrt = 0, incvt = 0, decvt = 0;
	int8		mutet = 0, mutev = sndEnableLeft|sndEnableRight;
#endif
	while(sndChannelBusy(ch))
	{
	#ifdef WINUSERAPI
		if (GetAsyncKeyState(VK_ESCAPE))
			break;
		r = GetAsyncKeyState(VK_M)&1;
		if (mutet!=r && r)
		{
			mutev^= sndEnableLeft|sndEnableRight;
			sndMgrMute(mgr,mutev);
		}
		r = GetAsyncKeyState(VK_SPACE)&1;
		if (pauset!=r && r)
			ch->actFlag^= 2;
		pauset = r;
		r = GetAsyncKeyState(VK_ADD)&1;
		if (incrt!=r && r)
			setChSettings(ch,vol,rmul=(rmul*33+16)/32);
		incrt = r;
		r = GetAsyncKeyState(VK_SUBTRACT)&1;
		if (decrt!=r && r)
			setChSettings(ch,vol,rmul=(rmul*32+16)/33);
		decrt = r;
		if (GetAsyncKeyState(VK_NUMPAD0))
			setChSettings(ch,vol,rmul=0x10000);
		r = GetAsyncKeyState(VK_OEM_PLUS)&1;
		if (incvt!=r && r)
			setChSettings(ch,vol+=0x200020,rmul);
		incvt = r;
		r = GetAsyncKeyState(VK_OEM_MINUS)&1;
		if (decvt!=r && r)
			setChSettings(ch,vol-=0x200020,rmul);
		decvt = r;
	#endif
		sndMgrProcWait(mgr);
	}
	sndMgrWait(mgr);
}

static int loadSndPlay(FILE *f,
	long z,
	long o,
	int cmdc,
	int32 vol,
	ufixed rmul)
{
	sndCommand	cmd;
	sndManager	*mgr = NULL;
	sndChannel	*ch = NULL;
	uint8		*p;
	const uint8	*s;
	long int	act;
	int		r, v, c;

	if ((c=cmdc)<=0)
		return -1;
	if ((p=calloc(z,1))==NULL)
		return -2;
	if ((act=fread(p,1,z,f))<z)
	{
		fprintf(stderr,"## EOF error reading %d bytes\n",
			z);
		r =-3;
		goto end;
	}
	z = act;
	if ((mgr=sndMgrOpen(0,1,&r))==NULL)
	{
		fprintf(stderr,"### unable to open sndManager (%d)\n",
			r);
		goto err;
	}
	if ((ch=sndNewChannel(mgr,sndInitStereo))==NULL)
	{
		fprintf(stderr,"### unable to create sndChannel\n");
		goto err;
	}
#ifdef SNDPLAY_MORECMDS //neither SndPlay nor ResEdit seem to do this
	v = MAX(c,2);
	if ((mgr->sndChannels=realloc(ch,OFFSET(sndChannel,queue)
			+sizeof(sndCommand)*v))==NULL)
	{
		fprintf(stderr,"## unable to resize channel for count\n");
		mgr->sndChannels = ch;
	} else
		ch->qLength = v;
#endif
	setChSettings(ch,vol,rmul);
	s = p;
	do
	{
		v = va_be16(s);
		cmd.param1 = va_be16(s);
		cmd.param2 = (void*)va_be32(s);
		if ((z-=8)<0)
		{
			fprintf(stderr,"## surpassed SndList end\n");
			break;
		}
		switch(v)
		{
		case 86: v = sndRateMultiplierCmd;
			cmd.param1 = 0;
			break;
		case 46: v = sndLRVolumeCmd;
			break;
		case sndNullCmd:
		case sndQuietCmd:
		case sndFlushCmd:
		case sndWaitCmd:
		case sndReInitCmd:
		case sndPauseCmd: /* not a great idea allowing all of these... */
		case sndResumeCmd:
		case sndSyncCmd: /* not the same! */
			break;
			/* these would just overwrite the SndCommand on the stack */
		case 24: /* availableCmd */
		case 25: /* versionCmd */
		case 47: /* getVolumeCmd */
		case 83: /* continueCmd - obsolete */
		case 85: /* getRateCmd - obsolete */
		case 87: /* getRateMultiplierCmd */
			continue;
		case 0x8050:
		case 0x8051:
			v = sndBufferCmd, cmd.param2 = p+(int32)cmd.param2-o;
			break;

		default:
			fprintf(stderr,"## unsupported SndCommand number #%d\n",
				v);
		}
		cmd.cmd = v;
		sndDoCommand(ch,&cmd,0);
	} while(--c);
	printf("Playing: %d command%s...\n",
		cmdc,cmdc==1 ? "":"s");
	sndWaitLoop(mgr,ch,vol,rmul);
	printf("Finished.\n");
err:	r =-128;
end:	if (mgr)
		sndMgrClose(mgr);
	if (p)
		free(p);
	return r;
}

static int playFile(FILE *in,
	sndFormat *f,
	int32 k,
	int32 vol,
	ufixed rmul)
{
	sndManager	*mgr = NULL;
	sndChannel	*ch = NULL;
	int		r = 0;

	r = f->rate+(f->rateFrac>>31);
	if (!mgr && (mgr=sndMgrOpen(r<32000 || rmul!=0x10000 ? 0:MIN(r,48000),
			f->chanCount>1 || (int16)vol!=(vol>>16),&r))==NULL)
	{
		fprintf(stderr,"### unable to open sndManager (%d)\n",
			r);
		goto err;
	}
	if ((ch=sndNewChannel(mgr,sndInitStereo))==NULL)
	{
		fprintf(stderr,"### unable to create sndChannel\n");
	err:	r = 0;
		goto end;
	}
	setChSettings(ch,vol,rmul);
	if ((r=sndPlayFile(in,ch,f,k))<0)
		goto end;
	sndWaitLoop(mgr,ch,vol,rmul);
	r = 1;
	sndPlayFileStop(ch);
end:	sndMgrClose(mgr);
	return r;
}
#endif

static long procSoundList(FILE *s,
	sndFormat *f,
	uint8 *n,
	long z,
	int32 vol,
	ufixed rmul)
{
	ufixed		rateMultiplier = 0x10000;
	long int	starto, datao = 0;
	int		v, c, m;

	if (!s || !f)
		return -1;
	starto = ftell(s);
	v = fgetc(s)<<8;
	if ((v|=fgetc(s))<0)
		return -3;
	if (v-1u>1u)
	{
		fprintf(stderr,"### bad soundlist format (%d)\n",
			v);
		return -206; /* badFormat */
	}
	if (v==1)
	{ /* parse modifiers */
		c = fgetc(s)<<8, v = m = 0;
		if ((c|=fgetc(s))<0)
			return -3;
		if ((c=(c^0x8000)-0x8000)<=0)
		{
			fprintf(stderr,"### bad modifier count (%d)\n",
				c);
			return -206;
		}
		v = fgetc(s)<<8;
		if ((v|=fgetc(s))<0)
			return -3;
		if (v!=5)
		{
			fprintf(stderr,"### snddec only works w. sampled synth (#5, not 0x%X)\n",
				v);
			return -206;
		}
		m = fgetc(s)<<16, m|= fgetc(s)<<8;
		if ((m|=fgetc(s))<0)
			return -3;
		if ((v=fgetc(s))<0)
			return -3;
		m = m<<8|v;
		fseek(s,6*c-6,SEEK_CUR);
	#if 0
		if (!(m&0x80))
		{
			fprintf(stderr,"### no channel init (%d)\n",
				m);
			return -206; /* no channel init */
		}
	#endif
		/* what the heck are srate flags? */
	} else /* refCount */
		fseek(s,2,SEEK_CUR);
	c = fgetc(s)<<8;
	if ((c|=fgetc(s))<0)
		return -3;
	if ((c=(c^0x8000)-0x8000)<=0)
		return c ? -5:-1;
	if (z)
	{
		v = ftell(s)-starto;
		if ((z-=v)<=0)
			return -5;
#ifndef NO_SNDMGR
		return loadSndPlay(s,z,v,c,vol,rmul);
#endif
	}
	/* parse sndCommands */
	do
	{
		m = fgetc(s)<<8;
		if ((m|=fgetc(s))<0)
			return -3;
		switch(m)
		{ /* #15 on = param2 is an absolute offset (off=ptr, invalid here) */
		case 86: /* rateMultiplier - apply afterward */
			fseek(s,2,SEEK_CUR); /* param1 - SM-ignored */
			v = fgetc(s)<<16, v|= fgetc(s)<<8;
			if ((v|=fgetc(s))<0)
				return -3;
			rateMultiplier = v<<8;
			if ((v=fgetc(s))<0)
				return -3;
			rateMultiplier|= v;
			break;
		case 0x8050: /* soundCmd */
		case 0x8051: /* bufferCmd */
			if (datao)
			{
				fprintf(stderr,"## more than one bufferCmd/soundCmd\n");
				goto skip;
			}
			fseek(s,2,SEEK_CUR); /* param1 - ignored */
			v = fgetc(s)<<16, v|= fgetc(s)<<8;
			if ((v|=fgetc(s))<0)
				return -3;
			datao = v<<8;
			if ((v=fgetc(s))<0)
				return -3;
			datao|= v;
		case sndNullCmd:
			break;

		default:
			fprintf(stderr,"## unsupported SndCommand for exporting (#%d)\n",
				m);
		skip:	fseek(s,6,SEEK_CUR);
		}
	} while(--c);

	if (!datao)
		return -5;
	fseek(s,datao+starto,SEEK_SET);
	if ((starto=readBufferCmdHeader(s,f,n))<=0)
		return starto;
	if (rateMultiplier!=0x10000)
	{ /* normally, this and volumeCmd would be resampled through a mixer */
		uint64		rate;

		mulu64u32(rate,f->rate<<16|f->rateFrac>>16,rateMultiplier);
		f->rate = int64hi(rate), f->rateFrac = int64lo(rate);
	}
	return datao+starto;
}

static int readExtHeader(FILE *s,
	sndFormat *f,
	uint8 *n,
	uint8 m)
{
	uint8		b[64];
	int		r;

	if (!s || !f)
		return -1;
	memset(b,0,22);
	fread(b+22,1,sizeof(b)-22,s);
	b[7] = (m&1)+1, b[8] = 0x80, b[20] = 0xFF, b[21] = 0x3C;
	r = parseBufferCmdHeader(b,f,n);
	if (r<=0)
		return r;
	r-= 22;
	/* ROM ext. headers contain a TOTAL 4-byte count, not numFrames */
	/* (at least in the stereo case) */
	if (m&4)
	{
		f->frameCount = b[22]<<24|b[23]<<16|b[24]<<8|b[25];
		if (f->depth<8 || f->frameCount>0x200000u)
			return -1;
		f->frameCount = ((f->frameCount<<2)-42)
			/(((f->depth+7)>>3)*f->chanCount); //&127);
		if (f->frameCount>0x3FFFFFFF)
			return -1;
	}
	return !parseExt80Rate(b+26,f) ? -1:r;
}

static int readDOCSSampleHeader(FILE *s,
	sndFormat *f,
	int32 z)
{
	uint8		b[10];
	long		c;

	if (fread(b,10,1,s)<=0)
		return -3;
	for(c=0,z-=10; z>0 && fgetc(s)>0; --z)
		++c;
	return parseSoundSampleHeader(b,f,c);
}

static int outWAVE(FILE *in,
	sndFormat *f,
	const char *outn,
	int k,
	int8 wr)
{
	FILE		*out;
	const uint8	*h;
	uint8		*buf;
	long		t, z = 0;
	int32		packo = sndMaxFrameSize(f);
	int		r = 0;
	/* allocate the buffers */
	if ((buf=calloc(packo+f->maxRead*f->blockSize,1))==NULL)
	{
		fprintf(stderr,"### failed to allocate conversion buffers\n");
		return 1;
	}
	/* set up the output file */
	h = setupWAVEHeader(f,&z);
	if (!wr)
		out = NULL;
	else if ((out=fopen(outn,"wb"))==NULL)
	{
		fprintf(stderr,"### failed to open '%s' for writing\n",
			outn);
		r = 1;
		goto end;
	} else
	{
		fseek(out,0,SEEK_SET);
		fwrite(h,z,1,out);
	}
	t = 0;
	/* some formats don't have starting samples in their packets */
	/* the universal method for seek would be to just decompress and skip */
	while((z=decompressSndFrame(buf,buf+packo,f,
			readSndFrame(buf+packo,in,f))-buf)>0)
	{
		if (k>0)
			{k-= f->maxRead; continue;}
		z = writeSndSamples(buf,z,out,f,STYPE_WAVE32,
			!out ? sndwrite_null:sndwrite_fwrite);
		if (z<=0)
			{r|= 1; break;}
		t+= z;
	}
	if (out)
	{
		h = setupWAVEHeader(f,&t);
		fseek(out,0,SEEK_SET);
		fwrite(h,t,1,out);
	} else
		printf("\n % 7lu sample bytes accounted for\n\n",
			t);
	if (r)
		fprintf(stderr,"## error(s) occurred during %s (%d)\n",
			out ? "conversion":"verification",r);
	r = 0;
end:	free(buf);
	return r;
}

static void printDuration(const sndFormat *f,
	int userSet)
{
	uint32		dur, mins;

	if ((dur=sndFrameToTime(IMAX(int32),f,0))==0)
		return;
	printf(userSet ? " - for":"\n Duration:");
	if ((mins=dur/60000)!=0)
	{
		uint32		hrs = mins/60;

		dur-= mins*60000, mins-= hrs*60;
		printf(" %d:%02d:%c%g\n",
			hrs,mins,dur/10000+'0',dur%10000/1000.);
	} else
		printf(" %g %s\n",
			dur/1000.,dur==1000 ? "second":"seconds");
}

static void printSndInfo(const sndFormat *f)
{
	static char	*types[] =
	{
		"PCM",
		"NPCM",
		"DPCM",
		"TDPCM",
		"ADPCM",
		"MPEG"
	};

	const char	*n;
	long double	rate = f->rate+f->rateFrac*(1/4294967296.);
	int32		sperpack = f->frameSize/((f->chanCount&127)*f->mSampleSize)*1000;
	uint8		stype = f->type, sdepth = f->mSampleSize*8;

	printf(f->depth!=sdepth ? "    Depth: %d/%d bits\n"
		:"    Depth: %d bits\n",
		f->depth,sdepth);
	if (stype&15)
		printf("    Codec: %s\n",
			(n=sndGetCodecName(f->codec))==NULL ? "???":n);
	printf(
"     Type: %s%s %s\n"
"     Rate: %s%dx%.*f Hz, %.1f kbps\n",
		stype&STYPE_FP ? "floating point":stype&STYPE_ISGN ? "unsigned":"signed",
		stype&15 || sdepth==8 ? "":stype&STYPE_LSB ? " (LE)":" (BE)",types[stype&15],
		f->chanCount&128 ? "r":"",f->chanCount&127,f->rateFrac ? 6:0,rate,
		f->blockSize*rate*8/sperpack);
}

typedef enum
{
	_none,
	_bufC,
	_extM,
	_extS,
	_romM,
	_romS,
	_sndL,
	_sndS
} sndType;

int main(int argc,
	char **argv)
{
	sndFormat	format;
	FILE		*in;
	char const	*n;
	char		**a = argv+1, *s;
	long int	o =-1, k = 0, dur =-1, vol = 0x1000100;
	int32		r = 0, z = IMAX(int32);
	ufixed		rmul = 0x10000;
	sndType		m = _none;
	uint8		play = 1, dinfo = 0, dcvt = 0, dchk = 0, bn = 0;

	if (argc<1)
		return 0;
	while(argc>1 && *(s=*a)=='-')
		if (--argc,++a,!(++s)[1])
			if (*s=='k')
			{
				if (!readanyinteger(*a,strlen(*a),&k) || k<0)
				{
					fprintf(stderr,"### bad seek value '%s'\n",
						*a);
					argc = 0;
					break;
				}
				--argc, ++a;
			} else if (*s=='d')
			{
				if (!readanyinteger(*a,strlen(*a),&dur) || dur<0)
				{
					fprintf(stderr,"### bad duration value '%s'\n",
						*a);
					argc = 0;
					break;
				}
				--argc, ++a;
			} else if (*s=='v')
			{
				long int	rv;

				if ((n=readanyinteger(*a,strlen(*a),&vol))==NULL || vol<0)
					goto badvol;
				rv = vol;
				if (*n++==',' && !readanyinteger(n,strlen(n),&rv))
				{
				badvol:	fprintf(stderr,"### bad volume value '%s'\n",
						*a);
					argc = 0;
					break;
				}
				vol = (vol*64+12)/25, rv = (rv*64+12)/25;
				vol = vol<<16|rv;
				--argc, ++a;
			} else if (*s=='r')
			{
				if ((rmul=atof(*a)*65536+.5)==0)
				{
					fprintf(stderr,"### bad rate value '%s'\n",
						*a);
					argc = 0;
					break;
				}
				--argc, ++a;
			} else if (*s=='p')
				play = 1;
			else if (*s=='i')
				dinfo = 1;
			else if (*s=='z')
				dchk = 1, play = 0;
			else if (*s=='c' || *s=='w')
				dcvt = 1, play = 0;
			else
				goto unk;
		else if (!strcmp(s,"bc"))
			{m = _bufC; goto co;}
		else if (!strcmp(s,"xm"))
			{m = _extM; goto co;}
		else if (!strcmp(s,"xs"))
			{m = _extS; goto co;}
		else if (!strcmp(s,"rm"))
			{m = _romM; goto co;}
		else if (!strcmp(s,"rs"))
			{m = _romS; goto co;}
		else if (!strcmp(s,"sl"))
			{m = _sndL; goto co;}
		else if (!strcmp(s,"ss"))
		{
			m = _sndS;
		co:	if ((n=readanyinteger(*a,strlen(*a),&o))==NULL || o<0)
			{
				fprintf(stderr,"### bad offset value '%s'\n",
					*a);
				argc = 0;
				break;
			}
			if (m>_romS && *n++==',')
			{
				long int	ssz;

				if (!readanyinteger(n,strlen(n),&ssz))
				{
					fprintf(stderr,"### bad size value '%s'\n",
						*a);
					argc = 0;
					break;
				}
				z = ssz;
			} else if (m==_sndL)
				z = 0;
			--argc, ++a;
		} else
		{
		unk:	fprintf(stderr,"### unknown option '%s'\n",
				s-1);
			goto usage;
		}
	if (argc<2)
	{
	usage:	fprintf(stderr,
"# usage: %s [options] infile\n"
"#\n"
"# -bc <offset>   bufferCmd\n"
"# -xm <offset>   extended, mono\n"
"# -xs <offset>   extended, stereo\n"
"# -rm <offset>   ROM extended, mono\n"
"# -rs <offset>   ROM extended, stereo\n"
"# -sl <ofs[,sz]> Mac SoundList - virtual SM with sz+play\n"
"# -ss <ofs[,sz]> DOC rSoundSample\n"
"# -k <ms>        starting millisecond\n"
"# -d <ms>        max. duration (milliseconds)\n"
"# -p             play (default on single path)\n"
"# -i             print info\n"
"# -v             set playback volume\n"
"# -r             set playback rate multiplier\n"
"# -c  |  -w      convert (to WAVE; path follows infile)\n"
"# -z             check for errors (only)\n",
			*argv);
		return 1;
	}
	in = fopen(*a,"rb");
	if (!in)
	{
		fprintf(stderr,"### failed to open '%s' for reading\n",
			*a);
		return 1;
	}
	if (o<0)
		z = readAudioHeader(in,&format);
	else
	{
		fseek(in,o,SEEK_SET);
 		if (m>_bufC)
	 		if (m==_sndL)
			{
				if (!play)
					z = 0;
				if (z && dinfo)
					printf("Parsing SoundList within '%s'...\n",
						*a);
				z = procSoundList(in,&format,&bn,z,vol,rmul);
				if (z==-128)
					goto end;
	 		} else if (m==_sndS)
				z = readDOCSSampleHeader(in,&format,z);
	 		else
				z = readExtHeader(in,&format,&bn,m);
		else
			z = readBufferCmdHeader(in,&format,&bn);
	}
	if (z>0 && k && (k=sndTimeToFrame(k,&format))<0)
		z = k;
	if (z>0 && format.frameCount==-1u)
		z =-1;
	if (z<=0)
		goto err;
	if ((format.chanCount&127)>2)
	{
		fprintf(stderr,"### implementation cannot handle more than two channels\n");
		goto end;
	}
	o = MAX(o,0)+z;
	/* check positive each time in case of overflow */
	if (dur>0 && (dur=sndTimeToFrame(dur,&format))>0)
	{
		dur-= k;
		if (dur<format.frameCount)
			format.frameCount = dur;
	}
	/* for higher PCM seek precision */
	if (k>0 && format.maxRead>1)
	{ /* none of the formats in this case need pre-decomp */
	//	t = k%format.maxRead, o+= t*format.blockSize, k-= t;
		o+= k*format.blockSize;
		if (k>format.frameCount)
			format.frameCount = 0;
		else
			format.frameCount-= k;
		k = 0;
	}
	/* initialize variables, seek */
	decompressSndFrame((uint8*)format.variables,NULL,&format,0);
	fseek(in,o,SEEK_SET);
	if (play)
	{
	#ifndef NO_SNDMGR
		printf(m ? "Playing sound within '%s'...\n":"Playing '%s'...\n",
			*a);
		printDuration(&format,dur>0);
		if (dinfo)
		{
			printSndInfo(&format);
			if (bn)
				printf("Base Note: %u\n",
					bn);
			if (rmul!=0x10000)
				printf("Play Rate: %.10g\n",
					round(rmul*(50000/65536.))/50000);
		}
		z = playFile(in,&format,k,vol,rmul);
	#endif
		if (z<=0)
			if (!z)
				fputs("## failed to set up the sound output\n",stderr);
			else
			err:	fprintf(stderr,!z ? "### unsupported file format\n"
					:z==-1 ? "### bad header format or parameter\n"
					:z==-2 ? "### out of memory\n"
					:z==-3 ? "### read/EOF error\n"
					:z==-5 ? "### no sound\n"
					:z==-8 ? "### unsupported codec\n"
					:"### error(s) occurred parsing sound header (%d)\n",
					z);
		else
	#ifndef NO_SNDMGR
			printf("\nDone playing.\n\n");
	#else
			printf("### Sound playback not implemented for this platform.\n\n");
	#endif
		goto end;
	} else
	{
		printf("Decoding '%s'...\n",
			*a);
		if (dinfo)
		{
			printDuration(&format,dur>0);
			printSndInfo(&format);
		}
	}
	r = dcvt|dchk ? outWAVE(in,&format,a[1],k,dcvt):0;
end:	fclose(in);
	return r;
}
