/*
 snddec.h
  sound decoder package

  Created: 2014.01.13 23:39 EST
  Revised: 2014.11.02

 * You can redistribute this file and/or modify it under the
 * terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.  You should have
 * received a copy of the license along with this file.

 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See LICENSE.txt more details.
 */

#ifndef __SNDDEC_H
 #define __SNDDEC_H
 #include <pshpack2.h>

/* ---- constants */

/* for integer PCM, #0=-.5, as lowest (-32768) is 1 bigger than its positive (32767) */

#define CTYPE_PCM		0 /* linear PCM */
#define CTYPE_NPCM		1 /* non-linear PCM */
#define CTYPE_DPCM		2 /* delta modulated PCM */
#define CTYPE_TDPCM		3 /* table-delta modulated PCM */
#define CTYPE_ADPCM		4 /* Adaptive Differential PCM */
#define CTYPE_MPEG		5

/* these codes found in Interfaces/CIncludes/Sound.h (Mac OS Interfaces 3.4.2) */
/* -- PCM */
#define CODEC_PCM_U8		MK4CC('r','a','w',' ') /* offset-binary: 0=127.5 */
#define CODEC_PCM_DEFAULT	MK4CC('N','O','N','E')
#define CODEC_PCM_BE_S16	MK4CC('t','w','o','s') /* int16 MSB */
#define CODEC_PCM_LE_S16	MK4CC('s','o','w','t') /* int16 LSB */
#define CODEC_PCM_BE_S24	MK4CC('i','n','2','4') /* int24 MSB */
#define CODEC_PCM_BE_S32	MK4CC('i','n','3','2') /* int32 MSB */
#define CODEC_PCM_LE_S32	MK4CC('2','3','n','i') /* int32 LSB */
#define CODEC_PCM_BE_F32	MK4CC('f','l','3','2') /* float MSB */
#define CODEC_PCM_BE_F64	MK4CC('f','l','6','4') /* double MSB */
/* -- non-linear PCM */
#define CODEC_ALAW		MK4CC('a','l','a','w') /* CCITT 12-bit alaw 2:1 */
#define CODEC_ULAW		MK4CC('u','l','a','w') /* CCITT 12-bit [micro]law 2:1 */
/* -- table DPCM */
/* Macintosh Audio Compression/Expansion */
#define CODEC_MACE_3		MK4CC('M','A','C','3') /* 9-bit/pair MACE 3:1 */
#define CODEC_MACE_6		MK4CC('M','A','C','6') /* 10-bit MACE 6:1 */
/* Apple IIgs Audio Compression/Expansion */
#define CODEC_ACE_2		MK4CC('A','C','E','2') /* ACE 2:1 */
#define CODEC_ACE_8		MK4CC('A','C','E','8') /* ACE 8:3 */
/* -- ADPCM */
#define CODEC_AIFC_IMA		MK4CC('i','m','a','4') /* 16-bit AIFC IMA 4:1 (64:17) */
/* the CD/XA codecs are h/w decompressed using the Enhanced Apple Sound Chip (EASC) */
#define CODEC_CDXA_4		MK4CC('c','d','x','4') /* 16-bit BRR CD/XA 4:1 (56:15) */
#define CODEC_CDXA_2		MK4CC('c','d','x','2') /* 16-bit BRR CD/XA 2:1 */

#define CODEC_ADP4		MK4CC('A','D','P','4')
/* 4XM is LE IMA but with all-channels-in-one packets (pred.w[ch],ix.w[ch]); */
/* same w. DK4, but header is (samp.w,ix.w)[ch] */
/* nybbles are swapped for ISS (pred.w,ix.w)[ch] */
/* EA EACS is 32-bit (ix.l[ch],pred.l[ch]):ads=6, and APC has no header */
/* -- MPEG */
#define CODEC_MPEG_L3P		MK4CC('.','m','p','3') /* MPEG CBR/VBR - QT 4.1+ */

/* other */
#define CODEC_DVCA		MK4CC('d','v','c','a') /* DV Audio */
#define CODEC_QDESIGN		MK4CC('Q','D','M','C') /* QDesign music */
#define CODEC_QDESIGN2		MK4CC('Q','D','M','2') /* QDesign2 music */
#define CODEC_QUALCOMM		MK4CC('Q','c','l','p') /* QUALCOMM PureVoice */
/* newer */
#define CODEC_GSM		MK4CC('G','S','M',' ')
#define CODEC_G726_ADPCM	MK4CC('G','7','2','6')
#define CODEC_DWVW		MK4CC('D','W','V','W')

/* unofficial little endian ordering types */
#define CODEC_PCM_LE_F32	MK4CC('2','3','l','f') /* float LSB */
#define CODEC_PCM_LE_F64	MK4CC('4','6','l','f') /* double LSB */

/* Microsoft WAVE Format types - ID 0=unknown */
/* most of these are here merely for completeness */
#define CODEC_MS_PCM		MK4CC('m','s',0,1)
#define CODEC_MS_ADPCM		MK4CC('m','s',0,2)	/* 16-bit Microsoft ADPCM 4:1 (253:64) */
#define CODEC_MS_FLOAT		MK4CC('m','s',0,3)	/* 32/64-bit floating point PCM */
#define CODEC_MS_VSELP		MK4CC('m','s',0,4)	/* Compaq Computer Corporation */
#define CODEC_IBM_CVSD		MK4CC('m','s',0,5)	/* IBM Corporation */
#define CODEC_MS_ALAW		MK4CC('m','s',0,6)
#define CODEC_MS_ULAW		MK4CC('m','s',0,7)
#define CODEC_OKI_ADPCM		MK4CC('m','s',0,16)	/* 12-bit OKI */
/* the sample rate must be divisible by two for a workable byte-rate field */
#define CODEC_DVI_IMA		MK4CC('m','s',0,17)	/* 16-bit DVI IMA 4:1 (255:64) */
#define CODEC_MEDIASPACE	MK4CC('m','s',0,18)	/* Videologic MediaSpace ADPCM */
#define CODEC_SIERRA_ADPCM	MK4CC('m','s',0,19)	/* Sierra Semiconductor Corp */
#define CODEC_G723_ADPCM	MK4CC('m','s',0,20)	/* Antex Electronics Corp. */
#define CODEC_DSP_DIGISTD	MK4CC('m','s',0,21)	/* DSP Solutions, Inc. */
#define CODEC_DSP_DIGIFIX	MK4CC('m','s',0,22)	/* DSP Solutions, Inc. */
#define CODEC_DIALOGIC_OKI	MK4CC('m','s',0,23)	/* Dialogic Corporation */
#define CODEC_MEDIAVISION	MK4CC('m','s',0,24)	/* Media Vision, Inc. */
#define CODEC_CU_HP		MK4CC('m','s',0,25)	/* CU codec - Hewlett-Packard Company */
#define CODEC_YAMAHA_ADPCM	MK4CC('m','s',0,32)	/* Yamaha Corporation of America */
#define CODEC_SONARC		MK4CC('m','s',0,33)	/* Speech Compression */
#define CODEC_TRUESPEECH	MK4CC('m','s',0,34)	/* DSP Group TrueSpeech[TM] */
#define CODEC_ECHOSC1		MK4CC('m','s',0,35)	/* Echo Speech Corp. */
#define CODEC_AUDIOFILE_AF36	MK4CC('m','s',0,36)	/* Audiofile, Inc. */
#define CODEC_APTX		MK4CC('m','s',0,37)	/* Audio Processing Technology */
#define CODEC_AUDIOFILE_AF10	MK4CC('m','s',0,38)	/* Audiofile, Inc. */
#define CODEC_PROSODY_1612	MK4CC('m','s',0,39)	/* Aculab plc */
#define CODEC_LRC		MK4CC('m','s',0,40)	/* Merging Technologies S.A. */
#define CODEC_DOLBY_AC2		MK4CC('m','s',0,48)	/* Dolby Laboratories */
#define CODEC_MS_GSM_610	MK4CC('m','s',0,49)	/* GSM 6.10 - Microsoft Corp */
#define CODEC_MSNAUDIO		MK4CC('m','s',0,50)	/* Microsoft Corp. */
#define CODEC_ANTEX_ADPCME	MK4CC('m','s',0,51)	/* Antex Electronics Corp. */
#define CODEC_CONTROLRES_VQLPC	MK4CC('m','s',0,52)	/* Control Resources Ltd. */
#define CODEC_DIGIREAL		MK4CC('m','s',0,53)	/* DSP Solutions, Inc. */
#define CODEC_DIGIADPCM		MK4CC('m','s',0,54)	/* DSP Solutions, Inc. */
#define CODEC_CONTROLRES_CR10	MK4CC('m','s',0,55)	/* Control Resources Ltd. */
#define CODEC_NMS_VBXADPCM	MK4CC('m','s',0,56)	/* Natural MicroSystems */
#define CODEC_ROLAND_RDAC	MK4CC('m','s',0,57)
#define CODEC_ECHOSC3		MK4CC('m','s',0,58)	/* Echo Speech Corp. */
#define CODEC_ROCKWELL_ADPCM	MK4CC('m','s',0,59)	/* Rockwell International */
#define CODEC_ROCKWELL_DIGITALK	MK4CC('m','s',0,60)	/* Rockwell International */
#define CODEC_XEBEC		MK4CC('m','s',0,61)	/* Xebec Multimedia Solutions Ltd */
#define CODEC_G721_ADPCM	MK4CC('m','s',0,64)	/* Antex Electronics Corp. */
#define CODEC_G728_CELP		MK4CC('m','s',0,65)	/* Antex Electronics Corp. */
#define CODEC_G723_1		MK4CC('m','s',0,66)	/* Microsoft G.723.1 */
#define CODEC_MS_MPEG		MK4CC('m','s',0,80)	/* Microsoft MPEG */
#define CODEC_RT24		MK4CC('m','s',0,82)	/* InSoft Inc. */
#define CODEC_PAC		MK4CC('m','s',0,83)	/* InSoft Inc. */
#define CODEC_MPEG_L3		MK4CC('m','s',0,85)	/* MPEG - CBR-only, pre-QT 4.1 */
#define CODEC_LUCENT_G723	MK4CC('m','s',0,89)	/* Lucent Technologies */
#define CODEC_CIRRUS		MK4CC('m','s',0,96)	/* Cirrus Logic */
#define CODEC_ESPCM		MK4CC('m','s',0,97)	/* ESS Technology */
#define CODEC_MS_G726_ADPCM	MK4CC('m','s',0,100)	/* APICOM */
#define CODEC_MS_G722_ADPCM	MK4CC('m','s',0,101)	/* APICOM */
#define CODEC_MS_DSAT		MK4CC('m','s',0,102)	/* Microsoft Corp. */
#define CODEC_DSAT_DISPLAY	MK4CC('m','s',0,103)	/* Microsoft Corp. */
/* Voxware Inc. */
#define CODEC_VOXWARE_ALIGNED	MK4CC('m','s',0,105)	
#define CODEC_VOXWARE_AC8	MK4CC('m','s',0,112)
#define CODEC_VOXWARE_AC10	MK4CC('m','s',0,113)
#define CODEC_VOXWARE_AC16	MK4CC('m','s',0,114)
#define CODEC_VOXWARE_AC20	MK4CC('m','s',0,115)
#define CODEC_VOXWARE_RT24	MK4CC('m','s',0,116)
#define CODEC_VOXWARE_RT29	MK4CC('m','s',0,117)	/* also, Meta-Sound */
#define CODEC_VOXWARE_RT29HW	MK4CC('m','s',0,118)
#define CODEC_VOXWARE_VR12	MK4CC('m','s',0,119)
#define CODEC_VOXWARE_VR18	MK4CC('m','s',0,120)
#define CODEC_VOXWARE_TQ40	MK4CC('m','s',0,121)

#define CODEC_SOFTSOUND		MK4CC('m','s',0,128)	/* Softsound, Ltd. */
#define CODEC_VOXWARE_TQ60	MK4CC('m','s',0,129)	/* Voxware Inc. */
#define CODEC_MS_RT24		MK4CC('m','s',0,130)	/* Microsoft Corp. */
#define CODEC_G729A		MK4CC('m','s',0,131)	/* AT&T Labs, Inc. */
#define CODEC_DF_GSM610		MK4CC('m','s',0,134)	/* DataFusion Systems (Pty) (Ltd) */
#define CODEC_SBC24		MK4CC('m','s',0,145)	/* Siemens Business Communications Systems */
#define CODEC_ZYXEL_ADPCM	MK4CC('m','s',0,151)	/* ZyXEL Communications, Inc. */

#define CODEC_AAC_FF		MK4CC('m','s',0,255)
#define CODEC_RHETOREX_ADPCM	MK4CC('m','s',1,0)	/* 256 - Rhetorex, Inc. */
#define CODEC_ACELP_NET		MK4CC('m','s',1,0x30)	/* 304: Sipro Lab Telecom ACELP.net */
#define CODEC_WMA1		MK4CC('m','s',1,0x60)	/* 352: Windows Media Audio v1 */
#define CODEC_WMA2		MK4CC('m','s',1,0x61)	/* 353: Windows Media Audio v2 */
#define CODEC_WMAPRO		MK4CC('m','s',1,0x62)	/* 354 */
#define CODEC_WMALOSSLESS	MK4CC('m','s',1,0x63)	/* 355 */
#define CODEC_ADPCM_CT		MK4CC('m','s',2,0)	/* 512 */
#define CODEC_ATRAC3		MK4CC('m','s',2,112)	/* 624 */
#define CODEC_ADPCM_G722	MK4CC('m','s',2,143)	/* 655 - unofficial...? */
#define CODEC_IMC		MK4CC('m','s',4,1)	/* 1025 */
#define CODEC_IAC_2		MK4CC('m','s',4,2)	/* 1026: Indeo(R) v2-32 */
#define CODEC_VME_VMPCM		MK4CC('m','s',6,0x80)	/* 1664 - AT&T Labs, Inc. */
#define CODEC_LH		MK4CC('m','s',17,0)	/* 4352: Lernout & Hauspie */
#define CODEC_GSM_MS		MK4CC('m','s',21,0)
#define CODEC_TRUESPEECH_X	MK4CC('m','s',21,1)	/* 5377 - another TrueSpeech ID */
#define CODEC_AAC_1600		MK4CC('m','s',22,0)	/* 5632 - ADTS AAC */
#define CODEC_AAC_LATM		MK4CC('m','s',22,2)	/* 5634 */
#define CODEC_DVM		MK4CC('m','s',32,0)	/* 8192 - FAST Multimedia AG, AC3 */
#define CODEC_DTS		MK4CC('m','s',32,1)	/* 8193 */
#define CODEC_AAC_WAV		MK4CC('m','s','A','C')
#define CODEC_PCM_ULAW		MK4CC('m','s','l','u')
#define CODEC_AAC_PM		MK4CC('m','s','p','m')
#define CODEC_SPEEX		MK4CC('m','s',161,9)
#define CODEC_FLAC		MK4CC('m','s',0xF1,0xAC)

#define STYPE_ISGN		0x80 /* in: invert sign (U8, U16, U32); out: U8 for S8 */
#define STYPE_LSB		0x40 /* native or stored little endian */
#define STYPE_FP		0x20 /* floating point format (float/double) */
#define STYPE_MAX16		0x10 /* maximum storage depth of 16 bits (i.e., WAVE) */
/* values to pass as an stype parameter */
/* ISGN in this case means u8 for 8-bit */
#define STYPE_STDMAX16		(STYPE_ISGN|STYPE_LSB|STYPE_MAX16)
#define STYPE_WAVE32		STYPE_STDMAX16
#define STYPE_WAVE64		(STYPE_ISGN|STYPE_LSB|STYPE_FP)

#define STYPE_ENDIAN_NATIVE	(!BIG_ENDIAN<<6)

/* sound command numbers */

enum
{ /* most of these are in Mac OS order, and some are not compatible */
	sndNullCmd,
	sndAllocCmd,			/* allocates channel's unpacking buffer */
	sndFreeCmd,			/* frees channel's unpacking buffer (silence till init) */
	sndQuietCmd,			/* ends current sound */
	sndFlushCmd,			/* clears everything, including wait time */
	sndReInitCmd,			/* param2=initFlag */
	sndFormatCmd,			/* load a raw sndFormat */
	sndRawCmd,			/* new samples/current fmt.; param1=frameCount (uint16) */
	sndBufferCmd,			/* Mac OS MSB bufferCmd - samplePtr not supported */
	sndRateMultiplierCmd,		/* param2=fixed-point multiplier; param1=pre-fraction */
	sndWaitCmd,			/* param2=half-millisecond delay */
	sndPauseCmd,			/* stop playback until resume command */
	sndResumeCmd,
	sndCallBackCmd,			/* calls param2 as sndCallBack (NULL ignored here) */
	sndSyncCmd,			/* wait until next out-block (helps sync channels) */
	sndLRVolumeCmd,
	sndRLVolumeCmd
	/* getVolumeCmd: ->LRVolume */
	/* getRateMultiplierCmd: ->sndFrame.rateMul */
};

/* new channel/reinit channel and sndMgrMute flags */
#define sndEnableLeft			1
#define sndEnableRight			2
#define sndInitPanMask			3	/* volume muting of channel(s) */
#define sndInitNoInterpolation		4	/* no linear interpolation in rate conversion */
#define sndInitNoNearDropSample		8	/* disable near-1:1 rate conversion */
#define sndInitStereo			0x40	/* otherwise, mono */

/* ---- typedefs */

typedef struct sndFormat/*(72)*/
{
	uint32		codec;		/* OSType */
	uint32		rate, rateFrac;	/* sample rate and fraction */
	uint8		chanCount;	/* channel count - bit #7 = reverse ordering */
	uint8		type;		/* [isgn.1] [LSB.1] [fp.1] [max16.1] [cmptype.4] */
	uint8		depth;		/* codec precision in bits */
	uint8		mSampleSize;	/* unpacked single-channel byte size (1, 2, 4 or 8) */
/*0x10*/uint32		frameCount;	/* number of uncompressed block frames */
	int32		frameSize;	/* size of one multi-channel frame in native PCM */
	uint16		packSize;	/* compressed size for one channel in bytes */
	uint16		blockSize;	/* packSize*chanCount, w. any necessary alignment */
	/* the user can set this field after the codec has been initialized */
	uint32		maxRead;	/* maximum block read count per readSndFrame */
	uint64		totalCount;	/* total number of decompressed samples */
	/* for additional channels, allocate 16 bytes for each extra channel */
/*0x20*/int16		variables[16];	/* interframe data; MACE needs 6/chan; CD/XA 8/chan */
} sndFormat;

typedef struct adpcm_base/*(16)*/
{
	int8		coef[10];	/* applicable coefficients */
	uint8		kx, mx;		/* applicable pre-sample mask/index, maximum */
	int16		l0, l1;		/* out-sample and modifier */
} adpcm_base;

#define DEF_MIX_RTN(n)		void n(int32*d, const void*p, sndMixerFrame*m, int c)
#define TYP_MIX_RTN(n)		void(*n)(int32*d, const void*p, struct sndMixerFrame*, int c)

typedef struct sndMixerFrame
{
	TYP_MIX_RTN(mixerRtn); /* loaded automatically - set to NULL */

	uint8		*cmpSamples; /* advances on each read */
	uint8		*dcmpBuf; /* calloc allocated - must be >sndEndFrameCount*2 */
	int32		dcmpSize; /* byte count at dcmpBuf */
	sndFormat	format;
	ufixed		rateMul;
	uint32		LRAmp; /* ifix16 left/right amplitudes */
	/* these are automatically calculated */
	uint32		pos; /* byte position in dcmpBuf */
	ufixed		cnvRate; /* src/dst */
	uint16		cnvFrc; /* pos. fraction; start at d/s, zero, or 0.5 if s>=2*d */
} sndMixerFrame;

/* unlike the SndCommand structure, param2 is a ptr first */
/* in other words, it can work on 64-bit systems */
typedef struct sndCommand/*(10+)*/
{
	uint16		cmd;
	uint32		param1; /* expanded for 32-bit raw frameCounts */
	void		*param2;
} sndCommand;

typedef struct sndChannel
{
	struct sndChannel	*next;

	void		*private;
	void		*userInfo;
	uint32		waitCnt;
	uint32		LRVolume;
	sndMixerFrame	sndFrame;
	uint8		actFlag, initFlag;
	int16		qLength, qHead, qTail;
	sndCommand	queue[128];
} sndChannel;

typedef struct sndManager
{
	int8	private[1];
} sndManager;

typedef uint32 (*sndwfunc_t)(const void *p,
	uint32 byteCount,
	void *stream);

typedef void(*sndCallBack)(sndChannel*,
	sndCommand*);

TYP_MIX_RTN(mixerRoutine);

/* ---- globals */

extern const adpcm_base	brr_default; /* PSX values */
extern const int16	Alaw16btable[256], ulaw16btable[256];

/* ---- prototypes */

/*
 sndMaxFrameSize
	Computes the maximum size for one multi-channel sample frame.
	This is meant for the allocating of space for readSndFrame calls.
  Result:
	The byte size; zero on error.
*/
int sndMaxFrameSize(const sndFormat *f);

/*
 decompressSndFrame
	Decompresses memory-stored data into nativized PCM.
	This also decrements the frame count by maxFrame.
	When out of source data, the destination does not advance.
	Pass NULL for the source to initialize the codec's variables.
  Result:
	The destination end pointer; NULL on error.
*/
uint8 *decompressSndFrame(uint8 *d,
	const uint8 *s,
	sndFormat *f,
	int inByteSize);
/*
 sndFixCmpType
	Updates the sound format to fit the passed codec OSType.
  Result:
	Zero on success; -8 if the codec is not supported.
*/
int sndFixCmpType(uint32 codec,
	sndFormat *f);
/*
 sndFixCodecFrame
	This is used to finish sound description parameters made from scratch.
	Used internally, you'll probably never need to use this.
*/
void sndFixCodecFrame(sndFormat *f);

/*
 sndGetCodecName
	Returns the name for a codec type.
	If it fails to identify the type, NULL is returned.
*/
const char *sndGetCodecName(int32 type);

/*
 parseBufferCmdHeader
	Parses a Mac OS buffer command header: standard, extended or compressed.
  Result:
	Zero or an error bitmask, with bit #3 or higher being more serious.
*/
int32 parseBufferCmdHeader(const uint8 *s,
	sndFormat *f,
	uint8 *baseNote); /* MIDI note for middle C (default: 60); can be NULL */
/*
 parseSoundSampleHeader
	Parses an Apple IIgs rSoundSample header.
  NOTE:
	The user must pass the length, as it's not stored in the header.
*/
int parseSoundSampleHeader(const uint8*,
	sndFormat*,
	int32 length);
/*
 parseExt80Rate
	Parses an "Extended" 80-bit floating point sample rate.
	The value is stored in the codec frame's 64-bit fixed-point representation.
  Result:
	1 if valid; 0 if out of range; undefined if NaN.
*/
int parseExt80Rate(const uint8*,
	sndFormat*);
/*
 sndTimeToFrame
	Converts a millisecond time to a frame count, rounding.
  Result:
	The frame count; negative on error.
*/
int32 sndTimeToFrame(uint32 ms,
	const sndFormat *f);
/*
 sndFrameToTime
	Converts a frame index to milliseconds time.
	The index is clipped to the remaining count.
*/
uint32 sndFrameToTime(uint32 x,
	const sndFormat *f,
	int round); /* treat rate rounded to nearest whole */

/* reading/writing */

/*
 readAudioHeader
	Detects and reads a supported audio format.
	The full sample rate & fraction is rendered.
  Result:
	The offset to the first sample in the file; non-positive on error.
*/
long int readAudioHeader(struct FILE *s,
	sndFormat *f);

/*
 readSndFrame
	Performs a block read for a sound format, given maxRead.
  Result:
	Byte size read to be passed to decompressSndFrame; otherwise, an error.
*/
int readSndFrame(uint8 *d,
	struct FILE*,
	const sndFormat *f);

/*
 writeSndSamples
	Writes PCM samples to an out-stream.
	It also performs any necessary conversions to fit the stype parameter.
  Result:
	The total out-length; the sum of what's returned by the write function.
*/
long writeSndSamples(const void *p,
	long byteSize,
	void *outStream,
	const sndFormat *f,
	uint8 stype, /* use the STYPE constants above */
	sndwfunc_t wf);

/*
 setupWAVEHeader
	Sets up a 32-bit WAVE header according to the codec type & passed size.
	This should be called before writing samples and after finishing.
	Because WAVE rates cannot have a fraction, the encoded rate is rounded.
  Result:
	The header (static global-allocated; do not dispose); NULL on error.
	'size' will be set to the header size.
*/
const uint8 *setupWAVEHeader(const sndFormat *f,
	long *size); /* can be NULL */

/* an assortment of write fuctions */

uint32 sndwrite_fwrite(const void *p,
	uint32 z,
	void *stream); /* FILE* */
uint32 sndwrite_memcpy(const void *p,
	uint32 z,
	void *stream); /* uint8** - advanced in position; pass something of &dst */
uint32 sndwrite_null(const void *p,
	uint32 z,
	void *stream); /* ignored */

/* ---- sound mixer */

/*
 sndMixSamples
	Adds stereo samples, performing decompression if necessary.
	Interpolation is also performed if enabled in rflag.
  Result:
	The remaining count.
*/
int sndMixSamples(int32 *d, /* must be valid */
	sndMixerFrame*, /* must be valid */
	uint32 rflag, /* upper 8-bits=init-flags, low 24=out-rate */
	int dcmpMinCount, /* buffer must be adequate for this count */
	int outMaxCount);
/*
 sndMixFill
	Adds a stereo repeat of the last sample.
	The last is the current (if pos!=0), or final in dcmpBuf, if it exists.
*/
void __fastcall sndMixFill(int c,
	uint32 rflag, /* upper 8-bits=init-flags */
	int32 *d, /* must be valid */
	const sndMixerFrame*); /* must be valid */
/*
 s32to16
	Performs an in-place clipping, of 32 to 16 bits.
	If rflag indicates mono, every other entry is skipped.
*/
void __fastcall s32to16(int32 c,
	uint32 rflag, /* upper 8-bits=init-flags */
	int32 *b); /* must be valid */

/* ---- sound manager */

/*
 sndNewChannel
	Creates and inserts a new channel into a manager's list.
	Use the init-flags constants above; will be mono without sndInitStereo.
	Channels are initially enabled; use sndReInitCmd to disable left/right.
*/
sndChannel *sndNewChannel(sndManager*,
	uint8 initFlags);
/*
 sndDisposeChannel
	Disposes a sound channel, removing it from its manager's list.
  Result:
	Returns zero on success; negative on error.
*/
int sndDisposeChannel(sndChannel*);
/*
 sndChannelBusy
	Returns a positive value if a sound is playing or the queue is active.
*/
int sndChannelBusy(sndChannel*);
/*
 sndDoCommand
	In the immediate case, the command is inserted & processed from the start.
	In the delayed case, the command is merely inserted at the queue end.
  Result:
	Zero on success; a 'full' error code if the queue is full; -1 if qLength<=0.
*/
int sndDoCommand(sndChannel *ch,
	const sndCommand *cmd,
	int8 immediate);

/*
 sndPlayFile
	Plays a file, given the parameters.
	Use fopen and readAudioHeader for 'inStream' and 'format'.
	Use sndPauseCmd and sndResumeCmd for pause/play.
	Passing the binary complement (~) for frameSeek starts play paused.
  Notes:
	This routine uses the userInfo field; it must be NULL prior to calling.
	It completely takes over the channel, and will flush the queue.
  Result:
	Positive on success; zero if seek surpassed end; otherwise, error.
*/
int sndPlayFile(struct FILE *inStream,
	sndChannel *ch,
	const sndFormat *format,
	int32 frameSeek);
/*
 sndPlayFileStop
	Stops and disposes of the frame set up by sndPlayFile.
  Notes:
	This routine flushes the command queue to ensure stop.
	As sndPlayFile does not open a file, neither does this close.
*/
int sndPlayFileStop(sndChannel *ch);

/*
 sndMgrOpen
	Opens a virtual sound manager, given a sample rate.
	If the rate is non-positive, the default rate is used.
	The out-rate is pinned to the range [1000,100000].
  Result:
	A system-compatible output (e.g., wave mapper); NULL on error.
*/
sndManager *sndMgrOpen(int outRate,
	int stereo,
	int *err); /* can be NULL */
/*
 sndMgrClose
	Waits until last packet is finished, frees everything.
	All of its channels are disposed.
  Result:
	The number of elements disposed; negative on error.
*/
int sndMgrClose(sndManager*);
/*
 sndMgrMute
	Sets a manger's mute flags, updating its channels in accordance.
	Use the sndEnableLeft/Right constants above.  (Zero for full mute.)
*/
int sndMgrMute(sndManager*,
	uint8 LREnable);
/*
 sndMgrWait
	If a manager is active, wait one sound packet.
	This is useful for waiting until all sound has completed.
*/
int sndMgrWait(sndManager*);

/*
 sndMgrProcWait
	Use this to ensure that the manager gets cycles.
  Result:
	Zero when finished; 1 when full; otherwise, 2.
*/
void sndMgrProcWait(sndManager*);
/*
 sndMgrMixRate
	Sets the mixing rate, detuning the results & timing.
	Passed rate must be positive.
  Result:
	Old rate if positive; negative on error.
*/
int sndMgrMixRate(sndManager*,
	int rate);

#ifdef WINUSERAPI
HWAVEOUT sndMgrGetHWO(sndManager*);
#endif

/* ---- individual codecs */

/* either Sony (mx>8) or SNES(SPC-700) (mx<=8) */
uint8 *brr_decode(uint8 *d,
	const uint8 *s,
	int chanCount,
	int sampleByteSize, /* not counting the shift/filter byte */
	adpcm_base *f);

/* non-multiplier (lossy conditional downshifts, but faster) */
uint8 *imanm_decode(uint8 *d,
	const uint8 *s,
	int chanCount,
	int sampleByteSize, /* not counting the shift/filter byte */
	adpcm_base *f);
uint8 *ima_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f);
uint8 *oki_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f);
uint8 *ms_adpcm_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f); /* adpcm_base[2] */
uint8 *adpcm_yamaha_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	adpcm_base *f); /* adpcm_base[2] */

uint8 *MACE3_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	int16 *f); /* int16[2] */
uint8 *MACE6_decode(uint8 *d,
	const uint8 *s,
	int m,
	int z,
	int16 *f); /* int16[6] */
 #include <poppack.h>
#endif
