/*
 isc.c
  VM prototype
*/

#include <enviro64.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum
{ /* addressing modes */
	_m_sp,	/* 0-[-S] */
	_m_ip,	/* 0-<implied> */
	_m_ia,	/* 0-A-modifier */
	_m_gw,	/* 4-<global:i32> */
	_m_fb,	/* 1-i8[F] */
	_m_fw,	/* 4-i32[F] */
	_m_rb,	/* 1-*+i8 */
	_m_rw,	/* 4-*+i32 */
	_m_ib,	/* 1-i8 */
	_m_iw,	/* 4-i32 */
	_m_if,	/* 4-d32 */
	_m_id	/* 8-d64 */
};

enum
{ /* operation indices */
	_o_ret,	/* >;	return */
	_o_cst,	/* #	cast type */
	_o_alc,	/* )	allocate (count) */
	_o_del,	/* )	delete (count) */
	_o_pll,	/* )	pull-list (count) */
	_o_call,/* x()	call function */
	_o_bra,	/* :>	branch (goto) */
	_o_bnz,	/* <?	branch if not zero */
	_o_brz,	/* ?>	branch if zero */
	_o_apd,	/* ++	append */
	_o_stv,	/* ==>	store */
	_o_stx,	/* ==>[ store index */
	_o_ldx,	/* x[y] load index */
	_o_and,	/* &	bit-and */
	_o_andt,/* &=>	and-to - mode 0=and-self=int (integerize) */
	_o_ior,	/* |	inclusive or */
	_o_iort,/* |=>	ior-to - mode 0=or-self=uint (integerize) */
	_o_eor,	/* ^x	exclusive or */
	_o_eort,/* ^=>  eor to - mode 0=eor-self=ldz (load zero) */
	_o_shl,	/* <<	shift left */
	_o_shlt,/* <<=> shl-to */
	_o_shr,	/* >>	shift right */
	_o_shrt,/* >>=> shr-to */
	_o_add,	/* +	add */
	_o_addt,/* +=>	add-to - mode 0=add-self=double (*2) */
	_o_sub,	/* -	subtract */
	_o_subf,/* -=>	sub from - mode 0=sub-self=unreal (NaN if INF/NaN) */
	_o_mul,	/* *	multiply */
	_o_mult,/* *=>	mul into - mode 0=mul-self=square (**2) */
	_o_div,	/* /	divide by */
	_o_divt,/* /=>	div into */
	_o_mod,	/* %	modulus (remainder) */
	_o_modt,/* %=>	mod to - mode 0=reserved */
	_o_pow,	/* **	raise to power */
	_o_clt,	/* <	compare less */
	_o_cge,	/* >=	.. greater or equal */
	_o_cle,	/* <=	.. less or equal */
	_o_cgt,	/* >	.. greater */
	_o_ceq,	/* =	.. equal */
	_o_cne,	/* !=	.. inequal */
	_o_cid,	/* ==	.. identical */
	_o_cni,	/* ~=	.. not identical */
	/* monadic */
	_o_phv,	/* ,	push */
	_o_ldv,	/* x	load - mode 0=ldv-self=nop (no operation) */
	_o_abs,	/* +	absolute value */
	_o_neg,	/* -	negate */
	_o_not,	/* ~	bit-not */
	_o_bno,	/* !	boolean-not */
	_o_dno,	/* !!	double boolean-not */
	_o_srt,	/* ~/	square root */
	_o_res	/*	reserved */
};

#define OPS(a)		(_o_##a|_m_sp<<8)
#define OPF(a)		(_o_##a|_m_gw<<8),(_o_##a|_m_fb<<8),(_o_##a|_m_fw<<8)
#define OPi(a)		(_o_##a|_m_ib<<8)
#define OPw(a)		(_o_##a|_m_iw<<8)
#define OPD(a)		(_o_##a|_m_if<<8),(_o_##a|_m_id<<8)
#define OPM(a)		(_o_##a|_m_ip<<8)
#define OPA(a)		(_o_##a|_m_ia<<8)
#define OPR(a)		(_o_##a|_m_rb<<8),(_o_##a|_m_rw<<8)

#define OPI(a)		OPi(a),OPw(a)
#define OPe(a)		OPS(a),OPF(a)
#define OPE(a)		OPA(a),OPF(a)
#define OPl(a)		OPe(a),OPI(a)
#define OPv(a)		OPI(a),OPD(a)
#define OPV(a)		OPe(a),OPv(a)

uint16		ins_ops[256] =
{
/*x00*/	OPe(phv),OPM(ret),OPi(cst),OPI(alc),	OPe(call),OPI(pll),OPI(del),
/*x10*/	OPl(and),OPR(brz),	OPE(andt),OPE(bno),
/*x20*/	OPl(ior),OPR(bnz),	OPE(iort),OPE(dno),
/*x30*/	OPl(eor),OPR(bra),	OPE(eort),OPE(not),
/*x40*/	OPl(shl),OPI(stx),	OPE(shlt),OPE(stx),
/*x50*/	OPl(shr),OPI(ldx),	OPE(shrt),OPE(ldx),
/*x60*/	OPV(add),	OPE(addt),OPE(abs),
/*x70*/	OPV(sub),	OPE(subf),OPE(neg),
/*x80*/	OPV(mul),	OPE(mult),OPE(pow),
/*x90*/	OPV(div),	OPE(divt),OPE(srt),
/*xA0*/	OPV(mod),	OPE(modt),OPv(pow),
/*xB0*/	OPV(ldv),	OPe(stv), OPe(apd),
/*xC0*/	OPV(clt),	OPV(cge),
/*xD0*/	OPV(cle),	OPV(cgt),
/*xE0*/	OPV(ceq),	OPV(cne),
/*xF0*/	OPV(cid),	OPV(cni)
};

typedef double		evalvalue;

typedef struct isc_state
{
	evalvalue	a;
} isc_state;

int isc_execute(const int8 *p,
	isc_state *f);
int isc_execute(const int8 *p,
	isc_state *f)
{
	const int8	*s = p;

	for(;;)
	{
		evalvalue	x;
		uint16		o = ins_ops[*s++];

		switch(o>>8)
		{
		case _m_ib: x = (int8)*s++; break;
		case _m_iw: x = (int32)_be32read(s); break;
		case _m_if:
			((int32*)&x)[BIG_ENDIAN] = 0;
			((int32*)&x)[!BIG_ENDIAN] = _be32read(s);
			break;
		case _m_id:
			((int32*)&x)[!BIG_ENDIAN] = _be32read(s);
			((int32*)&x)[BIG_ENDIAN] = _be32read(s);
			break;
		case _m_ia: x = f->a;
		case _m_ip: break;
		case _m_fb:
		case _m_fw:
		case _m_gw:
		case _m_sp:

		default: fprintf(stderr,"### isc_execute - illegal instruction @ %X\n",
				s-p);
			return 1;
		}
	}
}

int main(int argc,
	char **argv)
{
	FILE		*f;
	long int	z;
	int8		*a;
	isc_state	st;

	if (argc<=1)
		return 0;
	f = fopen(argv[1],"r");
	if (!f)
	{
		fprintf(stderr,"### %s - could not open '%s'\n",
			*argv,argv[1]);
		return 1;
	}
	fseek(f,0,SEEK_END);
	z = ftell(f);
	if (z<=0 || z>0x7FFFFFF0u)
	{
		fprintf(stderr,"### %s - bad length for '%s'\n",
			*argv,argv[1]);
		return 1;
	}
	a = malloc(z);
	if (!a)
	{
		fprintf(stderr,"### %s - out of memory\n",
			*argv);
		return 1;
	}
	fseek(f,0,SEEK_SET);
	if (fread(a,z,1,f)!=1)
	{
		fprintf(stderr,"### %s - read failed\n",
			*argv);
		z = 1;
		goto end;
	}
	memset(&st,0,sizeof(st));
	z = isc_execute(a,&st);
end:	if (a)
		free(a);
	return z;
}
