#if 0
C -o script script.o var.c.o interpreter.c.o -s Interpreter=Main -di interpreter.h -i :: -d mac -d EVAL_FLOATING -d ALLOW_RUNON_WSP -d ALLOW_LONG_COMMENTS -d ALLOW_ELITERALS -d ALLOW_ESCNL -d ALLOW_DSTMOD -d ALLOW_REP {"Parameters"}
Exit 0
#endif
/*
 script.c
  basic interpreter (no functions yet)

  Created: 2011.08.20
  Version: 1.9r2
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "interpreter.h"

/* ---- mid-level */

#if TARGET_OS_MAC
 #define commandPeriod()		(*((volatile char*)0x179)&*((volatile char*)0x17A))
#else
 #define commandPeriod()		0 /* you lose... */
#endif

static uint8	user_abort, modal;

void global_idle_proc(void)
{
	if (!commandPeriod() && !user_abort)
		return;
	if (user_abort && !modal)
	{
		fputs("\n# User abort.\n",stderr);
		exit(1);
	}
	user_abort = 1;
}

/* ---- evaluation */

static evalvalue evalindir(evalvalue v)
{
	return ((unsigned)v&0xFFFF)*12345;
}

/* ---- script */

/* converts tabs to spaces so the pointer (^) places right */
static void untab(char *s,
	const char *e)
{
	if (!s)
		return;
	while(s<e)
		if (*s++=='\t')
			s[-1] = ' ';
}

const char *err_strs[] =
{
	"syntax error",
	"memory error",
	"read error",
	"write error",
	"incomplete expression",
	"interrogator already open",
	"bad/missing identifier",
	"handler not supplied",
	"stack over/underflow",
	"value out of range",
	"comment still open",
	"unterminated string",
	"identifier expected here",
	"internal op error",
	"cannot place labels here",
	"language not supported",
	"no such macro",
	"cannot optimize this",
	"value already set to something else",
	"destination store is a number",
	"data too large for store",
	"no such struct/list field",
	"no expression for loop"
};

static void report_error(int err,
	char *s,
	long z,
	long r)
{
	untab(s,s+z);
	if (err==_e_param_separation)
		err = _e_general_error;
	if (err<0 && (err=~err)<GET_ARRAY_COUNT(err_strs))
		fprintf(stderr,
"### %.*s\n"
"### % *^\n"
"### %s\n",
			z,s,r+1,err_strs[err]);
}

//s = malloc(0);
//while(z=~0,l=x,f && (o=ftell(f),s=freadline(f,&z,s,NULL,&x,'\\'))!=NULL)

static long evaluate_script(char *s,
	const char *e,
	void *u)
{
	evalvalue	v = 0;
	char		*le, *lasts;
	long		lastl;
	uint32		l, x;
	long		r;
	int		err = 0;
#ifdef ALLOW_LONG_COMMENTS
	uint8		com = 0;
#endif
	for(x=l=1,lasts=NULL,lastl=-1; s<e; global_idle_proc())
	{
		uint32		z = sreadline(s,e-s,&x,'\\');

		le = s+z, r = evaluate_line(s,le,&v,&err,
#ifdef ALLOW_LONG_COMMENTS
			&com,
#endif
			eval_getv,eval_setv,u,evalindir);
		if (err)
			goto erro;
		if (r>=_e_repeat_end)
		{
			if (r==_e_return)
				break;
			if (r==_e_repeat)
				{x = l; continue;}
			if (r==_e_repeat_last)
			{
				if (!lasts)
					goto expo;
				s = lasts, x = lastl;
				continue;
			}
			if (!lasts)
			{
			expo:	r = 0, err = _e_no_expr;
			erro:	report_error(err,s,z,r);
				goto end;
			}
		} else
		{
			if (s+r<e && s[r]==',')
				{report_error(_e_general_error,s,z,r); goto end;}
			if (r) /* substantive line */
				lasts = s, lastl = l;
		}
		l = x, s = le+1;
	}
#ifdef ALLOW_LONG_COMMENTS
	if (com)
		fprintf(stderr,
"### %s\n",
			err_strs[~_e_open_comment]);
#endif
	else
	{
		printf("# result: %.16g\n# %d %s\n",
			(double_t)v,l,l==1 ? "line":"lines");
		l = 0;
	}
end:	return l;
}

static int delim(const char *s,
	const char *e)
{
	while(s<e)
	{
		uint8		c = *s++;

		if ((c-='\t')!=0 && c!=' '-'\t')
			{--s; break;}
	}
	return s+3>e || *s!='>' || s[1]!=';';
}

static void dumpIDs(const uint8 *l)
{
	char		c;

	if (!l)
		return;
	fputc('\n',stdout);
	for(; (c=*l++)!=0; l+=sizeof(evalvalue))
	{
		const char	*s;
		evalvalue	v;
		int		z = *l++;

		s = (const char*)l, l+= z, ++l;
		memcpy(&v,l,sizeof(v));
		fprintf(stdout,"%c%.*s%* = %.16g\n",
			(c&0x3F)|0x40,z,s,9-MIN(z,9),v);
	}
}

static long do_script(FILE *f,
	char *s,
	int8 dump)
{
	void		*vbuf = NULL; /* will be list_item */
	int32		z = 0;
	int		err = 0;

	eval_setv("pi",(char*)-1,3.14159265358979323846,&err,&vbuf); /* circumf./diameter */
	eval_setv("C",(char*)-1,299792458,&err,&vbuf); /* light in vaccuum, m/sec */
	if (err)
		return err;
	if (f)
		s = freadfile(f,&z,f==stdin ? delim:NULL);
	else
		z = strlen(s);
	err = !s ? (z ? _e_memory_error:0)
		:evaluate_script(s,s+z,&vbuf);
	if (dump)
		dumpIDs(vbuf);
	eval_free(vbuf);
	if (f)
		free(s);
	return err;
}

int main(int argc,
	char **argv)
{
	FILE		*f = NULL;
	const char	*fn = NULL;
	char		**a, *s, *str = NULL;
	int		r = 0;
	int8		dump = 0;

	if (argc<=0 || (a=argv)==NULL)
		return 0;
	if (argc==1)
	{
		fprintf(stderr,"### usage: %s filename\n",
			*argv);
		fputs(
"#\n"
"#  -            use stdin\n"
"#  -D           dump IDs\n"
"#  -E<expr>     evaluate expression\n"
"#\n",
			stderr);
		return 1;
	}
	while(--argc && (s=*++a)!=NULL)
		if (*s=='-')
		{
			if (!s[1])
			{
				if (f || str)
					goto opn;
				f = stdin, s = "dev::stdin";
			} else if (s[1]=='E')
			{
				if (f || str)
					goto opn;
				str = s+2;
			} else if (s[1]=='D')
				dump = 1;
			else
			{
				fprintf(stderr,"### %s - unknown option '%s'.\n",
					*argv,s);
				return 1;
			}
		} else if (f)
		opn:	fprintf(stderr,"### %s - one input at a time.\n",
				*argv);
		else if ((f=fopen(fn=s,"r"))==NULL)
		{
			fprintf(stderr,"### %s - could not get/open '%s'.\n",
				*argv,s);
			return 1;
		}
	if (!str && !fn)
	{
		fprintf(stderr,"### %s - no input.\n",
			*argv);
		return 1;
	}
#if !TARGET_API_MAC_CARBON && TARGET_OS_MAC
	if (((volatile char*)0x174)[60>>3]&0x80>>(60&7)) /* keymap - control */
		Debugger();
#endif
	r = do_script(f,str,dump);
	if (f!=stdin)
		fclose(f);
	if (r)
	{
		if (r<0)
			fprintf(stderr,"### %s - fatal error (%d).\n",
				*argv,r);
		else if (f)
			fprintf(stderr,
"File \"%s\"; line %d\n",
				fn,r);
		r = 1;
	}
	return r;
}
